#!/bin/bash

#usage: $o <file to save expeirment id into> program [args...]
# executes program (several times) and measures various performance counter events. Saves the whole lot into a huge log file. The lines for this experiment are identified by <experiment_id> which is saved into the file given as the first argument

#settings file must contain the exports of the following variables:
# DETLOG  filename of the detailed log file. THe individual events are written into this log file
# TEMPDIR is a directory (without the trailing /) where to save temporary files (small files for communicating between processes of the script)
# LIMITS is an argument sequence to ulimit


# read settings
if [ -z "$FIMIDIR" ] ; then
    if . ./settings ; then
        x=1 # do nothing
    else
        echo Error: either create ./settings or set FIMIDIR variable to the directory of it >&2
        exit 1
    fi
else 
    if . $FIMIDIR/settings ; then
        x=1
    else
        echo Error: $FIMIDIR/settings not found. Set FIMIDIR variable correctly. >&2
        exit 1
    fi
fi


#calculate experiment ID
EXPIDFILE=$1
shift
export EXPID=`date +%Y%m%d%H%M%S_%s%N`
echo Experiment id is $EXPID >&2
#save experiment id for the caller to process
echo $EXPID > $EXPIDFILE

export PROGARGS="$@"


# this runs a single instance of the program.
# arg 1 is the counters to measure (with the spaces between the different arguments to perfex)
# writes the counter event counts into the destination log file and fd 3
# arg 2 is the run ID
# arg 3 is a space-separated list of the counters measured by arg1 (not including tsc which is always measured)
function runprog() {
    COUNTERS="$1"
    RUNID=$2
    TEMPPRGFILE=$TEMPDIR/perfexout.$$.$RUNID
    DESC=($3)
    ( ulimit $LIMITS ; perfex -o $TEMPPRGFILE $COUNTERS $PROGARGS )
    STATUS=$?
    echo $EXPID:status:$RUNID:$STATUS >> $DETLOG
    if (( $STATUS )) ; then
        return 1
    fi
    exec 4< $TEMPPRGFILE
    read a b <&4
    echo $EXPID:tsc:$RUNID:$b >> $DETLOG
    idx=0
    while read a b c <&4 ; do 
        if [ "$a" != "event" ] ; then
            break; #this is the PEBS_blabla lines
        fi
        echo $EXPID:${DESC[$idx]}:$RUNID:$c >> $DETLOG
        idx=$(( $idx + 1 ))
    done
    rm -f $TEMPPRGFILE
    return 0
}


CTRS=""
DESC=""

CTRS="$CTRS -e 0x00410023"
DESC="$DESC LS2_buffer_full"
CTRS="$CTRS -e 0x00410265"
DESC="$DESC WC_buffer_flush_plus_streaming_store"
CTRS="$CTRS -e 0x00418065"
DESC="$DESC streaming_store"
CTRS="$CTRS -e 0x0041074b"
DESC="$DESC prefetch_instr"

runprog "$CTRS" 1 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x00410040"
DESC="$DESC L1D_access"
CTRS="$CTRS -e 0x00411e42"
DESC="$DESC L1D_miss_L2_hit"
CTRS="$CTRS -e 0x00410142"
DESC="$DESC L1D_miss_L2_miss"
CTRS="$CTRS -e 0x00410047"
DESC="$DESC L1D_misalign_access"

runprog "$CTRS" 2 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x00410045"
DESC="$DESC L1DTBL_miss_L2DTBL_hit"
CTRS="$CTRS -e 0x00410046"
DESC="$DESC L1DTBL_miss_L2DTBL_miss"
CTRS="$CTRS -e 0x00410167"
DESC="$DESC hw_prefetch_cachehit"
CTRS="$CTRS -e 0x00410267"
DESC="$DESC hw_prefetch_all"

runprog "$CTRS" 5 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x0041016d"
DESC="$DESC FSB_write_8byte"
CTRS="$CTRS -e 0x0041077e"
DESC="$DESC L2_miss_all"
CTRS="$CTRS -e 0x0041047e"
DESC="$DESC L2_miss_TLB_walk"
CTRS="$CTRS -e 0x004100c0"
DESC="$DESC instr_all"

runprog "$CTRS" 4 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x01c100c1"
DESC="$DESC uop_histo_all_0"
CTRS="$CTRS -e 0x02c100c1"
DESC="$DESC uop_histo_all_1"
CTRS="$CTRS -e 0x03c100c1"
DESC="$DESC uop_histo_all_2"
CTRS="$CTRS -e 0x00410076"
DESC="$DESC uop_histo_all_3"

runprog "$CTRS" 3 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x004100c2"
DESC="$DESC branch_all"
CTRS="$CTRS -e 0x004100c3"
DESC="$DESC branch_mispred"
CTRS="$CTRS -e 0x00410ecb"
DESC="$DESC mmx_oper"
#CTRS="$CTRS -e 0x004100c6"
#DESC="$DESC far_control_transfer"
CTRS="$CTRS -e 0x004100c1"
DESC="$DESC uop_all"

#CTRS="$CTRS -e 0x00410076"
#DESC="$DESC ticks_not_halted"

runprog "$CTRS" 6 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x004100d0"
DESC="$DESC stall_decoder_empty"
CTRS="$CTRS -e 0x004100d1"
DESC="$DESC stall_dispatch"
CTRS="$CTRS -e 0x004100d2"
DESC="$DESC stall_branch_abort"
CTRS="$CTRS -e 0x004107e3"
DESC="$DESC dram_turnaround_all"

runprog "$CTRS" 7 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS -e 0x004100d5"
DESC="$DESC stall_reorder_buffer_full"
CTRS="$CTRS -e 0x004100d6"
DESC="$DESC stall_resrvstat_full"
CTRS="$CTRS -e 0x004100d7"
DESC="$DESC stall_fpu_full"
CTRS="$CTRS -e 0x004100d1"
DESC="$DESC stall_LS_full"

runprog "$CTRS" 8 "$DESC" || exit 1


CTRS=""
DESC=""
CTRS="$CTRS -e 0x004101e0"
DESC="$DESC dram_page_hit"
CTRS="$CTRS -e 0x004102e0"
DESC="$DESC dram_page_miss"
CTRS="$CTRS -e 0x004104e0"
DESC="$DESC dram_page_conflict"
CTRS="$CTRS -e 0x004100e1"
DESC="$DESC dram_page_table_overflows"

runprog "$CTRS" 9 "$DESC" || exit 1


exit 0

CTRS=""
DESC=""
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "

runprog "$CTRS" 10 "$DESC" || exit 1

CTRS=""
DESC=""
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "
CTRS="$CTRS "
DESC="$DESC "

runprog "$CTRS" 10 "$DESC" || exit 1



CTRS="$CTRS -e 0x0103d000/0x06508204@0x80000001"
DESC="$DESC FSB_write"
CTRS="$CTRS -e 0x0103d000/0x07504204@0x80000002"
DESC="$DESC FSB_read_plus_prefetch"
CTRS="$CTRS -e 0x00031000/0x12000204@0x80000004"
DESC="$DESC uop_from_TC_build"
CTRS="$CTRS -e 0x00031000/0x12000404@0x80000006"
DESC="$DESC uop_from_TC_deliver"
CTRS="$CTRS -e 0x00035000/0x0a003804@0x80000005"
DESC="$DESC mispred_noncond_branch"
CTRS="$CTRS -e 0x00035000/0x08000804@0x80000007"
DESC="$DESC total_function_calls"
CTRS="$CTRS -e 0x00033000/0x02004004@0x8000000c"
DESC="$DESC resource_stall_store_buffer"
CTRS="$CTRS -e 0x0003b000/0x10000204@0x80000010"
DESC="$DESC store_nbogus"
CTRS="$CTRS -e 0x00035000/0x04000804@0x80000011"
DESC="$DESC tag_store1"
CTRS="$CTRS -e 0x0003B000/0x12000604@0x8000000E --p4pe=0x01000001 --p4pmv=0x1"
DESC="$DESC L1_read_miss"
CTRS="$CTRS -e 0x00039000/0x02000204@0x8000000d"
DESC="$DESC uops_nbogus"
CTRS="$CTRS -e 0x00039000/0x04000604@0x8000000f"
DESC="$DESC instr_nbogus"


runprog "$CTRS" 1 "$DESC" || exit 1

CTRS=""
DESC=""

CTRS="$CTRS -e 0x0003b000/0x04001804@0x80000008" 
DESC="$DESC memory_cancel_all"
CTRS="$CTRS -e 0x00035000/0x10000604@0x80000009"
DESC="$DESC memory_complete_split_access"
CTRS="$CTRS -e 0x0003b000/0x0a000604@0x8000000a"
DESC="$DESC wc_buffer_evictions"
CTRS="$CTRS -e 0x00033000/0x35000004@0x8000000b"
DESC="$DESC 128bit_mmx_oper"
CTRS="$CTRS -e 0x00031000/0x12000804@0x80000004"
DESC="$DESC uop_from_ROM"
CTRS="$CTRS -e 0x0103d000/0x0750c204@0x80000000"
DESC="$DESC fsb_r_w_prefetch"
CTRS="$CTRS -e 0x0007d000/0x3550c204@0x80000002"
DESC="$DESC fsb_tickof_r_w_prefetch_pending"
CTRS="$CTRS -e 0x00035000/0x06007404@0x80000001"
DESC="$DESC MOB_load_replay"
#
CTRS="$CTRS -e 0x00035000/0x0a000404@0x80000005"
DESC="$DESC branch_mispred_conditional"
CTRS="$CTRS -e 0x00035000/0x08000404@0x80000006"
DESC="$DESC branch_all_conditional"
CTRS="$CTRS -e 0x0003B000/0x10000204@0x8000000C"
DESC="$DESC load_nbogus"
CTRS="$CTRS -e 0x00035000/0x04000404@0x8000000D"
DESC="$DESC tag_load2"
CTRS="$CTRS -e 0x0003B000/0x12000604@0x8000000E --p4pe=0x01000002 --p4pmv=0x1"
DESC="$DESC L2_read_miss"
CTRS="$CTRS -e 0x00039000/0x02000404@0x8000000F"
DESC="$DESC uops_bogus"
CTRS="$CTRS -e 0x00039000/0x04001804@0x80000010"
DESC="$DESC instr_bogus"

runprog "$CTRS" 2 "$DESC" || exit 1

CTRS=""
DESC=""


#for run 3
CTRS="$CTRS -e 0x0103d000/0x0650c204@0x80000000"
DESC="$DESC fsb_r_w"
CTRS="$CTRS -e 0x0007d000/0x3450c204@0x80000002"
DESC="$DESC fsb_tickof_r_w_pending"


CTRS="$CTRS -e 0x00035000/0x08002004@0x80000007"
DESC="$DESC branch_indirect"

CTRS="$CTRS -e 0x000f9000/0x02000604@0x8000000c"
DESC="$DESC uop_histo_all_0"
CTRS="$CTRS -e 0x001f9000/0x02000604@0x8000000d"
DESC="$DESC uop_histo_all_1"
CTRS="$CTRS -e 0x002f9000/0x02000604@0x80000010"
DESC="$DESC uop_histo_all_2"
#
CTRS="$CTRS -e 0x000f9000/0x02000204@0x8000000e"
DESC="$DESC uop_histo_nbogus_0"
CTRS="$CTRS -e 0x001f9000/0x02000204@0x8000000f"
DESC="$DESC uop_histo_nbogus_1"
CTRS="$CTRS -e 0x002f9000/0x02000204@0x80000011"
DESC="$DESC uop_histo_nbogus_2"


runprog "$CTRS" 3 "$DESC" || exit 1