#!/bin/bash

# usage: $0 fimilogfile database support ["nbogus"]

# read settings
if [ -z "$FIMIDIR" ] ; then
    if . ./settings ; then
        x=1 # do nothing
    else
        echo Error: either create ./settings or set FIMIDIR variable to the directory of it >&2
        exit 1
    fi
else 
    if . $FIMIDIR/settings ; then
        x=1
    else
        echo Error: $FIMIDIR/settings not found. Set FIMIDIR variable correctly. >&2
        exit 1
    fi
fi

fin=$1
db=$2
supp=$3


if [ ! -f "$fin" ] ; then
	echo usage: $0 fimilogfile database support bogusspec
        echo if you give \"nbogus\" as the 4th parameter, the graph will be created only nbogus uops.
	exit 1
fi

fimilogf=$TEMPDIR/supplogf.$$
detlogf=$TEMPDIR/sdbf.$$
plotfile=./plotrunnerb.gnuplot

grep "$db" $fin | grep ":$supp:" | sort > $fimilogf
sort $DETLOG > $detlogf
if [ "$4" == "nbogus" ] ; then
    asp="nbogus"
    title="nbogus uops on "
else
    asp="all"
    title="all uops on "
fi
    
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep uop_histo_${asp}_0 | awk ' BEGIN{FS=":"} {print $5,NR,$4/1e9}' > /tmp/all-0-$db-$supp.$$
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep uop_histo_${asp}_1 | awk ' BEGIN{FS=":"} {print $5,NR,$4/1e9}' > /tmp/all-1-$db-$supp.$$
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep uop_histo_${asp}_2 | awk ' BEGIN{FS=":"} {print $5,NR,$4/1e9}' > /tmp/all-2-$db-$supp.$$
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep tsc:3 | awk ' BEGIN{FS=":"} {print $5,NR,$4/1e9}' > /tmp/all-3-$db-$supp.$$

join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep fsb_tickof_r_w_prefetch_pending | awk ' BEGIN{FS=":"} {print $5,NR+0.375,$4/1e9}' > /tmp/rwpref-$db-$supp.$$
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep fsb_tickof_r_w_pending | awk ' BEGIN{FS=":"} {print $5,NR+0.375,$4/1e9}' > /tmp/rw-$db-$supp.$$

cat $detlogf | grep uops_nbogus > /tmp/uopnb.$$
cat $detlogf | grep uops_bogus > /tmp/uopb.$$
join -t : -1 1 -2 1 /tmp/uopnb.$$ /tmp/uopb.$$ > /tmp/uopall.$$
join -t : -1 1 -2 1 /tmp/uopall.$$  $fimilogf  | awk ' BEGIN{FS=":"} {print $8,NR,($4+$7)/1e9}' > /tmp/uopall-$db-$supp.$$
join -t : -1 1 -2 1 $detlogf  $fimilogf  | grep uops_nbogus | awk ' BEGIN{FS=":"} {print $5,NR,$4/1e9}' > /tmp/uopnb-$db-$supp.$$


if [ "${PLOTEPS:-n}" != "n" ] ; then 
    extx="eps";

    echo "set terminal postscript eps mono" > $plotfile
    echo "set size 0.7,0.7" >>$plotfile

else
    extx="png";
    echo "set terminal png  small picsize 720 600 " > $plotfile
fi

cat >>$plotfile <<EOF
	set encoding iso_8859_1
	set nologscale
	set data style boxes
	set key right top
	set ylabel "GClockticks"
       set boxwidth 0.5
EOF

if [ "${PLOTEPS:-n}" != "n" ] ; then 

    echo "set style fill pattern" >> $plotfile

else
    echo "set style fill solid" >> $plotfile
fi




count=`cat /tmp/all-0-$db-$supp.$$ | wc -l`
echo nump is $count file is /tmp/all-0-$db-$supp.$$
if [ "$count" == 0 ] ; then
    echo Error: no data ponts. Check that the given database and support value has measurements! >&2
    exit 1
fi
echo "set xrange[0.5:${count}.75]" >> $plotfile
echo "set yrange[0:*]" >> $plotfile
echo "set angles degrees" >> $plotfile

if [ "${PLOTEPS:-n}" != "n" ] ; then 
    echo "set key right top" >> $plotfile
    echo "set bmargin 3" >> $plotfile
    echo -n "set xtics axis rotate by -10 (" >> $plotfile

else 
    echo "set xtics (1 " " )" >> $plotfile
    echo -n "set x2tics rotate (" >> $plotfile
fi

awk 'BEGIN{first=1} {if (!first) printf",";first=0; printf "\"%s\" %d",$1,$2}' < /tmp/all-0-$db-$supp.$$ >>$plotfile
echo ")" >> $plotfile
echo "set title \"$title $db at $supp\"" >> $plotfile
echo "set output \"outputs/stall-${asp}-${db}-${supp}.${extx}\"" >> $plotfile
echo "plot \"/tmp/all-3-$db-$supp.$$\" using 2:3 title \"3 uops/tick\",  \"/tmp/all-2-$db-$supp.$$\" using 2:3 title \"2 uops/tick\", \"/tmp/all-1-$db-$supp.$$\" using 2:3 title \"1 uop/tick\", \"/tmp/all-0-$db-$supp.$$\" using 2:3 title \"stall\", \"/tmp/uopall-$db-$supp.$$\" using 2:3:(0.25) title \"bogus uops\", \"/tmp/uopnb-$db-$supp.$$\" using 2:3:(0.25) title \"nbogus uops\", \"/tmp/rwpref-$db-$supp.$$\" using 2:3:(0.25) title \"prefetch pending\", \"/tmp/rw-$db-$supp.$$\" using 2:3:(0.25) title \"r/w pending\"" >> $plotfile

gnuplot $plotfile