#ifndef SeqAprioriSelector_HPP
#define SeqAprioriSelector_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/SeqFrequent2Filter.cpp"


#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/sequence/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/trie/trie_manipulators/sequence/support_counter/SupportCounterLookupSeq.hpp"
#include "apriori/OneByOneSupportCounter.hpp"

#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/sequence/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/sequence/SimplePruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/sequence/IntersectProPruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/sequence/CandGenInfreqRemoveNopruneMerge.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/InfreqRemover.hpp"
#include "apriori/bodon/trie/trie_manipulators/InfreqRemover.hpp"
#include "apriori/Apriori.hpp"


template <class TRIE, class LEAF, class T_R, class DF_D> 
class SeqAprioriSelector
{
   public:
      SeqAprioriSelector( 
	 counter_t min_supp, char* algorithm, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);
};

template <class TRIE, class LEAF, class T_R, class DF_D>
SeqAprioriSelector<TRIE, LEAF, T_R, DF_D>::SeqAprioriSelector( 
   counter_t min_supp, char* algorithm, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef typename bracz::BuildTreeDBCache< Coder<T_R, DF_D>,  
      std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;
//   typedef Coder<T_R, DF_D>  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);

   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   SeqFrequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE main_trie;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, false> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::sequence::SupportCounterLookupSeq<TRIE> SUPP_C_BASE;
//   typedef Bodon::sequence::SupportCounterLookupEdge<TRIE> SUPP_C_BASE;
//   typedef Bodon::sequence::SupportCounterLookupEdgeCommute<TRIE> SUPP_C_BASE;
//   typedef Bodon::sequence::SupportCounterMerge<TRIE> SUPP_C_BASE;
//   typedef Bodon::sequence::SupportCounterMerge2<TRIE> SUPP_C_BASE;
//   typedef Bodon::sequence::SupportCounterIterArray<TRIE> SUPP_C_BASE;
   typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;

   typedef bracz::singleualloc<LEAF, 64 * 1024> LEAF_ALLOCATOR;

   char *kind=algorithm+7;
   const NEELevel NEE = NEE_Off;

   if( strstr(kind,"-noprune") )
   {
      log_status(0,"Complete pruning is disabled.");
      typedef Bodon::sequence::FrequentPairInserterNoprune<DF_D, TRIE, LEAF, LEAF_ALLOCATOR> FPI;
      typedef Bodon::sequence::inhomogeneous_trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, LEAF, LEAF_ALLOCATOR> CGIR;

      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.")
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
   }
   else if( strstr(kind,"-intersectprune") )
   {
      log_status(0,"Intersection-based pruning is selected.");
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::sequence::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF, LEAF_ALLOCATOR> PRUNER;
      typedef Bodon::sequence::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF, LEAF_ALLOCATOR> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF, LEAF_ALLOCATOR, NEE> IR;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else
   {
      log_status(0,"Simple pruning is selected.");
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::sequence::inhomogeneous_trie::SimplePruner<DF_D, TRIE, LEAF, LEAF_ALLOCATOR> PRUNER;
      typedef Bodon::sequence::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF, LEAF_ALLOCATOR> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF, LEAF_ALLOCATOR, NEE> IR;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.")
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
   }
}

#endif
