/***************************************************************************
                          CandidateGeneratorNoprune.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorNoprune_prefix_HPP
#define CandidateGeneratorNoprune_prefix_HPP

/**
  *@author Ferenc Bodon
  */


#include "datastructures/PrefixArrayList.hpp"
#include "apriori/bodon-vector/CandidateGeneratorNoprune.hpp"
#include "common/log.h"
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace prefix_array
{
   template <class DATASTRUCTURE, class D, class DUMMY>
   class CandidateGeneratorNoprune 
   {
      protected:
	 DATASTRUCTURE& candidates; 
	 D& decoder;
	 DUMMY& dummy; 
	 bool newCandidate;
      public:
	 CandidateGeneratorNoprune<DATASTRUCTURE, D, DUMMY> 
	 (DATASTRUCTURE& candidates, D& decoder, DUMMY& dummy) :
	    candidates(candidates), decoder(decoder), dummy(dummy){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void afterWorkCandGen(){}

	 bool isThereAnyCandidate()
	 {
	    return newCandidate;
	 }
      protected:
	 void generateCandidateSizeThree();
   };


   template <class DATASTRUCTURE, class D, class DUMMY> inline
   void CandidateGeneratorNoprune<DATASTRUCTURE, D, DUMMY>::
   generateCandidate(const unsigned int candidate_size)
   {
      newCandidate = false;
      if(candidate_size == 3)
	 generateCandidateSizeThree();
      else
      {
	 for( typename DATASTRUCTURE::iterator it1 = candidates.begin();
	      it1 != candidates.end(); ++it1)
	    for(diag_array_row_t::iterator it2 = (*it1).begin();it2 != (*it1).end(); ++it2)
	    {
	       Bodon::vector_based::CandidateGeneratorNoprune<D, DUMMY> 
		  cand_gen(*it2, decoder, dummy);
	       cand_gen.generateCandidate(candidate_size);
	       if(!(*it2).empty())
		  newCandidate = true;
	    }
      }
   }

   template <class DATASTRUCTURE, class D, class DUMMY> inline
   void CandidateGeneratorNoprune<DATASTRUCTURE, D, DUMMY>::
   generateCandidateSizeThree()
   {
      itemvector new_candidate;
      for(typename DATASTRUCTURE::size_type index1 = 0; 
	  index1 != candidates.size(); ++index1)
	 for(diag_array_row_t::size_type index2
		= 0;index2 < candidates[index1].size(); ++index2)
	 {
	    if(!candidates[index1][index2].empty())
	    {
	       candidates[index1][index2].pop_back();
	       for(diag_array_row_t::size_type index3
		      = index2 + 1; index3 < candidates[index1].size(); ++index3)
		  if(!candidates[index1][index3].empty())
		  {
		     new_candidate.push_back(index3 + index1 + 1);
		     candidates[index1][index2].push_back(itemvecCounterPair(new_candidate, 0));
		     newCandidate = true;
		     new_candidate.pop_back();	       
		  }
	    }
	 }
   }
}
}
#endif

