/***************************************************************************
                          CandidateGeneratorSimplePrune.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorSimplePrune_HPP
#define CandidateGeneratorSimplePrune_HPP

/**
  *@author Ferenc Bodon
  */


#include "apriori/bodon-vector/vector_typedef.hpp"
#include "common/log.h"
//#include <cstdio>
//#include <iterator>	//for test




bool lessItemvector( const itemvecCounterPair& elem1, 
		     const itemvecCounterPair& elem2 )
{
	return elem1.first < elem2.first;
}

namespace Bodon
{
namespace vector_based
{
   template <class D, class DUMMY>
   class CandidateGeneratorSimplePrune 
   {
      protected:
	 cand_vector_t& candidates; 
      public:
	 CandidateGeneratorSimplePrune<D, DUMMY> 
	 ( cand_vector_t& candidates, D& decoder, DUMMY& dummy) :
	    candidates(candidates){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void afterWorkCandGen(){}

	 bool isThereAnyCandidate()
	 {
	    return !candidates.empty();
	 }
   };


   template <class D, class DUMMY> inline
   void CandidateGeneratorSimplePrune<D, DUMMY>::
   generateCandidate(const unsigned int candidate_size)
   {
      if(!candidates.empty())
      {
	 cand_vector_t::size_type orig_size(candidates.size());
	 itemvector new_candidate;
	 cand_vector_t::size_type index=0, index2;
	 for(;index < orig_size -1; ++index) // minden genertorra
	 {
	    new_candidate.clear();
	    new_candidate.insert(new_candidate.end(), // j jellt az egyik genertorbl kpzdik
				 candidates[index].first.begin(), // genertor els eleme
				 candidates[index].first.end()); // genertor utls eleme
	    index2 = index +1;
	    while(index2 < orig_size && // minden nagyobb genertorra, amg a prefixek megyeznek
		  candidates[index].first.haveSamePrefix(candidates[index2].first))
	    {
		  new_candidate.push_back(candidates[index2].first.back()); // az j jellt vgre fzi a msik genertor utols elemt
		  // innentl rtam n a kdot
		  // Kollr Balzs, N2ZZL4
		  // ide kell majd egy elgazs, hogy csak akkor veszi fl jelltnek,
		  // ha minden egyes kisebb mret rszhalmaza szerepel a cand_vector-ban
		  // a new_candidate elemeit egyesvel kiveszi, s ellenrzi, hogy a maradk benne van-e
		  // a candidates vektor els felben
		  bool isAllSubsetFrequent = true;
		  std::vector<item_t>::iterator item_it = new_candidate.end()-1;
		  register item_t deleted_item = *item_it, temp_item;
		  new_candidate.erase(item_it);
		  while (isAllSubsetFrequent && item_it != new_candidate.begin())
		  {
			// 0..orig_size-1 intervallumban kell keresni new_candidate-et a candidates vektorban
			// btw elg lower_bound..orig_size-1 intervallumban is keresni,
			// ahol lower_bound az elz keress eredmnye
			// de a binary_search nem adja vissza, hogy milyen indexen tallta meg a vektorban amit keresett
			if (!std::binary_search(
					candidates.begin(),
					candidates.begin()+orig_size,
					//new_candidate // ez lenne, ha a lessItemVector funktor mkdne
							itemvecCounterPair(new_candidate, 0)
				   ,::lessItemvector // ez is akkor lenne
					)
			   )
					// negyedik funktor paramter azrt nem kell, mert a pair funktora megfelel
					//  mindig tud majd dnteni a pr els eleme alapjn
					// ksbb ezt ki kell cserlni sajt funktorra, hogy ne kelljen prt adni a
					// binris keresnek paramterl
			{
				isAllSubsetFrequent = false; // ez volt a ciklustrzs utols futsa
			}
			else // csak akkor kell a ciklusvltozt kezelni, ha lefut mg a ciklustrzs
			{
				--item_it;
				temp_item = *item_it;
				*item_it = deleted_item;
				deleted_item = temp_item;
			}
		  }
		  new_candidate.insert(item_it, deleted_item); // vissza a trlt elemet
		  if (isAllSubsetFrequent)
		  // idig tart az n ltalam rt kd
		  // Kollr Balzs, N2ZZL4
		  {
			candidates.push_back( // felveszi a jelltek kz az j jelltet nulls tmogatottsggal
			     itemvecCounterPair(new_candidate, 0));
		  }
		  new_candidate.pop_back(); // levgja az j jelltbl a genertor utols elemt
		  ++index2;
	    }
	 }
	 candidates.erase(candidates.begin(), candidates.begin()+orig_size); // trli a genertorokat
      }
   }
}
}
#endif
