/***************************************************************************
                          Trie.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef Trie_HPP
#define Trie_HPP

/**
  *@author Ferenc Bodon
  */

enum MatchFind
{
   /// Merge scan
   MERGE,
   /// lookup by index
   LOOKUP,
   /// LOOKUP by index if that is preferred, otherwise MERGE
   HYBRID
};


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/InnerNodeSpecific.hpp"
#include <vector>
//#include <iterator>	//for test


namespace Bodon
{
   /** This struct represent an edge of a Trie.
       
   An edge has a label, and an edge points to a subtrie.
   */

   template <class LEAF, class EDGELIST>
   class Trie : public LEAF, public InnerNodeSpecific<EDGELIST>
   {
      public:
	 static std::vector<item_t> neelist;
	 Trie(){}
	 Trie(const counter_t counter) : LEAF(counter), 
					 InnerNodeSpecific<EDGELIST>() {}


	 bool neeFind(item_t item) const{return true;} //this should never be called

	 void neePushBackSorted(const std::vector<item_t>& new_neelist){} //this should never be called
	 void neeAddSorted(const std::vector<item_t>& new_neelist){} //this should never be called
	 void neeInsertItem(const item_t new_item){} //this should never be called
	 /// It decides whether the given itemset is included in the trie or not.
	 const Trie* isIncluded( const std::vector<item_t>& an_itemset, 
				 std::vector<item_t>::const_iterator item_it ) const;
   };
   template <class LEAF, class EDGELIST> std::vector<item_t> Trie<LEAF, EDGELIST>::neelist;
   
/**
   \param an_itemstd::set The given itemstd::set.
   \param item_it This iterator shows the element of the basket 
   that has to be processed.
   \return NULL, if the itemstd::set is not included, otherwise the trie, 
   that represents the itemstd::set.
*/

   template <class LEAF, class EDGELIST> const Trie<LEAF, EDGELIST>* Trie<LEAF, EDGELIST>::
   isIncluded( const std::vector<item_t>& an_itemset, 
      std::vector<item_t>::const_iterator item_it ) const
   {
      if( item_it == an_itemset.end() ) 
	 return this;
      else
      {
	 Trie<LEAF, EDGELIST>* subtrie = static_cast<Trie<LEAF, EDGELIST>*>(
	    InnerNodeSpecific<EDGELIST>::edgelist.find(*item_it));
	 if(subtrie)
	    return subtrie->isIncluded( an_itemset, ++item_it );
	 else 
	    return NULL;
      }
   }
}

#endif
