/***************************************************************************
                          TrieNEE.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef TrieNEE_HPP
#define TrieNEE_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "apriori/bodon/InnerNodeSpecific.hpp"
#include <vector>
//#include <iterator>	//for test


namespace Bodon
{
   /** This struct represent an edge of a TrieNEE.
       
   An edge has a label, and an edge points to a subtrie.
   */
   template <class TRIE>
   class TrieNEE : public TRIE
   {
      public:
	 std::vector<item_t> neelist;	 
	 TrieNEE():TRIE(){}
	 TrieNEE(counter_t counter): TRIE(counter){}

	 /// It decides whether the given itemset is included in the trie or not.
	 const TrieNEE<TRIE>* isIncluded( 
	    const std::vector<item_t>& an_itemset, 
	    std::vector<item_t>::const_iterator item_it ) const;

	 bool neeFind(const item_t item) const
	 {
	    std::vector<item_t>::const_iterator it(neelist.begin());
	    while(it != neelist.end() && *it < item)
	       ++it;
	    if(it != neelist.end() && *it == item)
	       return true;
	    else
	       return false;
	 }
	 void neePushBackSorted(const std::vector<item_t>& new_neelist)
	 {
	    neelist.reserve(neelist.size() + new_neelist.size());
	    neelist.insert(neelist.end(), new_neelist.begin(),new_neelist.end());

	 }
	 void neeAddSorted(const std::vector<item_t>& new_neelist)
	 {
	    neePushBackSorted(new_neelist);
	    std::inplace_merge(
	       neelist.begin(), neelist.end() - new_neelist.size(), neelist.end());
	 }
	 void neeInsertItem(const item_t new_item)
	 {
	    std::vector<item_t>::iterator it(neelist.begin());
	    while(it != neelist.end() && *it < new_item)
	       ++it;
	    neelist.insert(it, new_item);
	 }
   };
   

/**
   \param an_itemstd::set The given itemstd::set.
   \param item_it This iterator shows the element of the basket 
   that has to be processed.
   \return NULL, if the itemstd::set is not included, otherwise the trie, 
   that represents the itemstd::set.
*/

   template <class TRIE> const TrieNEE<TRIE>* 
   TrieNEE<TRIE>::isIncluded( 
      const std::vector<item_t>& an_itemset, 
      std::vector<item_t>::const_iterator item_it ) const
   {
      if( item_it == an_itemset.end() ) 
	 return this;
      else
      {
	 TrieNEE<TRIE>* subtrie(static_cast<TrieNEE<TRIE>*>(
					     TRIE::edgelist.find(*item_it)));
	 if(subtrie)
	    return subtrie->isIncluded( an_itemset, ++item_it );
	 else 
	 {
	    if(neeFind(*item_it))
	       return isIncluded( an_itemset, ++item_it );
	    else 
	       return NULL;			
	 }
      }
   }
}

#endif
