/***************************************************************************
                          CandGenInfreqRemoveNopruneMerge.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandGenInfreqRemoveNopruneMerge_HPP
#define CandGenInfreqRemoveNopruneMerge_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/log.h"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandidateGeneratorNoprune.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE>
   class CandGenInfreqRemoveNopruneMerge : 
	 public CandidateGeneratorNoprune<
      DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>
   {
      private:
	 typedef CandidateGeneratorNoprune<
	    DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE> PARENT;
      protected:
	 unsigned int nr_of_deleted;
	 unsigned int total_nr_of_deleted;
	 unsigned int nr_of_prefix_equiitem;
	 unsigned int total_nr_of_prefix_equiitem;

      public:
	 static unsigned int total_nr_of_candidates;
	 CandGenInfreqRemoveNopruneMerge( 
	    TRIE_OEL& main_trie, DF_D& df_decoder, LEAF_ALLOCATOR& s_alloc ) : 
	    PARENT(main_trie, df_decoder, s_alloc), total_nr_of_deleted(0), 
	    total_nr_of_prefix_equiitem(0){}

	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void deleteInfrequent( const counter_t min_supp, 
				unsigned int candidate_size );
	 void afterWorkDel();
      protected:

//	 template <class TRIE> void writeTrie(
//	    TRIE* subtrie, unsigned int step_to_cand_par);

	 /// Deletes the tries that represent infrequent itemsets.
	 template <class TRIE> void delete_infrequent_subtrie( 
	    TRIE* subtrie, const counter_t min_supp,
	    unsigned int step_to_cand_par );

	 template <class TRIE> void delete_infrequent_subtrieNEE( 
	    TRIE* subtrie, const counter_t min_supp,
	    unsigned int step_to_cand_par );


      public:
	 // No public members
   };

   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   unsigned int CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::total_nr_of_candidates;

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidate(const unsigned int candidate_size)
   {
      if(candidate_size == 3)
      {
	 PARENT::nr_of_new_candidates = 0;
	 if(NEE > NEE_Off)
	 {
	    generateCandidateFindParentNEE( &this->main_trie,
					    candidate_size -2 );
	 }
	 else
	    generateCandidateFindParent( &this->main_trie, 
					 candidate_size -2 );
	 log_info_det(0,"Number of newly generated candidates: %d", 
		      PARENT::nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_INFO
	 total_nr_of_candidates += nr_of_new_candidates;
#endif
      }
   }

/**
   \param min_supp The threshold of absolute support.
   \param candidate_size The size of the candidate itemstd::set.
*/
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   deleteInfrequent( const counter_t min_supp, unsigned int candidate_size )
   {
      assert(candidate_size > 1);
      nr_of_deleted = 0;
      PARENT::nr_of_new_candidates = 0;
      if(NEE > NEE_Off)
      {
	 nr_of_prefix_equiitem = 0;
	 PARENT::df_decoder.pushEquisupportItemSet(
	    this->main_trie.neelist);
	 delete_infrequent_subtrieNEE<TRIE_OEL>( &this->main_trie, min_supp,
						 candidate_size - 1 );
#if DEBUG_LEVEL >= LEVEL_INFO
	 total_nr_of_prefix_equiitem += nr_of_prefix_equiitem;
#endif

      }
      else
	 delete_infrequent_subtrie<TRIE_OEL>( &this->main_trie, min_supp, 
					      candidate_size - 1 );
      log_info_det(0,"Number of infrequent candidates: %d", nr_of_deleted);
      log_info_det(0,"Number of newly generated candidates: %d", 
	       PARENT::nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_INFO
      total_nr_of_candidates += nr_of_new_candidates;
      total_nr_of_deleted += nr_of_deleted;
#endif
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, 
					LEAF, LEAF_ALLOCATOR, NEE>::
   afterWorkDel()
   {
      log_info(0,"The total number of infrequent candidates: %d", 
	      total_nr_of_deleted);
      log_info(0,"The total number of generated candidates: %d", 
	       total_nr_of_candidates);
      if(NEE > NEE_Off)
	 log_info(0,"The total number of prefix equisupport item: %d", 
		  total_nr_of_prefix_equiitem);
   }

/**
   \param min_supp The support threshold
*/
   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, 
					LEAF, LEAF_ALLOCATOR, NEE>::
   delete_infrequent_subtrie( TRIE* subtrie, const counter_t min_supp,
			      unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       delete_infrequent_subtrie<TRIE_OEL>( 
		  static_cast<TRIE_OEL*>((*itEdge).second),
		  min_supp, step_to_cand_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       delete_infrequent_subtrie<TRIE_OI>( 
		  static_cast<TRIE_OI*>((*itEdge).second),
		  min_supp, step_to_cand_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    PARENT::df_decoder.popItem();

	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_INFO
		  ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else
	       ++itEdge;
	 }
	 generateCandidateAtParent(subtrie);
      }
   }
   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, 
					LEAF, LEAF_ALLOCATOR, NEE>::
   delete_infrequent_subtrieNEE( TRIE* subtrie, const counter_t min_supp,
				 unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OEL*>((*itEdge).second)->neelist);
	       delete_infrequent_subtrieNEE( 
		  static_cast<TRIE_OEL*>((*itEdge).second), min_supp,
		  step_to_cand_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() )
	       {
		  PARENT::df_decoder.write( 
		     static_cast<TRIE_OEL*>((*itEdge).second)->getCounter()  & TWO_POW31_1);
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OI*>((*itEdge).second)->neelist);		 
	       delete_infrequent_subtrieNEE( 
		  static_cast<TRIE_OI*>((*itEdge).second), min_supp,
		  step_to_cand_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() )
	       {
		  PARENT::df_decoder.write( static_cast<TRIE_OI*>((*itEdge).second)->getCounter());
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    PARENT::df_decoder.popItem();
	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_INFO
	       ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else if(static_cast<LEAF*>((*itEdge).second)->getCounter() == 
		    (subtrie->getCounter() & TWO_POW31_1) )
	    {
	       subtrie->neelist.push_back((*itEdge).first);
	       PARENT::df_decoder.pushEquisupportItem((*itEdge).first);
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
#if DEBUG_LEVEL >= LEVEL_INFO
	       ++nr_of_prefix_equiitem;
#endif
	    }
	    else
	       ++itEdge;
	 }
	 generateCandidateAtParentNEE(subtrie);
      }
   }
/*   This function is for debugging!
   template <class DF_D, class TRIE_OEL, class TRIE_OI, NEELevel NEE> 
   template <class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, NEE>::
   writeTrie(TRIE* subtrie, unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    cout<<"Edge with label "<<(*itEdge).first<<" leads to ";
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       cout<<"an OEL with counter"<< (static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31_1)<<endl;;
	       writeTrie<TRIE_OEL>(static_cast<TRIE_OEL*>((*itEdge).second), step_to_cand_par);
	    }
	    else
	    {
	       cout<<"an OI with counter"<<static_cast<LEAF*>((*itEdge).second)->getCounter()<<endl;;
	       writeTrie<TRIE_OI>(static_cast<TRIE_OI*>((*itEdge).second), step_to_cand_par);
	    }
	    ++itEdge;
	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    cout<<"Edge with label "<<(*itEdge).first<<" leads to Leaf!"<<endl;;
	    cout<<"with counter: "<<(static_cast<Leaf*>((*itEdge).second)->getCounter())<<endl;
	    ++itEdge;
	 }
      }

      }*/
}
}
#endif
