/***************************************************************************
                          CandidateGeneratorNoprune.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorNoprune_HPP
#define CandidateGeneratorNoprune_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/log.h"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE>
   class CandidateGeneratorNoprune : public Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR>
   {
      private:
	 typedef Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR> PARENT;
      protected:
	 std::vector<Edge> extenders;
	 std::vector<Edge> replace_list;

	 unsigned int nr_of_new_candidates;
	 

      public:
	 CandidateGeneratorNoprune( TRIE_OEL& main_trie, DF_D& df_decoder, 
				    LEAF_ALLOCATOR& s_alloc ) : 
	    PARENT(main_trie, df_decoder, s_alloc){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      	 
	 void afterWorkCandGen();
      protected:

	 template <class TRIE> void generateCandidateAtParent( TRIE* trie );
	 template <class TRIE> void generateCandidateAtParentNEE( 
	    TRIE* trie);

	 /// Generates candidate of size more than two
	 template <class TRIE> void generateCandidateFindParent( 
	    TRIE* trie, unsigned int step_to_freq_par);
	 template <class TRIE> void generateCandidateFindParentNEE( 
	    TRIE* trie, unsigned int step_to_freq_par);

      public:
	 // No public members
   };


   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   inline void CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF, 
					 LEAF_ALLOCATOR, NEE>::
   generateCandidate(unsigned int candidate_size)
   {
      assert(candidate_size > 1);
      if(NEE > NEE_Off)
      {
	 PARENT::df_decoder.pushEquisupportItemSet(
	    PARENT::main_trie.neelist);
	 generateCandidateFindParentNEE<TRIE_OEL>( 
	    &PARENT::main_trie, candidate_size -2 );
      }
      else
	 generateCandidateFindParent<TRIE_OEL>( 
	    &PARENT::main_trie, candidate_size -2 );

   }
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   void CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   afterWorkCandGen()
   {
   }


   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF, 
			     LEAF_ALLOCATOR, NEE>::
   generateCandidateAtParent(TRIE* trie)
   {
      typename TRIE::iterator itEdge2;
      LEAF* toExtendAsLeaf;
      replace_list.clear();
      for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	   itEdge != trie->edgelist.end(); ++itEdge )
      {
	 toExtendAsLeaf = static_cast<LEAF*>((*itEdge).second);
	 PARENT::df_decoder.pushItemWithWrite( (*itEdge).first, 
					       toExtendAsLeaf->getCounter() );
	 PARENT::df_decoder.popItem();
	 itEdge2 = itEdge;
	 ++itEdge2;
	 extenders.clear();
	 for( ; itEdge2 != trie->edgelist.end(); ++itEdge2 )
	 {
	    extenders.push_back(Edge((*itEdge2).first, PARENT::s_alloc.allocate()));
	    static_cast<LEAF*>(extenders.back().second)->setCounter(0);
	       
	 }
	 if(!extenders.empty())
	 {
	    if( 2 * extenders.size() < extenders.back().first - extenders.front().first + 3 )
	    {
	       TRIE_OEL* toExtend = new TRIE_OEL(toExtendAsLeaf->getCounter() | TWO_POW31 );
	       replace_list.push_back(Edge((*itEdge).first, toExtend));
	       toExtend->edgelist.insert(extenders);
	    }
	    else
	    {
	       TRIE_OI* toExtend = new TRIE_OI(toExtendAsLeaf->getCounter());
	       replace_list.push_back(Edge((*itEdge).first, toExtend));
	       toExtend->edgelist.insert(extenders);
	    }
#if DEBUG_LEVEL >= LEVEL_INFO
	    nr_of_new_candidates +=extenders.size();
#endif
	 }
	 PARENT::s_alloc.free(toExtendAsLeaf);
      }
      trie->edgelist.clear();
      trie->edgelist.insert(replace_list);
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF, 
			     LEAF_ALLOCATOR, NEE>::
   generateCandidateAtParentNEE( TRIE* trie )
   {
      typename TRIE::iterator itEdge2;
      LEAF* toExtendAsLeaf;
      replace_list.clear();
      for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	   itEdge != trie->edgelist.end(); ++itEdge )
      {
	 toExtendAsLeaf = static_cast<LEAF*>((*itEdge).second);
	 itEdge2 = itEdge;
	 ++itEdge2;
	 extenders.clear();
	 for( ; itEdge2 != trie->edgelist.end(); ++itEdge2 )
	 {
	    extenders.push_back(Edge((*itEdge2).first, PARENT::s_alloc.allocate()));
	    static_cast<LEAF*>(extenders.back().second)->setCounter(0);
	 }
	 if(extenders.empty())
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    PARENT::df_decoder.write( toExtendAsLeaf->getCounter());
	    PARENT::df_decoder.popItem();
	 }
	 else
	 {
	    if( 2 * extenders.size() < extenders.back().first - extenders.front().first + 3 )
	    {
	       TRIE_OEL* toExtend = new TRIE_OEL(toExtendAsLeaf->getCounter() | TWO_POW31 );
	       replace_list.push_back(Edge((*itEdge).first, toExtend));
	       toExtend->edgelist.insert(extenders);
	    }
	    else
	    {
	       TRIE_OI* toExtend = new TRIE_OI(toExtendAsLeaf->getCounter());
	       replace_list.push_back(Edge((*itEdge).first, toExtend));
	       toExtend->edgelist.insert(extenders);
	    }
#if DEBUG_LEVEL >= LEVEL_PERF
	    nr_of_new_candidates +=extenders.size();
#endif
	 }
	 PARENT::s_alloc.free(toExtendAsLeaf);
      }
      trie->edgelist.clear();
      trie->edgelist.insert(replace_list);
   }


   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> void
   CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, 
			     LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidateFindParent( TRIE* trie, 
				unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       generateCandidateFindParent<TRIE_OEL>( 
		  static_cast<TRIE_OEL*>((*itEdge).second), step_to_freq_par);
	       if(static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty())
	       {
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       generateCandidateFindParent<TRIE_OI>( 
		  static_cast<TRIE_OI*>((*itEdge).second), step_to_freq_par);
	       if(static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty())
	       {
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    PARENT::df_decoder.popItem();
	 }
      }
      else
	 generateCandidateAtParent(trie);
   }
   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> void
   CandidateGeneratorNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF, 
			     LEAF_ALLOCATOR, NEE>::
   generateCandidateFindParentNEE( TRIE* trie, 
				unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OEL*>((*itEdge).second)->neelist);
	       generateCandidateFindParentNEE<TRIE_OEL>( 
		  static_cast<TRIE_OEL*>((*itEdge).second), step_to_freq_par);
	       if(static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty())
	       {
		  PARENT::df_decoder.write( static_cast<TRIE_OEL*>((*itEdge).second)->getCounter()  & TWO_POW31_1);
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(static_cast<TRIE_OI*>((*itEdge).second)->neelist);		  
	       generateCandidateFindParentNEE<TRIE_OI>( 
		  static_cast<TRIE_OI*>((*itEdge).second), step_to_freq_par);
	       if(static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty())
	       {
		  PARENT::df_decoder.write( static_cast<TRIE_OI*>((*itEdge).second)->getCounter());
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    PARENT::df_decoder.popItem();
	 }
      }
      else
	 generateCandidateAtParentNEE(trie);
   }
}
}
#endif
