/***************************************************************************
                          CandidateGeneratorPrune.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorPrune_HPP
#define CandidateGeneratorPrune_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/allocators.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{
   template <class P, class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   class CandidateGeneratorPrune : public P
   {
      protected:
	 unsigned int total_nr_of_candidates;
	 unsigned int total_nr_of_nonprefix_equiitem;

      public:
	 CandidateGeneratorPrune( TRIE_OEL& main_trie, DF_D& df_decoder, 
				  LEAF_ALLOCATOR& s_alloc) : 
	    P(main_trie, df_decoder, s_alloc), 
	    total_nr_of_candidates(0), total_nr_of_nonprefix_equiitem(0){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void afterWorkCandGen();
      protected:

	 /// Generates candidate of size more than two
	 template <class TRIE> void generateCandidateFindParent( 
	    TRIE* trie, std::vector<item_t>& maybe_candidate,
	    unsigned int step_to_freq_par);
	 template <class TRIE> void generateCandidateFindParentNEE( 
	    TRIE* trie, std::vector<item_t>& maybe_candidate,
	    unsigned int step_to_freq_par);
   };


   template <class P, class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   void CandidateGeneratorPrune<P, DF_D, TRIE_OEL, TRIE_OI, LEAF, 
				LEAF_ALLOCATOR, NEE>::
   generateCandidate(unsigned int candidate_size)
   {
      assert(candidate_size > 1);
      P::nr_of_new_candidates = 0;
      std::vector<item_t> maybe_candidate;
      if(NEE > NEE_Off)
      {
	 P::nr_of_nonprefix_equiitem = 0;
	 P::df_decoder.pushEquisupportItemSet(
	    P::main_trie.neelist);
	 generateCandidateFindParentNEE( &this->main_trie, maybe_candidate,
					 candidate_size -2 );      
	 log_info_det(0,"Number of nonprefix equisupport items: %d", P::nr_of_nonprefix_equiitem);
#if DEBUG_LEVEL >= LEVEL_INFO
	 total_nr_of_nonprefix_equiitem += nr_of_nonprefix_equiitem;
#endif

      }
      else
	 generateCandidateFindParent( &this->main_trie, maybe_candidate,
				      candidate_size -2 );
      log_info_det(0,"Number of newly generated candidates: %d", 
	      P::nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_INFO
      total_nr_of_candidates += nr_of_new_candidates;
#endif
   }

   template <class P, class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   void CandidateGeneratorPrune<P, DF_D, TRIE_OEL, TRIE_OI, LEAF, 
				LEAF_ALLOCATOR, NEE>::
   afterWorkCandGen()
   {
      log_info(0,"The total number of generated candidates: %d", 
	       total_nr_of_candidates);
      if(NEE > NEE_Off)
	 log_info(0,"The total number of nonprefix equisupport item: %d", 
		  total_nr_of_nonprefix_equiitem);
   }

   template <class P, class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   template <class TRIE>
   void CandidateGeneratorPrune<P, DF_D, TRIE_OEL, TRIE_OI, 
				LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidateFindParent( TRIE* trie, 
				std::vector<item_t>& maybe_candidate,
				unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    maybe_candidate.push_back((*itEdge).first);
	    P::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       generateCandidateFindParent( static_cast<TRIE_OEL*>((*itEdge).second), 
					    maybe_candidate,
					    step_to_freq_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       generateCandidateFindParent( static_cast<TRIE_OI*>((*itEdge).second), 
					    maybe_candidate,
					    step_to_freq_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    P::df_decoder.popItem();
	    maybe_candidate.pop_back();
	 }
      }
      else
      {
	 generateCandidateAtParent(trie, maybe_candidate);
      }
   }

   template <class P, class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   template <class TRIE>
   void CandidateGeneratorPrune<P, DF_D, TRIE_OEL, TRIE_OI, 
				LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidateFindParentNEE( TRIE* trie, 
				   std::vector<item_t>& maybe_candidate,
				   unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    maybe_candidate.push_back((*itEdge).first);
	    P::df_decoder.pushItem( (*itEdge).first );
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       P::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OEL*>((*itEdge).second)->neelist);
	       generateCandidateFindParentNEE( static_cast<TRIE_OEL*>((*itEdge).second), 
					       maybe_candidate, step_to_freq_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() && 
		   ( NEE == NEE_Full ||
		     NEE == NEE_Level1 && P::df_decoder.nrOfEEItems() < step_to_freq_par + 2 ))
	       {
		  P::df_decoder.write( static_cast<TRIE_OEL*>((*itEdge).second)->getCounter() 
				    & TWO_POW31_1);
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else
		  ++itEdge;
	    }
	    else
	    {
	       P::df_decoder.pushEquisupportItemSet(static_cast<TRIE_OI*>((*itEdge).second)->neelist);		  
	       generateCandidateFindParentNEE( static_cast<TRIE_OI*>((*itEdge).second), 
					       maybe_candidate, step_to_freq_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() && 
		   ( NEE == NEE_Full ||
		     NEE != NEE_Level1 && P::df_decoder.nrOfEEItems() < step_to_freq_par + 2 ) )
	       {
		  P::df_decoder.write( static_cast<TRIE_OI*>((*itEdge).second)->getCounter() );
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else
		  ++itEdge;
	    }
	    P::df_decoder.popItem();
	    maybe_candidate.pop_back();
	 }
      }
      else
	 generateCandidateAtParentNEE(trie, maybe_candidate);

   }
}
}
#endif
