/***************************************************************************
                          FrequentItemInserter.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef FrequentItemInserterDyn_HPP
#define FrequentItemInserterDyn_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/trie/trie_manipulators/ManipulatorBase.hpp"
#include "apriori/bodon/Leaf.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test


namespace Bodon
{
namespace dynamic_trie
{
template <class DF_D, class TRIE_OEL, NEELevel NEE>
class FrequentItemInserter : public Bodon::ManipulatorBase<DF_D, TRIE_OEL>
{
  private:
      typedef Bodon::ManipulatorBase<DF_D, TRIE_OEL> PARENT;
   public:
      FrequentItemInserter( TRIE_OEL& trie, DF_D& df_decoder ) : 
	 PARENT(trie, df_decoder){}

      /// This set the support of the emptyset.
      void setEmptysetSupport(const counter_t emptyset_support)
      {
	 PARENT::main_trie.setCounter(emptyset_support | TWO_POW31);
      }

      /// Insert the frequent items and their counters into the data structure;
      void insertFrequentItems(const std::vector<counter_t>& freq_counters )
      {
	 std::vector<Edge> extenders;
      	 if(NEE > NEE_Off)
	 {
	    unsigned int nr_of_prefix_equiitem=0;
	    for(item_t index = 0; index < freq_counters.size(); ++index)
	    {

	       if( freq_counters[index] == 
		   PARENT::main_trie.getCounter() )
	       {
#if DEBUG_LEVEL >= LEVEL_PERF
		  ++nr_of_prefix_equiitem;
#endif
		  PARENT::main_trie.neelist.push_back(index);
	       }
	       else
		  extenders.push_back(Edge(index, new Bodon::Leaf(freq_counters[index])));
	    }
	    PARENT::main_trie.edgelist.insert(extenders);
//	    log_perf(0,"Number of prefix equisupport items: %d", nr_of_prefix_equiitem); 
	 }
	 else
	 {
	    PARENT::df_decoder.write(PARENT::main_trie.getCounter() 
	       & TWO_POW31_1);
	    
	    for(item_t index = 0; index < freq_counters.size(); ++index)
	    {
	       PARENT::df_decoder.pushItemWithWrite( index, freq_counters[index] );
	       extenders.push_back( Edge(index, new Bodon::Leaf(freq_counters[index])) );
	       PARENT::df_decoder.popItem();
	    }
	    PARENT::main_trie.edgelist.insert(extenders);
	 }
      }
};
}
}

#endif
