/***************************************************************************
                          InfreqRemover.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef InfreqRemover_HPP
#define InfreqRemover_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{

   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   class InfreqRemover : public Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR>
   {
      private:
	 typedef Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR> PARENT;
      protected:
	 std::vector<Edge> extenders;
	 std::vector<item_t> neelist;
	 unsigned int nr_of_deleted;
	 unsigned int total_nr_of_deleted;
	 unsigned int nr_of_prefix_equiitem;
	 unsigned int total_nr_of_prefix_equiitem;

      public:
	 InfreqRemover( TRIE_OEL& main_trie, DF_D& df_decoder,
			LEAF_ALLOCATOR& s_alloc ) : 
	    PARENT(main_trie, df_decoder, s_alloc), 
	    total_nr_of_deleted(0), total_nr_of_prefix_equiitem(0){}

	 /// Deletes infrequent itemsets.
	 void deleteInfrequent( const counter_t min_supp,
				unsigned int candidate_size );
	 void afterWorkDel();
      protected:

	 /// Deletes the tries that represent infrequent itemstd::sets.
	 template <class TRIE> void delete_infrequent_subtrie( 
	    TRIE* subtrie, const counter_t min_supp,
	    unsigned int step_to_cand_par );
	 template <class TRIE> void delete_infrequent_subtrieNEE( 
	    TRIE* subtrie, const counter_t min_supp,
	    unsigned int step_to_cand_par);
   };

/**
   \param min_supp The threshold of absolute support.
*/
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, class LEAF_ALLOCATOR, 
	     NEELevel NEE>
   void InfreqRemover<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   deleteInfrequent( const counter_t min_supp,
		     unsigned int candidate_size )
   {
      assert(candidate_size > 2);
      nr_of_deleted = 0;
      if(NEE > NEE_Off)
      {
	 nr_of_prefix_equiitem = 0;
	 PARENT::df_decoder.pushEquisupportItemSet(
	    PARENT::main_trie.neelist);
	 delete_infrequent_subtrieNEE<TRIE_OEL>( 
	    &this->main_trie, min_supp, candidate_size - 1 );
	 log_info_det(0,"Number of prefix equisupport items: %d", nr_of_prefix_equiitem);
#if DEBUG_LEVEL >= LEVEL_INFO
	 total_nr_of_prefix_equiitem += nr_of_prefix_equiitem;
#endif

      }
      else
	 delete_infrequent_subtrie<TRIE_OEL>( 
	    &this->main_trie, min_supp, candidate_size - 1 );
      log_info_det(0,"Number of infrequent candidates: %d", nr_of_deleted);
#if DEBUG_LEVEL >= LEVEL_INFO
      total_nr_of_deleted += nr_of_deleted;
#endif
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, class LEAF_ALLOCATOR, 
	     NEELevel NEE>
   void InfreqRemover<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   afterWorkDel()
   {
      log_info(0,"The total number of infrequent candidates: %d", 
	      total_nr_of_deleted);
      if(NEE > NEE_Off)
	 log_info(0,"The total number of prefix equisupport item: %d", 
		  total_nr_of_prefix_equiitem);
   }

/**
   \param min_supp The support threshold
*/
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, class LEAF_ALLOCATOR, 
	     NEELevel NEE>
   template <class TRIE>
   void InfreqRemover<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   delete_infrequent_subtrie( TRIE* subtrie, const counter_t min_supp,
			      unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       delete_infrequent_subtrie<TRIE_OEL>( 
		  static_cast<TRIE_OEL*>((*itEdge).second), min_supp, step_to_cand_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	    else
	    {
	       delete_infrequent_subtrie<TRIE_OI>( 
		  static_cast<TRIE_OI*>((*itEdge).second), min_supp, step_to_cand_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() )
	       {
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else ++itEdge;
	    }
	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_INFO
	       ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else
	       ++itEdge;
	 }
      }
   }
   template <class DF_D, class TRIE_OEL, class TRIE_OI, 
	     class LEAF, class LEAF_ALLOCATOR, NEELevel NEE>
   template <class TRIE>
   void InfreqRemover<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   delete_infrequent_subtrieNEE( TRIE* subtrie, const counter_t min_supp,
				 unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem((*itEdge).first);
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() & TWO_POW31 )
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OEL*>((*itEdge).second)->neelist);
	       delete_infrequent_subtrieNEE<TRIE_OEL>( 
		  static_cast<TRIE_OEL*>((*itEdge).second), min_supp,
		  step_to_cand_par);
	       if( static_cast<TRIE_OEL*>((*itEdge).second)->edgelist.empty() &&
		   ( NEE == NEE_Level1 && PARENT::df_decoder.nrOfEEItems() < step_to_cand_par + 1 ||
		   NEE ==NEE_Full && step_to_cand_par != 0 ) )
	       {
		  PARENT::df_decoder.write(
		     static_cast<TRIE_OEL*>((*itEdge).second)->getCounter() & TWO_POW31_1);
		  delete static_cast<TRIE_OEL*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else
		  ++itEdge;
	    }
	    else
	    {
	       PARENT::df_decoder.pushEquisupportItemSet(
		  static_cast<TRIE_OI*>((*itEdge).second)->neelist);		 
	       delete_infrequent_subtrieNEE<TRIE_OI>( 
		  static_cast<TRIE_OI*>((*itEdge).second), min_supp,
		  step_to_cand_par);
	       if( static_cast<TRIE_OI*>((*itEdge).second)->edgelist.empty() && 
		   (NEE == NEE_Level1 && PARENT::df_decoder.nrOfEEItems() < step_to_cand_par + 1  ||
		    NEE ==NEE_Full && step_to_cand_par != 0 ) )
	       {
		  PARENT::df_decoder.write(
		     static_cast<TRIE_OI*>((*itEdge).second)->getCounter());
		  delete static_cast<TRIE_OI*>((*itEdge).second);
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else
		  ++itEdge;
	    }
	    PARENT::df_decoder.popItem();
	 }
      }
      else
      {
	 neelist.clear();
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<LEAF*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_INFO
	       ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else if(static_cast<LEAF*>((*itEdge).second)->getCounter() == 
		    (subtrie->getCounter() & TWO_POW31_1) )
	    {
	       neelist.push_back((*itEdge).first);
	       PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
#if DEBUG_LEVEL >= LEVEL_INFO
	       ++nr_of_prefix_equiitem;
#endif
	    }
	    else
	       ++itEdge;
	 }
	 subtrie->neeAddSorted(neelist);
	 PARENT::df_decoder.pushEquisupportItemSet(neelist);
      }
   }
}
}
#endif
