/***************************************************************************
                          IntersectProPruner.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef IntersectProPruner_HPP
#define IntersectProPruner_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE>
   class IntersectProPruner : public Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR>
   {
      private:
	 typedef Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE_OEL, LEAF_ALLOCATOR> PARENT;
      protected:
	 std::vector<Edge> extenders;
	 mutable std::vector<item_t> ext_items;
	 mutable std::vector<item_t> leaf_neelist;
	 std::vector<Edge> replace_list;

	 unsigned int nr_of_new_candidates;
	 unsigned int nr_of_nonprefix_equiitem;
      public:
	 IntersectProPruner( TRIE_OEL& main_trie, DF_D& df_decoder,
			     LEAF_ALLOCATOR& s_alloc ) : 
	    PARENT(main_trie, df_decoder, s_alloc){}

      protected:
	 
	 template <class TRIE> void intersect( const TRIE* subset_trie ) const;

	 template <class TRIE> void intersectNEE( 
	    const TRIE* subset_trie ) const;

	 void filterNonExtenders( 
	    const std::vector<const TRIE_OEL*>& subset_oel_tries,
	    const std::vector<const TRIE_OI*>& subset_oi_tries,
	    std::vector<typename TRIE_OEL::const_iterator>& 
	    subset_oel_child_iters,
	    std::vector<typename TRIE_OI::const_iterator>& 
	    subset_oi_child_iters,
	    const item_t leaf_item ) const;

	 template <class TRIE> const void* isIncluded( 
	    const TRIE* trie, const std::vector<item_t>& an_itemset, 
	    std::vector<item_t>::const_iterator item_it ) const;

	 bool findSubsetTries( 
	    std::vector<item_t>& itemset, 
	    std::vector<const TRIE_OEL*>& subset_oel_trie,
	    std::vector<const TRIE_OI*>& subset_oi_trie) const;

	 template <class TRIE> void generateCandidateAtParent(
	    TRIE* trie, std::vector<item_t>& maybe_candidate);
	 template <class TRIE> void generateCandidateAtParentNEE(
	    TRIE* trie, std::vector<item_t>& maybe_candidate);
   };

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   intersect( const TRIE* subset_trie) const
   {
      typename TRIE::const_iterator it_e( subset_trie->edgelist.begin() );
      std::vector<item_t>::iterator it_i(ext_items.begin());
      while(it_e != subset_trie->edgelist.end() && it_i != ext_items.end())
      {
	 if(*it_e < *it_i)
	    ++it_e;
	 else if(*it_e > *it_i)
	    it_i = ext_items.erase(it_i);
	 else
	 {
	    ++it_e;
	    ++it_i;
	 }
      }
      ext_items.erase(it_i, ext_items.end());      
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   intersectNEE( const TRIE* subset_trie ) const
   {
      typename TRIE::const_iterator it_e( subset_trie->edgelist.begin() );
      std::vector<item_t>::const_iterator it_nee( subset_trie->neelist.begin() );
      std::vector<item_t>::iterator it_i(ext_items.begin());
      while(it_i != ext_items.end())
      {
	 while (it_e != subset_trie->edgelist.end() && *it_e < *it_i)
	    ++it_e;
	 if(it_e != subset_trie->edgelist.end() && *it_e == *it_i)
	 {
	    ++it_e;
	    ++it_i;
	 }
	 else
	 {
	    while(it_nee != subset_trie->neelist.end() && *it_nee < *it_i)
	       ++it_nee;
	    if(it_nee != subset_trie->neelist.end() && *it_nee == *it_i)
	    {
	       if(NEE == NEE_Full)
	       {
		  leaf_neelist.push_back(*it_i);
		  it_i = ext_items.erase(it_i);
	       }
	       else
	       {
		  ++it_i;
	       }
	       ++it_nee;
	    }
	    else
	       it_i = ext_items.erase(it_i);
	 }
      }
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> inline void 
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   filterNonExtenders( 
      const std::vector<const TRIE_OEL*>& subset_oel_tries,  
      const std::vector<const TRIE_OI*>& subset_oi_tries,  
      std::vector<typename TRIE_OEL::const_iterator>& subset_oel_child_iters,
      std::vector<typename TRIE_OI::const_iterator>& subset_oi_child_iters,
      const item_t leaf_item ) const
   {

      std::vector<item_t>::size_type index = 0;
      for( ; index < subset_oel_tries.size(); ++index)
      {
	 while( subset_oel_child_iters[index] !=  
		subset_oel_tries[index]->edgelist.end() 
		&& (*(subset_oel_child_iters[index])).first < leaf_item)
	    ++subset_oel_child_iters[index];
	 if( subset_oel_child_iters[index] !=  
	     subset_oel_tries[index]->edgelist.end() &&
	     (*(subset_oel_child_iters[index])).first == leaf_item)
	    if( static_cast<LEAF*>((*(subset_oel_child_iters[index])).second)->
		getCounter() & TWO_POW31 )
	       if(NEE != NEE_Off && !static_cast<const TRIE_OEL*>(
			    (*(subset_oel_child_iters[index])).second)->neelist.empty())
		  intersectNEE(static_cast<const TRIE_OEL*>(
			    (*(subset_oel_child_iters[index])).second));
	       else
		  intersect(static_cast<const TRIE_OEL*>(
			       (*(subset_oel_child_iters[index])).second));
	    else
	       if(NEE != NEE_Off && !static_cast<const TRIE_OI*>(
			    (*(subset_oel_child_iters[index])).second)->neelist.empty())
		  intersectNEE(static_cast<const TRIE_OI*>(
				  (*(subset_oel_child_iters[index])).second));
	       else
		  intersect(static_cast<const TRIE_OI*>(
			       (*(subset_oel_child_iters[index])).second));
	 else
	 {
	    ext_items.clear();
	    break;
	 }
      }
      if( index == subset_oel_tries.size())
      {
	 index = 0;
	 for( ; index < subset_oi_tries.size(); ++index)
	 {
	    while( subset_oi_child_iters[index] !=  
		subset_oi_tries[index]->edgelist.end() 
		   && (*(subset_oi_child_iters[index])).first < leaf_item)
	       ++subset_oi_child_iters[index];
	    if( subset_oi_child_iters[index] !=  
		subset_oi_tries[index]->edgelist.end() &&
		(*(subset_oi_child_iters[index])).first == leaf_item)
	       if( static_cast<LEAF*>((*(subset_oi_child_iters[index])).second)->
		   getCounter() & TWO_POW31 )
		  if(NEE != NEE_Off && !static_cast<const TRIE_OEL*>(
			    (*(subset_oi_child_iters[index])).second)->neelist.empty())
		     intersectNEE(static_cast<const TRIE_OEL*>(
				     (*(subset_oi_child_iters[index])).second));
		  else
		     intersect(static_cast<const TRIE_OEL*>(
				  (*(subset_oi_child_iters[index])).second));
	       else
		  if(NEE != NEE_Off && !static_cast<const TRIE_OI*>(
			(*(subset_oi_child_iters[index])).second)->neelist.empty())
		     intersectNEE(static_cast<const TRIE_OI*>(
				     (*(subset_oi_child_iters[index])).second));
		  else
		     intersect( static_cast<const TRIE_OI*>(
			        (*(subset_oi_child_iters[index])).second) );
	    else
	    {
	       ext_items.clear();
	       break;
	    }
	 }
      }
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> const void*
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   isIncluded( const TRIE* trie, const std::vector<item_t>& an_itemset, 
	       std::vector<item_t>::const_iterator item_it ) const
   {
      if( item_it == an_itemset.end() )
	 return static_cast<const void*>(trie);
      else
      {
	 void* subtrie;
	 subtrie = trie->edgelist.find(*item_it);
	 if(subtrie)
	 {
	    if( static_cast<LEAF*>(subtrie)->getCounter() & TWO_POW31 )
	       return isIncluded(static_cast<TRIE_OEL*>(subtrie), an_itemset, ++item_it);
	    else
	       return isIncluded(static_cast<TRIE_OI*>(subtrie), an_itemset, ++item_it);	  
	 }
	 else 
	    if(NEE == NEE_Level1 && trie->neeFind(*item_it) )
	       return isIncluded(trie, an_itemset, ++item_it);	  
	    else
	       return NULL;
      }
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> inline bool
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   findSubsetTries( std::vector<item_t>& itemset, 
		    std::vector<const TRIE_OEL*>& subset_oel_tries,
		    std::vector<const TRIE_OI*>& subset_oi_tries ) const
   {
      assert(!itemset.empty());
      std::vector<item_t>::iterator item_it = itemset.end()-1;
      register item_t deleted_item = *item_it, temp_item;
      const void* a_subset_trie;

      itemset.pop_back();
      a_subset_trie = isIncluded( &this->main_trie, itemset, itemset.begin() );
      if(a_subset_trie)
	 if( static_cast<const LEAF*>(a_subset_trie)->
	     getCounter() & TWO_POW31 )
	    subset_oel_tries.push_back(static_cast<const TRIE_OEL*>(a_subset_trie));
	 else
	    subset_oi_tries.push_back(static_cast<const TRIE_OI*>(a_subset_trie));
      else 
      {
	 itemset.push_back(deleted_item);
	 return false;
      }
      while(item_it != itemset.begin() )
      {
	 --item_it;
	 temp_item = *item_it;
	 *item_it = deleted_item;
	 deleted_item = temp_item;
	 a_subset_trie = isIncluded( &this->main_trie, itemset, itemset.begin() );
	 if(a_subset_trie)
	    if( static_cast<const LEAF*>(a_subset_trie)->
		getCounter() & TWO_POW31 )
	       subset_oel_tries.push_back(static_cast<const TRIE_OEL*>(a_subset_trie));
	    else
	       subset_oi_tries.push_back(static_cast<const TRIE_OI*>(a_subset_trie));
	 else
	 {
	    itemset.insert( item_it, deleted_item );
	    return false;
	 }
      }
      itemset.insert( item_it, deleted_item );
      return true;
   }

   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidateAtParent(TRIE* trie, std::vector<item_t>& maybe_candidate)
   {
      std::vector<const TRIE_OEL*> subset_oel_tries;
      std::vector<const TRIE_OI*> subset_oi_tries;
      const bool all_subset_included = findSubsetTries( 
	 maybe_candidate, subset_oel_tries, subset_oi_tries);

      std::vector<typename TRIE_OEL::const_iterator> subset_oel_child_iters;
      std::vector<typename TRIE_OI::const_iterator> subset_oi_child_iters;
      subset_oel_child_iters.reserve(subset_oel_tries.size());
      subset_oi_child_iters.reserve(subset_oi_tries.size());

      for(typename std::vector<const TRIE_OEL*>::const_iterator it = 
	     subset_oel_tries.begin(); it !=  subset_oel_tries.end(); ++it)
	 subset_oel_child_iters.push_back((*it)->edgelist.begin());

      for(typename std::vector<const TRIE_OI*>::const_iterator it = 
	     subset_oi_tries.begin(); it !=  subset_oi_tries.end(); ++it)
	 subset_oi_child_iters.push_back((*it)->edgelist.begin());

      typename TRIE::iterator itEdge2;
      LEAF* toExtendAsLeaf;
      replace_list.clear();
      for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	   itEdge != trie->edgelist.end(); ++itEdge )
      {
	 toExtendAsLeaf = static_cast<LEAF*>((*itEdge).second);
	 PARENT::df_decoder.pushItemWithWrite( (*itEdge).first, 
				       toExtendAsLeaf->getCounter() );
	 PARENT::df_decoder.popItem();
	 if(all_subset_included)
	 {
	    itEdge2 = itEdge;
	    ++itEdge2;
	    ext_items.clear();
	    for( ; itEdge2 != trie->edgelist.end(); ++itEdge2 )
	       ext_items.push_back((*itEdge2).first);
	    filterNonExtenders( subset_oel_tries, subset_oi_tries, 
				subset_oel_child_iters, subset_oi_child_iters,
				(*itEdge).first );
	    extenders.clear();
	    for(std::vector<item_t>::iterator it = ext_items.begin();
		it != ext_items.end(); ++it)
	    {
	       extenders.push_back(Edge(*it, PARENT::s_alloc.allocate()));
	       static_cast<LEAF*>(extenders.back().second)->setCounter(0);
	    }
	    if(!extenders.empty())
	    {
	       if( 2 * extenders.size() < extenders.back().first - extenders.front().first + 3 )
	       {
		  TRIE_OEL* toExtend = new TRIE_OEL(toExtendAsLeaf->getCounter() | TWO_POW31);
		  replace_list.push_back(Edge((*itEdge).first, toExtend));
		  toExtend->edgelist.insert(extenders);
	       }
	       else
	       {
		  TRIE_OI* toExtend = new TRIE_OI(toExtendAsLeaf->getCounter());
		  replace_list.push_back(Edge((*itEdge).first, toExtend));
		  toExtend->edgelist.insert(extenders);
	       }
#if DEBUG_LEVEL >= LEVEL_INFO
	       nr_of_new_candidates +=extenders.size();
#endif
	    }
	 }
	 PARENT::s_alloc.free(toExtendAsLeaf);
      }
      trie->edgelist.clear();
      trie->edgelist.insert(replace_list);
   }
   template <class DF_D, class TRIE_OEL, class TRIE_OI, class LEAF, 
	     class LEAF_ALLOCATOR, NEELevel NEE> 
   template <class TRIE> inline void 
   IntersectProPruner<DF_D, TRIE_OEL, TRIE_OI, LEAF, LEAF_ALLOCATOR, NEE>::
   generateCandidateAtParentNEE( TRIE* trie, 
				 std::vector<item_t>& maybe_candidate )
   {
      std::vector<const TRIE_OEL*> subset_oel_tries;
      std::vector<const TRIE_OI*> subset_oi_tries;
      replace_list.clear();
      if( findSubsetTries( maybe_candidate, subset_oel_tries, subset_oi_tries ) )
      {
	 std::vector<typename TRIE_OEL::const_iterator> subset_oel_child_iters;
	 std::vector<typename TRIE_OI::const_iterator> subset_oi_child_iters;
	 subset_oel_child_iters.reserve(subset_oel_tries.size());
	 subset_oi_child_iters.reserve(subset_oi_tries.size());
	 
	 for(typename std::vector<const TRIE_OEL*>::const_iterator it = 
		subset_oel_tries.begin(); it !=  subset_oel_tries.end(); ++it)
	    subset_oel_child_iters.push_back((*it)->edgelist.begin());
	 
	 for(typename std::vector<const TRIE_OI*>::const_iterator it = 
		subset_oi_tries.begin(); it !=  subset_oi_tries.end(); ++it)
	    subset_oi_child_iters.push_back((*it)->edgelist.begin());

	 typename TRIE::iterator itEdge2;
	 LEAF* toExtendAsLeaf;
	 for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	      itEdge != trie->edgelist.end(); ++itEdge )
	 {
	    toExtendAsLeaf = static_cast<LEAF*>((*itEdge).second);
	    itEdge2 = itEdge;
	    ++itEdge2;
	    ext_items.clear();
	    leaf_neelist.clear();
	    for( ; itEdge2 != trie->edgelist.end(); ++itEdge2 )
	       ext_items.push_back((*itEdge2).first);
	    filterNonExtenders( subset_oel_tries, subset_oi_tries, 
				subset_oel_child_iters, subset_oi_child_iters,
				(*itEdge).first );
	    std::sort(leaf_neelist.begin(), leaf_neelist.end());
	    extenders.clear();
	    for(std::vector<item_t>::iterator it = ext_items.begin();
		it != ext_items.end(); ++it)
	    {
	       extenders.push_back(Edge(*it, PARENT::s_alloc.allocate()));
	       static_cast<LEAF*>(extenders.back().second)->setCounter(0);
	    }
	    if(extenders.empty() && leaf_neelist.empty() && 
	       ( NEE == NEE_Full  || 
		 NEE == NEE_Level1 && PARENT::df_decoder.EEListEmpty() )
	       )
	    {
	       PARENT::df_decoder.pushItemWithWrite( (*itEdge).first,
					     toExtendAsLeaf->getCounter() );
	       PARENT::df_decoder.popItem();
	    }
	    else
	    {
	       if( extenders.empty() || 
		   2 * extenders.size() < extenders.back().first - extenders.front().first + 3 )
	       {
		  TRIE_OEL* toExtend = new TRIE_OEL(toExtendAsLeaf->getCounter() | TWO_POW31);
		  toExtend->edgelist.insert(extenders);
		  toExtend->neePushBackSorted( leaf_neelist );
		  replace_list.push_back(Edge((*itEdge).first, toExtend));
	       }
 	       else
	       {
		  TRIE_OI* toExtend = new TRIE_OI(toExtendAsLeaf->getCounter());
		  toExtend->edgelist.insert(extenders);
		  toExtend->neePushBackSorted( leaf_neelist );
		  replace_list.push_back(Edge((*itEdge).first, toExtend));
	       }
#if DEBUG_LEVEL >= LEVEL_INFO
	       nr_of_new_candidates +=extenders.size();
	       nr_of_nonprefix_equiitem += leaf_neelist.size();
#endif
	    }
	    PARENT::s_alloc.free(toExtendAsLeaf);
	 }
	 trie->edgelist.clear();
	 trie->edgelist.insert(replace_list);
      }
      else if( NEE == NEE_Full || PARENT::df_decoder.EEListEmpty())
      {
	 for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	      itEdge != trie->edgelist.end(); ++itEdge )
	 {
	    PARENT::df_decoder.pushItemWithWrite( 
	       (*itEdge).first, 
	       static_cast<LEAF*>((*itEdge).second)->getCounter() );
	    PARENT::df_decoder.popItem();
	    PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	 }
	 trie->edgelist.clear();
      }
      else
      {
	 TRIE_OEL* toExtend;
	 for( typename TRIE::iterator itEdge( trie->edgelist.begin() ); 
	   itEdge != trie->edgelist.end(); ++itEdge )
	 {
	    toExtend = new TRIE_OEL( 
	       static_cast<LEAF*>((*itEdge).second)->getCounter() | TWO_POW31);
	    PARENT::s_alloc.free(static_cast<LEAF*>((*itEdge).second));
	    replace_list.push_back(Edge((*itEdge).first, toExtend));
	 }
	 trie->edgelist.clear();
	 trie->edgelist.insert(replace_list);
      }
   }
}
}
#endif
