/***************************************************************************
                          SupportCounter.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounter_HPP
#define SupportCounter_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace dynamic_trie
{
   template <class TRIE_OEL, class TRIE_OI>
   class SupportCounter
   {
      private:
	 typedef Bodon::Leaf LEAF;
      public:
	 SupportCounter( ){}

	 void updateCounters(
	    TRIE_OEL& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }
      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE_OEL* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
	 void findCandidates( TRIE_OI* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE_OEL, class TRIE_OI> void 
   SupportCounter<TRIE_OEL, TRIE_OI>::findCandidates( 
      TRIE_OEL* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      typename TRIE_OEL::iterator it_edge(subtrie->edgelist.begin());
      if( step_to_candidate ) 
      {
	 while( it_edge != subtrie->edgelist.end() && 
		it_basket != it_basket_upper_bound )
	 {
	    if( (*it_edge).first < *it_basket) ++it_edge;
	    else if( (*it_edge).first > *it_basket) ++it_basket;
	    else
	    {
	       ++it_basket;
	       if(static_cast<LEAF*>((*it_edge).second)->getCounter() & TWO_POW31)
		  findCandidates( static_cast<TRIE_OEL*>((*it_edge).second),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr);
	       else
		  findCandidates( static_cast<TRIE_OI*>((*it_edge).second),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr);
	       ++it_edge;
	    }
	 }
      }
      else
      {
	 while( it_edge != subtrie->edgelist.end() && 
		it_basket != it_basket_upper_bound )
	 {
	    if( (*it_edge).first < *it_basket) ++it_edge;
	    else if( (*it_edge).first > *it_basket) ++it_basket;
	    else
	    {
	       ++it_basket;
	       static_cast<LEAF*>((*it_edge).second)
		  ->increaseCounter( counter_incr);
	       ++it_edge;
	    }
	 }
      }

   }

   template <class TRIE_OEL, class TRIE_OI> void 
   SupportCounter<TRIE_OEL, TRIE_OI>::findCandidates( 
      TRIE_OI* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      void* subsubtrie;
      item_t largest_label = subtrie->edgelist.largestEdgelabel();
      if( step_to_candidate ) 
      {
	 while( it_basket != it_basket_upper_bound && *it_basket <= largest_label)
	 {
	    subtrie->edgelist.lookupNoUppercheck(*it_basket, subsubtrie);
	    ++it_basket;
	    if(subsubtrie)
	       if(static_cast<LEAF*>(subsubtrie)->getCounter() & TWO_POW31)
		  findCandidates( static_cast<TRIE_OEL*>(subsubtrie),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr );
	       else
		  findCandidates( static_cast<TRIE_OI*>(subsubtrie),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr );
	 }
      }
      else
      {
	 while( it_basket != it_basket_upper_bound && *it_basket <= largest_label)
	 {
	    subtrie->edgelist.lookupNoUppercheck(*it_basket, subsubtrie);
	    ++it_basket;
	    if(subsubtrie)
	       static_cast<LEAF*>(subsubtrie)
		  ->increaseCounter( counter_incr);
	 }
      }
   }

}
}
#endif
