/***************************************************************************
                          CandidateGeneratorPrune.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorPrune_HPP
#define CandidateGeneratorPrune_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/allocators.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace inhomogeneous_trie
{
   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE = true>
   class CandidateGeneratorPrune : public P
   {
      protected:
	 unsigned int nr_of_new_candidates;
	 unsigned int nr_of_all_candidates;

      public:
	 CandidateGeneratorPrune( TRIE& main_trie, DF_D& df_decoder, 
				  LEAF_ALLOCATOR& s_alloc) : 
	    P(main_trie, df_decoder, s_alloc), 
	    nr_of_all_candidates(0){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void afterWorkCandGen();
	 bool isThereAnyCandidate() const
	 {
	    if(DEADENDPRUNE)
	       return P::isThereAnyCandidate();
	    else
	       return P::is_there_any_new_candidate;
		  
	 }
      protected:

	 /// Generates candidate of size more than two
	 void generateCandidateFindParent( 
	    TRIE* trie, std::vector<item_t>& maybe_candidate,
	    unsigned int step_to_freq_par);
	 void generateCandidateFindParentNEE( 
	    TRIE* trie, std::vector<item_t>& maybe_candidate,
	    std::vector<item_t>& NEEsum,
	    unsigned int step_to_freq_par);
	 void generateCandidateFindParentNEENoDeadend( 
	    TRIE* trie, std::vector<item_t>& maybe_candidate,
	    unsigned int step_to_freq_par);
   };


   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE>
   void CandidateGeneratorPrune<P, DF_D, TRIE, LEAF_ALLOCATOR, 
				NEE, DEADENDPRUNE>::
   generateCandidate(unsigned int candidate_size)
   {
      assert(candidate_size > 1);
      nr_of_new_candidates = 0;
      P::is_there_any_new_candidate = false;
      std::vector<item_t> maybe_candidate;
      if(NEE > NEE_Off)
      {
	 if(DEADENDPRUNE)
	 {
	    std::vector<item_t> NEEsum;
	    generateCandidateFindParentNEE( &this->main_trie, maybe_candidate,
					    NEEsum, candidate_size -2 );      
	 }
	 else
	    generateCandidateFindParentNEENoDeadend( 
	       &this->main_trie, maybe_candidate, candidate_size -2 );
      }
      else
	 generateCandidateFindParent( &this->main_trie, maybe_candidate,
				      candidate_size -2 );
      log_dbg(0,"Number of newly generated candidates: %d", 
	      nr_of_new_candidates);

#if DEBUG_LEVEL >= LEVEL_DBG
      nr_of_all_candidates += nr_of_new_candidates;
#endif
   }

   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE>
   void CandidateGeneratorPrune<P, DF_D, TRIE, LEAF_ALLOCATOR, 
				NEE, DEADENDPRUNE>::
   afterWorkCandGen()
   {
      log_dbg(0,"The total number of generated candidates (in del. p.): %d", 
	      nr_of_all_candidates);
   }

   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE>
   void CandidateGeneratorPrune<P, DF_D, TRIE, LEAF_ALLOCATOR, 
				NEE, DEADENDPRUNE>::
   generateCandidateFindParent( TRIE* trie, 
				std::vector<item_t>& maybe_candidate,
				unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    maybe_candidate.push_back((*itEdge).first);
	    P::df_decoder.pushItem( (*itEdge).first );
	    generateCandidateFindParent( static_cast<TRIE*>((*itEdge).second), 
					 maybe_candidate,
					 step_to_freq_par);
	    if( DEADENDPRUNE && 
	        static_cast<TRIE*>((*itEdge).second)->edgelist.empty() )
	    {
	       delete static_cast<TRIE*>((*itEdge).second);
	       itEdge = trie->edgelist.erase(itEdge);
	    }
	    else ++itEdge;
	    P::df_decoder.popItem();
	    maybe_candidate.pop_back();
	 }
      }
      else
      {
	 generateCandidateAtParent(trie, maybe_candidate);
#if DEBUG_LEVEL >= LEVEL_DBG
	 for( typename TRIE::iterator it( trie->edgelist.begin() );
	      it != trie->edgelist.end(); ++it)
	    nr_of_new_candidates += 
	       static_cast<TRIE*>((*it).second)->edgelist.size();
#endif
      }
   }

   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE>
   void CandidateGeneratorPrune<P, DF_D, TRIE, LEAF_ALLOCATOR, 
				NEE, DEADENDPRUNE>::
   generateCandidateFindParentNEE( TRIE* trie, 
				   std::vector<item_t>& maybe_candidate,
				   std::vector<item_t>& NEEsum,
				   unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 NEEsum.insert( NEEsum.end(), 
			trie->neelist.begin(), trie->neelist.end() );
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    maybe_candidate.push_back((*itEdge).first);
	    P::df_decoder.pushItem( (*itEdge).first );
	    generateCandidateFindParentNEE( static_cast<TRIE*>((*itEdge).second), 
					    maybe_candidate, NEEsum, 
					    step_to_freq_par);
	    if( static_cast<TRIE*>((*itEdge).second)->edgelist.empty() )
	    {
	       NEEsum.insert( NEEsum.end(), 
			      static_cast<TRIE*>((*itEdge).second)->neelist.begin(), 
			      static_cast<TRIE*>((*itEdge).second)->neelist.end() );
	       if( NEEsum.size() < step_to_freq_par + 2 )
	       {
		  P::df_decoder.writeNEE( static_cast<TRIE*>((*itEdge).second)->getCounter(), 
				       NEEsum);
		  NEEsum.resize( NEEsum.size() - 
				 static_cast<TRIE*>((*itEdge).second)->neelist.size() );
		  delete static_cast<TRIE*>((*itEdge).second);
		  itEdge = trie->edgelist.erase(itEdge);
	       }
	       else
	       {
		  NEEsum.resize( NEEsum.size() - 
				 static_cast<TRIE*>((*itEdge).second)->neelist.size() );
		  ++itEdge;
	       }
	    }
	    else ++itEdge;
	    P::df_decoder.popItem();
	    maybe_candidate.pop_back();
	 }
	 NEEsum.resize( NEEsum.size() - trie->neelist.size() );
      }
      else
      {
	 generateCandidateAtParentNEE(trie, maybe_candidate, NEEsum);
#if DEBUG_LEVEL >= LEVEL_DBG
	 for( typename TRIE::iterator it( trie->edgelist.begin() );
	      it != trie->edgelist.end(); ++it)
	    nr_of_new_candidates += 
	       static_cast<TRIE*>((*it).second)->edgelist.size();
#endif

      }

   }

   template <class P, class DF_D, class TRIE, class LEAF_ALLOCATOR, 
	     NEELevel NEE, bool DEADENDPRUNE>
   void CandidateGeneratorPrune<P, DF_D, TRIE, LEAF_ALLOCATOR, 
				NEE, DEADENDPRUNE>::
   generateCandidateFindParentNEENoDeadend( 
      TRIE* trie, std::vector<item_t>& maybe_candidate,
      unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    maybe_candidate.push_back((*itEdge).first);
	    generateCandidateFindParentNEENoDeadend( static_cast<TRIE*>((*itEdge).second), 
					    maybe_candidate, step_to_freq_par);
	    ++itEdge;
	    maybe_candidate.pop_back();
	 }
      }
      else
      {
	 generateCandidateAtParentNEENoDeadend(trie, maybe_candidate);
#if DEBUG_LEVEL >= LEVEL_DBG
	 for( typename TRIE::iterator it( trie->edgelist.begin() );
	      it != trie->edgelist.end(); ++it)
	    nr_of_new_candidates += 
	       static_cast<TRIE*>((*it).second)->edgelist.size();
#endif

      }

   }
}
}
#endif
