/***************************************************************************
                          CandidateGeneratorNoprune.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandidateGeneratorNoprune_HPP
#define CandidateGeneratorNoprune_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/log.h"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace sequence
{
namespace inhomogeneous_trie
{
   template <class DF_D, class TRIE, class LEAF, class LEAF_ALLOCATOR>
   class CandidateGeneratorNoprune : public 
   Bodon::inhomogeneous_trie::ManipulatorBase<DF_D, TRIE, LEAF_ALLOCATOR>
   {
      private:
	 typedef Bodon::inhomogeneous_trie::ManipulatorBase<DF_D, TRIE, LEAF_ALLOCATOR> PARENT;
      protected:
	 std::vector<Edge> extenders;
	 unsigned int nr_of_new_candidates;
	 unsigned int nr_of_all_candidates;
	 std::vector<Edge> replace_list;

      public:
	 CandidateGeneratorNoprune( TRIE& main_trie, DF_D& df_decoder,
				    LEAF_ALLOCATOR& s_alloc ) :
	    Bodon::inhomogeneous_trie::ManipulatorBase<DF_D, TRIE, LEAF_ALLOCATOR>(
	       main_trie, df_decoder, s_alloc),
	    nr_of_all_candidates(0){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      	 
	 void afterWorkCandGen();
      protected:

	 void generateCandidateAtParent( TRIE* trie );
	 /// Generates candidate of size more than two
	 void generateCandidateFindParent( 
	    TRIE* trie, unsigned int step_to_freq_par);

      public:
	 // No public members
   };


   template <class DF_D, class TRIE, class LEAF, class LEAF_ALLOCATOR> 
   void CandidateGeneratorNoprune<DF_D, TRIE, LEAF, LEAF_ALLOCATOR>::
   generateCandidate(unsigned int candidate_size)
   {
      assert(candidate_size > 1);
      nr_of_new_candidates = 0;
      generateCandidateFindParent( &this->main_trie,
				   candidate_size -2 );
      log_dbg(0, "Number of newly generated candidates: %d", nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_DBG
      nr_of_all_candidates += nr_of_new_candidates;
#endif
   }
   template <class DF_D, class TRIE, class LEAF, class LEAF_ALLOCATOR> 
   void CandidateGeneratorNoprune<DF_D, TRIE, LEAF, LEAF_ALLOCATOR>::
   afterWorkCandGen()
   {
      log_dbg(0, "The total number of generated candidates: %d", nr_of_all_candidates);
   }


   template <class DF_D, class TRIE, class LEAF, class LEAF_ALLOCATOR> inline void 
   CandidateGeneratorNoprune<DF_D, TRIE, LEAF, LEAF_ALLOCATOR>::
   generateCandidateAtParent(TRIE* trie)
   {
      typename TRIE::iterator itEdge2;
      LEAF* toExtendAsLeaf;
      TRIE* toExtend;
      replace_list.clear();
      for( typename TRIE::iterator itEdge( trie->edgelist.begin() );
	   itEdge != trie->edgelist.end(); ++itEdge)
      {
	 toExtendAsLeaf = static_cast<LEAF*>((*itEdge).second);
	 PARENT::df_decoder.pushItemWithWrite( (*itEdge).first, 
				       toExtendAsLeaf->getCounter() );
	 PARENT::df_decoder.popItem();
	 extenders.clear();
	 for( itEdge2 = trie->edgelist.begin(); 
	      itEdge2 != trie->edgelist.end(); ++itEdge2 )
	 {
	    extenders.push_back(Edge((*itEdge2).first, PARENT::s_alloc.allocate()));
	    static_cast<LEAF*>(extenders.back().second)->setCounter(0);
	 }
	 if(!extenders.empty())
	 {
	    toExtend = new TRIE(toExtendAsLeaf->getCounter());
	    toExtend->edgelist.insert(extenders);
	    replace_list.push_back(Edge((*itEdge).first, toExtend));
	 }
	 PARENT::s_alloc.free(toExtendAsLeaf);
      }
      trie->edgelist.clear();
      trie->edgelist.insert(replace_list);
   }


   template <class DF_D, class TRIE, class LEAF, class LEAF_ALLOCATOR> void
   CandidateGeneratorNoprune<DF_D, TRIE, LEAF, LEAF_ALLOCATOR>::
   generateCandidateFindParent( TRIE* trie, 
				unsigned int step_to_freq_par)
   {
      typename TRIE::iterator itEdge( trie->edgelist.begin() );
      if(step_to_freq_par)
      {
	 --step_to_freq_par;
	 while( itEdge != trie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem( (*itEdge).first );
	    generateCandidateFindParent( 
	       static_cast<TRIE*>((*itEdge).second), step_to_freq_par);
	    if(static_cast<TRIE*>((*itEdge).second)->edgelist.empty())
	    {
	       delete static_cast<TRIE*>((*itEdge).second);
	       itEdge = trie->edgelist.erase(itEdge);
	    }
	    else ++itEdge;
	    PARENT::df_decoder.popItem();
	 }
      }
      else
      {
	 generateCandidateAtParent(trie);
#if DEBUG_LEVEL >= LEVEL_DBG
	 for( typename TRIE::iterator it( trie->edgelist.begin() );
	      it != trie->edgelist.end(); ++it)
	    nr_of_new_candidates += 
	       static_cast<TRIE*>((*it).second)->edgelist.size();
#endif
      }
   }
}
}
}
#endif
