/***************************************************************************
                          FrequentItemInserter.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef FrequentItemInserter_HPP
#define FrequentItemInserter_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test

namespace Bodon
{

template <class DF_D, class TRIE, bool NEE>
class FrequentItemInserter : public ManipulatorBase<DF_D, TRIE>
{
   private:
      typedef ManipulatorBase<DF_D, TRIE> PARENT;
   public:
      FrequentItemInserter( TRIE& trie, DF_D& df_decoder ) : 
	 PARENT(trie, df_decoder){}

      /// Sets the support of the emptyset.
      void setEmptysetSupport(const counter_t emptyset_support)
      {PARENT::main_trie.setCounter(emptyset_support);}

      /// Insert the frequent items and their counters into the data structure;
      void insertFrequentItems(const std::vector<counter_t>& freq_counters )
      {
	 std::vector<Edge> extenders;
	 if(NEE)
	 {
	    for(item_t index = 0; index < freq_counters.size(); ++index)
	    {
	       if( freq_counters[index] == PARENT::main_trie.getCounter())
		  PARENT::main_trie.neelist.push_back(index);
	       else
		  extenders.push_back(Edge(index, 
					   new TRIE(freq_counters[index])));
	    }
	 }
	 else
	 {
	    PARENT::df_decoder.write(PARENT::main_trie.getCounter());
	    for(item_t index = 0; index < freq_counters.size(); ++index)
	    {
	       PARENT::df_decoder.pushItemWithWrite( index, freq_counters[index] );
	       extenders.push_back(Edge(index, 
					new TRIE(freq_counters[index])));
	       PARENT::df_decoder.popItem();
	    }
	 }
	 PARENT::main_trie.edgelist.insert(extenders);
      }
};

}



#endif
