/***************************************************************************
                          InfreqRemover.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef InfreqRemover_HPP
#define InfreqRemover_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/ManipulatorBase.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace trie
{
   template <class DF_D, class TRIE, class TRIE_ALLOCATOR, NEELevel NEE>
   class InfreqRemover : public Bodon::inhomogeneous_trie::
	 ManipulatorBase<DF_D, TRIE, TRIE_ALLOCATOR>
   {
      private:
	 typedef Bodon::inhomogeneous_trie::ManipulatorBase<DF_D, TRIE, TRIE_ALLOCATOR> PARENT;
      protected:
	 std::vector<Edge> extenders;
	 unsigned int nr_of_deleted;
	 unsigned int total_nr_of_deleted;

      public:
	 InfreqRemover( TRIE& main_trie, DF_D& df_decoder,
			TRIE_ALLOCATOR& s_alloc ) : 
	    Bodon::inhomogeneous_trie::ManipulatorBase<DF_D, TRIE, TRIE_ALLOCATOR>(
	       main_trie, df_decoder, s_alloc), 
	    total_nr_of_deleted(0){}

	 /// Deletes infrequent itemsets.
	 void deleteInfrequent( const counter_t min_supp,
				unsigned int candidate_size );
	 void afterWorkDel();
      protected:

	 /// Deletes the tries that represent infrequent itemsets.
	 void delete_infrequent_subtrie( 
	    TRIE* subtrie, const counter_t min_supp,
	    unsigned int step_to_cand_par );
	 void delete_infrequent_subtrieNEE( 
	    TRIE* subtrie, const counter_t min_supp,
	    std::vector<item_t>& NEEsum, unsigned int step_to_cand_par);
   };

/**
   \param min_supp The threshold of absolute support.
   \param candidate_size the size of the candidates
*/
   template <class DF_D, class TRIE, class TRIE_ALLOCATOR, NEELevel  NEE>
   void InfreqRemover<DF_D, TRIE, TRIE_ALLOCATOR, NEE>::
   deleteInfrequent( const counter_t min_supp,
		     unsigned int candidate_size )
   {
      assert(candidate_size > 2);
      nr_of_deleted = 0;
      if(NEE > NEE_Off)
      {
	 std::vector<item_t> NEEsum;
	 delete_infrequent_subtrieNEE( &this->main_trie, min_supp, 
				       NEEsum, candidate_size - 1 );
	 
      }
      else
	 delete_infrequent_subtrie( &this->main_trie, min_supp, 
				    candidate_size - 1 );
      log_dbg(0,"Number of infrequent candidates: %d", nr_of_deleted);
#if DEBUG_LEVEL >= LEVEL_DBG
      total_nr_of_deleted += nr_of_deleted;
#endif
   }

   template <class DF_D, class TRIE, class TRIE_ALLOCATOR, NEELevel NEE>
   void InfreqRemover<DF_D, TRIE, TRIE_ALLOCATOR, NEE>::
   afterWorkDel()
   {
      log_dbg(0,"The total number of infrequent candidates: %d", 
	      total_nr_of_deleted);
   }

/**
   \param subtrie the root of the current subtrie
   \param min_supp The support threshold
   \param step_to_cand_par the number of downward steps to reach 
   the depth where the parents of the candidates are.
*/
   template <class DF_D, class TRIE, class TRIE_ALLOCATOR, NEELevel NEE>
   void InfreqRemover<DF_D, TRIE, TRIE_ALLOCATOR, NEE>::
   delete_infrequent_subtrie( TRIE* subtrie, const counter_t min_supp,
			      unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    delete_infrequent_subtrie( static_cast<TRIE*>((*itEdge).second), min_supp,
				       step_to_cand_par);
	    if( static_cast<TRIE*>((*itEdge).second)->edgelist.empty() )
	    {
	       PARENT::s_alloc.free(static_cast<TRIE*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else ++itEdge;
	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<TRIE*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_DBG
	       ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<TRIE*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else
	       ++itEdge;
	 }
      }
   }
   template <class DF_D, class TRIE, class TRIE_ALLOCATOR, NEELevel NEE>
   void InfreqRemover<DF_D, TRIE, TRIE_ALLOCATOR, NEE>::
   delete_infrequent_subtrieNEE( TRIE* subtrie, const counter_t min_supp,
				 std::vector<item_t>& NEEsum,
				 unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 NEEsum.insert( NEEsum.end(), 
	       subtrie->neelist.begin(), subtrie->neelist.end() );
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    PARENT::df_decoder.pushItem((*itEdge).first);
	    delete_infrequent_subtrieNEE( static_cast<TRIE*>((*itEdge).second), min_supp,
					  NEEsum, step_to_cand_par);
	    if( static_cast<TRIE*>((*itEdge).second)->edgelist.empty() )
	    {
	       NEEsum.insert( NEEsum.end(), 
			      static_cast<TRIE*>((*itEdge).second)->neelist.begin(), 
			      static_cast<TRIE*>((*itEdge).second)->neelist.end() );
	       if( NEEsum.size() < step_to_cand_par + 1)
	       {

		  PARENT::df_decoder.writeNEE(
		     static_cast<TRIE*>((*itEdge).second)->getCounter(), NEEsum);
		  NEEsum.resize( NEEsum.size() - 
				 static_cast<TRIE*>((*itEdge).second)->neelist.size() );
		  PARENT::s_alloc.free(static_cast<TRIE*>((*itEdge).second));
		  itEdge = subtrie->edgelist.erase(itEdge);
	       }
	       else
	       {
		  NEEsum.resize( NEEsum.size() - 
				 static_cast<TRIE*>((*itEdge).second)->neelist.size() );
		  ++itEdge;
	       }	   
	    }
	    else ++itEdge;
	    PARENT::df_decoder.popItem();
	 }
	 NEEsum.resize( NEEsum.size() - subtrie->neelist.size() );
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<TRIE*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_DBG
	       ++nr_of_deleted;
#endif
	       PARENT::s_alloc.free(static_cast<TRIE*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else if(static_cast<TRIE*>((*itEdge).second)->getCounter() == subtrie->getCounter())
	    {
	       subtrie->neelist.push_back((*itEdge).first);
	       PARENT::s_alloc.free(static_cast<TRIE*>((*itEdge).second));
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else
	       ++itEdge;
	 }
      }
   }
}
}
#endif
