/***************************************************************************
                          CandGenInfreqRemoveNopruneMerge.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef CandGenInfreqRemoveNopruneMerge_HPP
#define CandGenInfreqRemoveNopruneMerge_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include "common/log.h"
#include "apriori/bodon/trie/trie_manipulators/sequence/CandidateGeneratorNoprune.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace sequence
{
   template <class DF_D, class TRIE>
   class CandGenInfreqRemoveNopruneMerge : 
	 public CandidateGeneratorNoprune<DF_D, TRIE>
   {
      protected:
	 unsigned int nr_of_deleted;
	 unsigned int total_nr_of_deleted;
      public:
	 CandGenInfreqRemoveNopruneMerge( TRIE& main_trie, DF_D& df_decoder ) : 
	    CandidateGeneratorNoprune<DF_D, TRIE>(main_trie, df_decoder), 
	    total_nr_of_deleted(0){}


	 /** Generates candidates 
	  
	 Generates candidates and writes frequent itemset 
	 that are obtained in the previous iteration to disk.
	 */
	 void generateCandidate(unsigned int candidate_size);
      
	 void deleteInfrequent( const counter_t min_supp, 
				unsigned int candidate_size );
	 void afterWorkDel();
      protected:

	 /// Deletes the tries that represent infrequent itemstd::sets.
	 void delete_infrequent_subtrie( TRIE* subtrie, 
					 const counter_t min_supp,
					 unsigned int step_to_cand_par );

      public:
	 // No public members
   };


   template <class DF_D, class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE>::
   generateCandidate(const unsigned int candidate_size)
   {
      if(candidate_size == 3)
      {
	 nr_of_new_candidates = 0;
/*	 if(NEE)
	 {
	    std::vector<item_t> NEEsum;
	    generateCandidateFindParentNEE( &main_trie, NEEsum,
					    candidate_size -2 );
					    }
					    else*/
	    generateCandidateFindParent( &main_trie, 
					 candidate_size -2 );
	 log_dbg(0,"Number of newly generated candidates: %d", nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_DBG
	    nr_of_all_candidates += nr_of_new_candidates;
#endif
      }
      else
	 log_dbg(0,"This algorithm does not generate any candidate in this phase.");

   }

/**
   \param min_supp The threshold of absolute support.
   \param candidate_size The size of the candidate itemstd::set.
*/
   template <class DF_D, class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE>::
   deleteInfrequent( const counter_t min_supp, unsigned int candidate_size )
   {
      assert(candidate_size > 1);
      nr_of_deleted = 0;
      nr_of_new_candidates = 0;
      delete_infrequent_subtrie( &main_trie, min_supp, 
				 candidate_size - 1 );
      log_dbg(0,"Number of infrequent candidates: %d", nr_of_deleted);
      log_dbg(0,"Number of newly generated candidates (in del. phase!): %d", 
	      nr_of_new_candidates);
#if DEBUG_LEVEL >= LEVEL_DBG
      nr_of_all_candidates += nr_of_new_candidates;
      total_nr_of_deleted += nr_of_deleted;
#endif
   }

   template <class DF_D, class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE>::
   afterWorkDel()
   {
      log_dbg(0,"The total number of generated candidates (in del. p.): %d", 
	      nr_of_all_candidates);
      log_dbg(0,"The total number of infrequent candidates: %d", 
	      total_nr_of_deleted);
   }

/**
   \param min_supp The support threshold
*/
   template <class DF_D, class TRIE>
   void CandGenInfreqRemoveNopruneMerge<DF_D, TRIE>::
   delete_infrequent_subtrie( TRIE* subtrie, const counter_t min_supp,
			      unsigned int step_to_cand_par)
   {
      typename TRIE::iterator itEdge( subtrie->edgelist.begin() );
      if(step_to_cand_par)
      {
	 --step_to_cand_par;
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    df_decoder.pushItem( (*itEdge).first );
	    delete_infrequent_subtrie( static_cast<TRIE*>((*itEdge).second), min_supp,
				       step_to_cand_par);
	    df_decoder.popItem();
	    if( static_cast<TRIE*>((*itEdge).second)->edgelist.empty() )
	    {
	       delete static_cast<TRIE*>((*itEdge).second);
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else ++itEdge;
	 }
      }
      else
      {
	 while( itEdge != subtrie->edgelist.end() )
	 {
	    if( static_cast<TRIE*>((*itEdge).second)->getCounter() < min_supp )
	    {
#if DEBUG_LEVEL >= LEVEL_DBG
		  ++nr_of_deleted;
#endif
	       delete static_cast<TRIE*>((*itEdge).second);
	       itEdge = subtrie->edgelist.erase(itEdge);
	    }
	    else
	       ++itEdge;
	 }
	 generateCandidateAtParent(subtrie);
#if DEBUG_LEVEL >= LEVEL_DBG
	 for( typename TRIE::iterator it( subtrie->edgelist.begin() );
	      it != subtrie->edgelist.end(); ++it)
	    nr_of_new_candidates += 
	       static_cast<TRIE*>((*it).second)->edgelist.size();
#endif
      }
   }
}
}
#endif
