/***************************************************************************
                          SupportCounterLookupSeq.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounterLookupSeq_HPP
#define SupportCounterLookupSeq_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace sequence
{
   template <class TRIE>
   class SupportCounterLookupSeq
   {
      public:
	 SupportCounterLookupSeq( ){}
	 void updateCounters(
	    TRIE& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }

      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_transaction_upper_bound,
	    std::vector<item_t>::const_iterator it_transaction, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE> void 
   SupportCounterLookupSeq<TRIE>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_transaction_upper_bound,
      std::vector<item_t>::const_iterator it_transaction, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      std::vector<item_t>::const_iterator it_transaction2;
      --step_to_candidate;
      for( typename TRIE::iterator it_edge(subtrie->edgelist.begin()); 
	   it_edge != subtrie->edgelist.end(); ++it_edge)
      {
	 for( it_transaction2 = it_transaction; 
	      it_transaction2 != it_transaction_upper_bound;
	      ++it_transaction2 )
	    if(*it_transaction2 == (*it_edge).first )
	    {
	       if( step_to_candidate ) 
		  findCandidates( static_cast<TRIE*>((*it_edge).second),
				  it_transaction_upper_bound + 1, it_transaction2+1, 
				  step_to_candidate, counter_incr);
	       else
		  static_cast<Leaf*>((*it_edge).second)->increaseCounter( counter_incr);
	       break;
	    }
      }
   }
}
}
#endif
