/***************************************************************************
                          SupportCounterHybridSimple.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounterHybridSimple_HPP
#define SupportCounterHybridSimple_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "apriori/bodon/Leaf.hpp"
#include <vector>
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
   template <class TRIE, unsigned int THRESHOLD>
   class SupportCounterHybridSimple
   {
      public:
	 SupportCounterHybridSimple( ){}

      public:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE, unsigned int THRESHOLD> void 
   SupportCounterHybridSimple<TRIE, THRESHOLD>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      if(subtrie->edgeNumber() < THRESHOLD)
      {
	 typename TRIE::iterator it_edge(subtrie->edgelist.begin());
	 while( it_edge != subtrie->edgelist.end() && 
		it_basket != it_basket_upper_bound )
	 {
	    if( (*it_edge).first < *it_basket) ++it_edge;
	    else if( (*it_edge).first > *it_basket) ++it_basket;
	    else
	    {
	       ++it_basket;
	       if( step_to_candidate ) 
		  findCandidates( static_cast<TRIE*>((*it_edge).second),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr);
	       else
		  static_cast<Leaf*>((*it_edge).second)
		     ->increaseCounter( counter_incr);
	       ++it_edge;
	    }
	 }
      }
      else
      {
	 void* subsubtrie;
	 item_t largest_label = subtrie->edgelist.largestEdgelabel();
	 while( it_basket != it_basket_upper_bound )
	 {
	    if( *it_basket > largest_label)
	       break;
	    else
	    {
	       subtrie->edgelist.lookupNoUppercheck(*it_basket, subsubtrie);
	       ++it_basket;
	       if(subsubtrie)
		  if( step_to_candidate ) 
		     findCandidates( static_cast<TRIE*>(subsubtrie),
				     it_basket_upper_bound + 1, it_basket, 
				     step_to_candidate, counter_incr );
		  else
		     static_cast<Leaf*>(subsubtrie)
			->increaseCounter( counter_incr);
	    }
	 }
      }
   }
}
#endif
