/***************************************************************************
                          SupportCounterLookupEdge.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounterLookupEdge_HPP
#define SupportCounterLookupEdge_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <iterator>	//for test




namespace Bodon
{
   template <class TRIE>
   class SupportCounterLookupEdge
   {
      protected:
	 typedef Leaf LEAF;
      public:
	 SupportCounterLookupEdge( ){}

	 void updateCounters(
	    TRIE& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }

      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE> void 
   SupportCounterLookupEdge<TRIE>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      void* subsubtrie;
      if( step_to_candidate ) 
      {
	 while( it_basket != it_basket_upper_bound )
	 {
	    if( subtrie->edgelist.lookup(*it_basket, subsubtrie) )
	    {
	       ++it_basket;
	       if(subsubtrie)
		  findCandidates( static_cast<TRIE*>(subsubtrie),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr );
	    }
	    else
	       break;
	 }
      }
      else
      {
	 while( it_basket != it_basket_upper_bound )
	 {
	    if( subtrie->edgelist.lookup(*it_basket, subsubtrie) )
	    {
	       ++it_basket;
	       if(subsubtrie)
		  static_cast<LEAF*>(subsubtrie)
		     ->increaseCounter( counter_incr);
	    }
	    else
	       break;
	 }
      }
   }
}
#endif
