/***************************************************************************
                          SupportCounterLookupTrans.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounterLookupTrans_HPP
#define SupportCounterLookupTrans_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <iterator>	//for test




namespace Bodon
{
   template <class TRIE>
   class SupportCounterLookupTrans
   {
      private:
	 typedef Leaf LEAF;
      public:
	 SupportCounterLookupTrans( ){}
	 
	 void updateCounters(
	    TRIE& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }

      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE> void 
   SupportCounterLookupTrans<TRIE>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      if( step_to_candidate ) 
      {
	 for( typename TRIE::iterator it_edge(subtrie->edgelist.begin());
	      it_edge != subtrie->edgelist.end(); ++it_edge)
	 {
	    it_basket = std::lower_bound( it_basket, it_basket_upper_bound, 
					  (*it_edge).first);
	    if(it_basket == it_basket_upper_bound)
	       break;
	    else
	    {
	       if(*it_basket == (*it_edge).first)
	       {
		  findCandidates( static_cast<TRIE*>((*it_edge).second),
				  it_basket_upper_bound + 1, it_basket, 
				  step_to_candidate, counter_incr);
		  ++it_basket;
	       }
	    }
	 }
      }
      else
      {
	 for( typename TRIE::iterator it_edge(subtrie->edgelist.begin());
	      it_edge != subtrie->edgelist.end(); ++it_edge)
	 {
	    it_basket = std::lower_bound( it_basket, it_basket_upper_bound, 
					  (*it_edge).first);
	    if(it_basket == it_basket_upper_bound)
	       break;
	    else
	    {
	       if(*it_basket == (*it_edge).first)
	       {
		  static_cast<LEAF*>((*it_edge).second)
		     ->increaseCounter( counter_incr);
		  ++it_basket;
	       }
	    }
	 }
      }
   }
}
#endif
