/***************************************************************************
                          SupportCounterMerge.hpp  -  description
                             -------------------
    begin                : cs dec 26 2004
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef SupportCounterMerge_HPP
#define SupportCounterMerge_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
   template <class TRIE>
   class SupportCounterMerge
   {
      protected:
	 typedef Leaf LEAF;
      public:
	 SupportCounterMerge( ){}

	 void updateCounters(
	    TRIE& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }

      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_basket_upper_bound,
	    std::vector<item_t>::const_iterator it_basket, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE> void 
   SupportCounterMerge<TRIE>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_basket_upper_bound,
      std::vector<item_t>::const_iterator it_basket, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      --step_to_candidate;
      typename TRIE::iterator it_edge(subtrie->edgelist.begin());
      if( step_to_candidate ) 
      {
	 while( it_edge != subtrie->edgelist.end() && 
		it_basket != it_basket_upper_bound )
	 {
	    if( (*it_edge).first < *it_basket) ++it_edge;
	    else if( (*it_edge).first > *it_basket) ++it_basket;
	    else
	    {
	       ++it_basket;
	       findCandidates( static_cast<TRIE*>((*it_edge).second),
			       it_basket_upper_bound + 1, it_basket, 
			       step_to_candidate, counter_incr);
	       ++it_edge;
	    }
	 }
      }
      else
      {
	 while( it_edge != subtrie->edgelist.end() && 
		it_basket != it_basket_upper_bound )
	 {
	    if( (*it_edge).first < *it_basket) ++it_edge;
	    else if( (*it_edge).first > *it_basket) ++it_basket;
	    else
	    {
	       ++it_basket;
	       static_cast<LEAF*>((*it_edge).second)
		     ->increaseCounter( counter_incr);
	       ++it_edge;
	    }
	 }
      }
   }
}
#endif
