#ifndef common_HPP
#define common_HPP

//! set this to empty if you compiler does not support it
#ifdef TRACE
#define DINLINE
#else
#define DINLINE __attribute__((always_inline))
#endif //TRACE

#define NOINLINE __attribute__((noinline))

//! set this to empty if your compiler does not support (and does not need) typenames
#define TYPENAME typename

#include <vector>
#include <algorithm>
#include <functional>
#include <limits>
#include <inttypes.h>

/**
  *@author Ferenc Bodon
  */

  /** The type of the item.
    *
    * Items are represented by non-negative integers. 
    * If we know that no itemcode is larger than \f$2^{16}\f$, we can recomplie
    * the code using unsigned short as type.
    */
typedef unsigned int item_t;
//typedef unsigned short item_t;
//typedef long item_t;

const unsigned int largest_itemsetsize = 
std::numeric_limits<unsigned int>::max(); // it is more than enough

class itemvector : public std::vector<item_t>
{
   public:
      bool haveSamePrefix(const itemvector& other) const
      {
	 return std::equal(
	    std::vector<item_t>::begin(), std::vector<item_t>::end()-1,
	    other.begin());
      }
      bool isContained(std::vector<item_t>::const_iterator transaction_beg,
		       std::vector<item_t>::const_iterator transaction_end) const 
      {
	 return std::includes( 
	    transaction_beg, transaction_end,
	    std::vector<item_t>::begin(), std::vector<item_t>::end() );
      }
      bool isContained(const std::vector<item_t>& transaction) const 
      {
	 return isContained(transaction.begin(), transaction.end());
      }
};

  /** The type of the counter for support counting.
    *
    * Counters are represented by non-negative integers. 
    * If we know that the transaction file does not contain more than n \f$2^{16}\f$ transactions, we can recomplie
    * the code using unsigned short as type.
    */
//typedef unsigned int counter_t;
typedef uint32_t counter_t;

typedef std::pair<itemvector, counter_t> itemvecCounterPair;

class itemvecCounterPair_less : public 
std::binary_function<itemvecCounterPair, itemvecCounterPair, bool>
{
   public:
      bool operator()( const itemvecCounterPair& it_vec_pair,
		       const itemvecCounterPair& it_vec ) const
      {
	 return it_vec_pair.first < it_vec.first;
      }
      
};

const counter_t TWO_POW31   = 0x80000000;
const counter_t TWO_POW31_1 = TWO_POW31 - 1;

//!values of FpGrowth feature NEE
enum NEELevel {
  NEE_Off,
  NEE_Level1,
  NEE_Full
};

//!values of FpGrowth feature FirstLevel
enum FirstLevel {
  //build a single tree and do normal projections on the first level
  FLBuildSingleTree,
  //create buildtrees for each item with L1 conditional content: insert each transaction into all the L1 trees (of the items appearing in the transaction)
  FLBuildAllL1Trees,
  //Create a single build tree and do simultaneous projection
  FLSimultProject
}; //enum FPGrowth_FirstLevel


#endif
