/***************************************************************************
                          Edgelist.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef Edgelist_HPP
#define Edgelist_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include <vector>
//#include <iterator>	//for test


namespace Bodon
{
   /** This struct represent an edge of a Edgelist.
       
   An edge has a label, and an edge points to a subtrie.
   */

   template <class VECTOR = std::vector<Edge> >
   class Edgelist : public VECTOR
   {
      public:
	 Edgelist() : VECTOR(){}

	 void* find(item_t item) const;
	 void*& findOrCreate(item_t label);
	 bool lookup(item_t label, void*& subtrie) const;
	 void insert(const std::vector<Edge>& new_edges)
	 {
//option I.	    VECTOR::swap(new_edges);
///option II.
	    VECTOR::reserve(new_edges.size());
	    for(std::vector<Edge>::const_iterator it = new_edges.begin();
	       it != new_edges.end(); ++it)
	       VECTOR::push_back(*it);
//option III.
//	    VECTOR::insert(VECTOR::end(), new_edges.begin(), new_edges.end());
//option IV.
//	    VECTOR::operator=(new_edges);
	 }
	 size_t edgeNumber() const
	 {
	    return VECTOR::size();
	 }
   };
   
   template <class VECTOR>
   inline void* Edgelist<VECTOR>::find(
      item_t label) const
   {
      typename VECTOR::const_iterator it = 
	 std::find( VECTOR::begin(), VECTOR::end(), label );
      if( it != VECTOR::end() )
	 return  (*it).second;
      else 
	 return NULL;
   }

   template <class VECTOR>
   inline void*& Edgelist<VECTOR>::findOrCreate(
      item_t label)
   {
      typename VECTOR::iterator it = 
	 std::find( VECTOR::begin(), VECTOR::end(), label );
      if( it == VECTOR::end() )
      {
	 VECTOR::push_back(Edge(label, NULL));
	 it = --VECTOR::end();
      }
      return (*it).second;
   }

   template <class VECTOR>
   inline bool Edgelist<VECTOR>::
   lookup(item_t label, void*& subtrie) const
   {
      bool return_value = false;
      for(typename VECTOR::const_iterator it = VECTOR::begin(); 
	  it != VECTOR::end(); ++it)
      {
	 if(*it == label)
	 {
	    subtrie = (*it).second;
	    return true;
	 }
	 else if(*it > label)
	    return_value = true;
      }
      return return_value;
   }
}

#endif
