/***************************************************************************
                          MapRepr.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@cs.bme.hu
 ***************************************************************************/

#ifndef MapRepr_HPP
#define MapRepr_HPP

/**
  *@author Ferenc Bodon
  */


#include "common.hpp"
#include "common/Edge.hpp"
#include <vector>
#include <iterator>	//for test


namespace Bodon
{
   /** This struct represent an edge of a MapRepr.
       
   An edge has a label, and an edge points to a subtrie.
   */
	 
   class MapRepr : public std::map<item_t,void*>
   {
      public:

	 MapRepr():std::map<item_t,void*>(){}

	 std::map<item_t,void*>::iterator erase(
	    std::map<item_t,void*>::iterator pos)
	 {
	    std::map<item_t,void*>::iterator return_it(pos);
	    ++return_it;
	    std::map<item_t,void*>::erase(pos);
	    return return_it;
	 }
	 void* find(item_t label) const
	 {
	    std::map<item_t,void*>::const_iterator it(std::map<item_t,void*>::find(label)); 
	    if(it == end() )
	       return NULL;
	    else 
	       return (*it).second;
  
	 }

	 void*& findOrCreate(item_t label)
	 {
	    return operator[](label);
	 }
	 bool lookup(item_t label, void*& subtrie) const;
	 void insert(const std::vector<Edge>& new_edges);

	 static bool isLookupPreferred()
	 {
	    return true;
	 }
   };
   inline bool MapRepr::lookup(item_t label, void*& subtrie) const
   {
      if( (*--end()).first < label)
	 return false;
      else
      {
	 std::map<item_t,void*>::const_iterator it(std::map<item_t,void*>::find(label)); 
	 if( it == end() )
	    subtrie = NULL;
	 else 
	    subtrie = (*it).second;
	 return true;
      }
   }

   inline void MapRepr::insert(const std::vector<Edge>& new_edges)
   {
      for(std::vector<Edge>::const_iterator it = new_edges.begin();
	  it != new_edges.end(); ++it)
	 std::map<item_t,void*>::insert( end(), std::map<item_t,void*>::
					 value_type((*it).first, (*it).second));
   }
}

#endif
