/**
 * SparseBitmatrix_create.hpp
 *   -- create a sparse item transaction incidence bitmatrix from transaction data.
 */

#ifndef __SPARSEBITMATRIX_CREATE_HPP_
#define __SPARSEBITMATRIX_CREATE_HPP_ t

// #include "SparseBitmatrix.hpp"
//#include "eclat/SparseBitmatrix_singleblock.hpp"
#include "common/debug.hpp"

// #include "io/input/transaction_reader/TransactionReader.hpp"
// #include "common.hpp"

#include <vector>

typedef std::vector<std::pair<counter_t, item_t> > CounterItemPairs;

/**
 * Create a sparse item transaction incidence bitmatrix from transaction data.
 * @param tdb transaction reader for transaction database that contains only frequent items.
 * @param freqItems pairs of items and occurrence counters (for frequent items).
 * @return a sparse item transaction incidence bitmatrix for all frequent items in the transaction database.
 */ 
template <class TRANSACTION_READER>
SparseBitmatrix* createSparseItemTransactionIncidenceMatrix(TRANSACTION_READER& tdb, CounterItemPairs& freqItems) {
  SparseBitmatrix* iti = NULL;

  // 0. get frequencies of items in a simple array:
  int numItems = freqItems.size();
  int* itemFrequencies = new int[numItems];
  CounterItemPairs::iterator it_freqItems;
  int i = 0;
  for( it_freqItems = freqItems.begin(); it_freqItems != freqItems.end(); ++it_freqItems )
    itemFrequencies[i++] = it_freqItems->first; 
  // [lst 2005/05/04] this is correct as the index is the recoded item ID.
  // It should *NOT* be itemFrequencies[it_freqItems->second] = it_freqItems->first; 

  // 1. create matrix:
  iti = new SparseBitmatrix(numItems, itemFrequencies);
  // [lst 2005/05/04] We do not use custom row labels via 
  //   iti = new SparseBitmatrix(numItems, tdb->freqItems, tdb->encoding->itemByCode);
  // as the recoding is handeled by the I/O framework.  
  /*
  int** rows = iti->rows;
  */

  // 3. fill matrix:
  tdb.rewind();
  std::vector<item_t> transaction;
  std::vector<item_t>::iterator it_transaction;
  i = 0;
  while(tdb.nextTransactionBIS(transaction))
  {      
    for(it_transaction = transaction.begin(); it_transaction != transaction.end(); ++it_transaction) {
      int item = *it_transaction;
      iti->push_back(item, i);
    }
    i++;
  }

  delete[] itemFrequencies;

  return iti;
}


#endif
