#ifndef FPGrowthSelector_HPP
#define FPGrowthSelector_HPP

/**
 *@author Balazs RACZ
 */

//#include "io/output/NonOptDFOutput.hpp"

#include "common.hpp"

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/output/NoOutput.hpp"
#include "io/output/StatOutput.hpp"

#include "fpgrowth/core.cpp"
#include "fpgrowth/nonordfp.cpp"
#include "fpgrowth/classicfp.cpp"
#include "fpgrowth/classicrebuildfp.cpp"
#include "fpgrowth/buildfptree.hpp"

#include "common/allocators.hpp"
#include "common/linmap.cpp"


/*enum FpGrowth_Type {
  fpt_nonordfp_basic,
  fpt_nonordfp_dense,
  fpt_nonordfp_sparse,
  fpt_classic_td,
  fpt_nonordfp_basic_nonee,
  fpt_nonordfp_dense_nonee,
  fpt_nonordfp_sparse_nonee,
  fpt_classic_td_nonee,
  fpt_nonordfp_basic_nosingle,
  fpt_nonordfp_dense_nosingle,
  fpt_nonordfp_sparse_nosingle,
  fpt_classic_td_nosingle,
  fpt_classic_rebuild
  };*/


template<NEELevel NEE> struct NEE_to_String {
  static char* getString() {
    return "unknown";
  }
};

template<> struct NEE_to_String<NEE_Off> {
  static char* getString() {
    return "Off";
  }
};
template<> struct NEE_to_String<NEE_Level1>{
  static char* getString() {
    return "Level1";
  }
};
template<> struct NEE_to_String<NEE_Full>{
  static char* getString() {
    return "Full";
  }
};


template<FirstLevel First> struct FirstLevel_to_String {
  static char* getString() {
    return "unknown";
  }
};

template<> struct FirstLevel_to_String<FLBuildSingleTree> {
  static char* getString() {
    return "FL_BuildSingleTree";
  }
};
template<> struct FirstLevel_to_String<FLBuildAllL1Trees>{
  static char* getString() {
    return "FL_BuildAllL1Trees";
  }
};
template<> struct FirstLevel_to_String<FLSimultProject>{
  static char* getString() {
    return "FL_SimultProject";
  }
};


  struct type_nonordfp_basic {};
  struct type_nonordfp_dense {};
  struct type_nonordfp_sparse {};
  struct type_nonordfp_nsparse {};
  struct type_nonordfp_classictd {};
  struct type_nonordfp_classicrebuild {};
  struct type_nonordfp_classictd_alll1trees {};
  struct type_nonordfp_classicrebuild_alll1trees {};
  

template<class S_C, class O_M, class B_T, bool SINGLE, NEELevel NEE, bool TD, bool PROJECT, bool PROJECTMERGENODES, bool PROJECTDELETECLOSED, FirstLevel PERITEMTREE, bool SPARSEAGGR> void doFpGrowth(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m) {
  log_perf(0,"Fp-growth configuration: +nonordfp (compact), %ssinglepath, %std, %sproject, %sProjMergeNodes, %sProjDeleteClosed NEE_%s, %s, BT_EndPatricia",SINGLE?"+":"-",TD?"+":"-",PROJECT?"+":"-",PROJECTMERGENODES?"+":"-",PROJECTDELETECLOSED?"+":"-",NEE_to_String<NEE>::getString(),FirstLevel_to_String<PERITEMTREE>::getString());

  typedef bracz::TDNonOrdFPStructure<S_C, B_T ,/*singlepath*/SINGLE,/*TD*/TD,
    /*project*/PROJECT,/*projectdeleteclosed*/PROJECTDELETECLOSED,/*projmergenodes*/PROJECTMERGENODES,
    /*per item tree*/PERITEMTREE,/*SPARSEAGGR*/SPARSEAGGR> NonOrdFpStructure;

  NonOrdFpStructure fpstr(sorted_coder,maxitem,min_supp);

  bracz::FPGrowthBaseNoNEE<NonOrdFpStructure,O_M,/*singlepath*/SINGLE,/*NEE*/NEE,/*per item tree*/PERITEMTREE> fpgr;

  fpgr.findFrequentItems(min_supp,fpstr,o_m,maxitem);
}


template<class S_C, class O_M, class B_T, bool SINGLE, NEELevel NEE, bool TD, FirstLevel PERITEMTREE> void doFpGrowthClassic(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m) {
  log_perf(0,"Fp-growth configuration: +classic (node-based), %ssinglepath, %std, -project, -ProjMergeNodes, -ProjDeleteClosed NEE_%s, %s, BT_EndPatricia",SINGLE?"+":"-",TD?"+":"-",NEE_to_String<NEE>::getString(),FirstLevel_to_String<PERITEMTREE>::getString());

  typedef bracz::ClassicFPStructure<S_C, B_T ,/*per item tree*/PERITEMTREE,/*TD*/TD,/*singlepath*/SINGLE
    > NonOrdFpStructure;

  NonOrdFpStructure fpstr(sorted_coder,maxitem);

  bracz::FPGrowthBaseNoNEE<NonOrdFpStructure,O_M,/*singlepath*/SINGLE,/*NEE*/NEE,/*per item tree*/PERITEMTREE> fpgr;

  fpgr.findFrequentItems(min_supp,fpstr,o_m,maxitem);
}


template<class S_C, class O_M, class B_T, bool SINGLE, NEELevel NEE, FirstLevel PERITEMTREE> void doFpGrowthClassicRebuild(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m) {
  log_perf(0,"Fp-growth configuration: +classic (node-based), %ssinglepath, -td, +project, +ProjMergeNodes, -ProjDeleteClosed NEE_%s, %s, BT_EndPatricia",SINGLE?"+":"-",NEE_to_String<NEE>::getString(),FirstLevel_to_String<PERITEMTREE>::getString());

  typedef bracz::ClassicRebuildFPStructure<S_C, /*per item tree*/PERITEMTREE,/*singlepath*/SINGLE
    > NonOrdFpStructure;

  NonOrdFpStructure fpstr(sorted_coder,maxitem,min_supp);

  bracz::FPGrowthBaseNoNEE<NonOrdFpStructure,O_M,/*singlepath*/SINGLE,/*NEE*/NEE,/*per item tree*/PERITEMTREE> fpgr;

  fpgr.findFrequentItems(min_supp,fpstr,o_m,maxitem);
}


template <class T_R, class DF_D> 
class FPGrowthSelector
{
private:
  typedef bracz::EndPatriciaBuildTree<false> B_T;
  enum EModify {
    EModifyNone,
    EModifySingleOff,
    EModifyNeeOff
  };
   public:
      FPGrowthSelector( 
	 counter_t min_supp, bool relative, double relminsupp, 
	 char* algorithm, char* input_file,  counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);

  template<class S_C,class O_M, bool SINGLE, NEELevel NEE> class Phase2 {
  public:
    //    template<class Type>
    //void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, Type);

    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_basic) {
      doFpGrowth<S_C,O_M,B_T,SINGLE,NEE,/*TD*/false,/*PROJECT*/false,   
        /*ProjectMergeNodes*/false, /*ProjectDeleteClosed*/false,
        FLBuildSingleTree,/*sparseaggr*/false>(sorted_coder,maxitem,min_supp,o_m);
    }

    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_dense) {
      doFpGrowth<S_C,O_M,B_T,SINGLE,NEE,/*TD*/true,/*PROJECT*/false,   
        /*ProjectMergeNodes*/false, /*ProjectDeleteClosed*/false,
        FLBuildSingleTree,/*sparseaggr*/false>(sorted_coder,maxitem,min_supp,o_m);
    }

    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_sparse) {
      doFpGrowth<S_C,O_M,B_T,SINGLE,NEE,/*TD*/false,/*PROJECT*/true,   
        /*ProjectMergeNodes*/true, /*ProjectDeleteClosed*/false,
        FLSimultProject,/*sparseaggr*/false>(sorted_coder,maxitem,min_supp,o_m);
    }
    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_nsparse) {
      doFpGrowth<S_C,O_M,B_T,SINGLE,NEE,/*TD*/false,/*PROJECT*/true,   
        /*ProjectMergeNodes*/true, /*ProjectDeleteClosed*/false,
        FLSimultProject,/*sparseaggr*/true>(sorted_coder,maxitem,min_supp,o_m);
    }
    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_classictd) {
      doFpGrowthClassic<S_C,O_M,B_T,SINGLE,NEE,/*TD*/true,
        FLBuildSingleTree>(sorted_coder,maxitem,min_supp,o_m);
    }

    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_classictd_alll1trees) {
      doFpGrowthClassic<S_C,O_M,B_T,SINGLE,NEE,/*TD*/true,
        FLBuildAllL1Trees>(sorted_coder,maxitem,min_supp,o_m);
    }

    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_classicrebuild) {
      doFpGrowthClassicRebuild<S_C,O_M,B_T,SINGLE,NEE,
        FLBuildSingleTree>(sorted_coder,maxitem,min_supp,o_m);
    }
    static void doPhase2(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, 
                  type_nonordfp_classicrebuild_alll1trees) {
      doFpGrowthClassicRebuild<S_C,O_M,B_T,SINGLE,NEE,
        FLBuildAllL1Trees>(sorted_coder,maxitem,min_supp,o_m);
    }


  };

  template<class S_C,class O_M, class Type> 
  void Phase1(S_C *sorted_coder, item_t maxitem,counter_t min_supp,O_M &o_m, Type t, EModify mod) {
    switch(mod) {
    case EModifyNone: {
      Phase2<S_C,O_M,true,NEE_Full>::doPhase2(sorted_coder,maxitem,min_supp,o_m,t);
      break;
    }
#if 1
    case EModifySingleOff: {
      Phase2<S_C,O_M,false,NEE_Full>::doPhase2(sorted_coder,maxitem,min_supp,o_m,t);
      break;
    }
    case EModifyNeeOff: {
      Phase2<S_C,O_M,true,NEE_Off>::doPhase2(sorted_coder,maxitem,min_supp,o_m,t);
      break;
      }
#endif
    default:
      log_err(0,"Unknown (or disabled) modification requested.");
    }
  }
  

};








template <class T_R, class DF_D>
FPGrowthSelector<T_R, DF_D> ::FPGrowthSelector(
   counter_t min_supp, bool relative, double relminsupp, 
   char* algorithm, char* input_file,  counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef SortedTransactionReader<Coder<T_R, DF_D>, true> S_C;
   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = DESC;
   S_C sorted_coder(&par_c);
   item_t maxitem=freq_items_with_counters.size();
   //no output
   /*
   typedef LStatOutput<OManager> NEWOM;
   OManager oldom("");
   NEWOM o_m(oldom,maxitem);
   */
   //output
   /*typedef LStatOutput<D> NEWOM;
   NEWOM o_m(decoder,maxitem);
   */
#if 0
#if 0
   typedef bracz::NonOptDFOutput NEWOM;
#else
   typedef NoOutput<true> NEWOM;
#endif
   NEWOM::params_t par2;
   NEWOM o_m(&par2);
#else
   typedef DF_D NEWOM;
   #define o_m df_decoder
#endif

  log_info(0,"Min supp %d, Freq item count %d, trans count %d",
           min_supp,maxitem,nr_of_transactions);

   //typedef bracz::TrieBuildTree B_T;
   

   char *kind=algorithm+9;
   char *sub=strchr(kind,'_');
   if (!sub) sub=kind+strlen(kind);
   std::string skind(kind, sub-kind);
   std::string ssub(sub);
   
   log_info(0,"kind >>%s<< sub >>%s<<",skind.c_str(),ssub.c_str());

   EModify mod=EModifyNone;

   if (ssub=="") {
     mod=EModifyNone;
   } else if (ssub=="_neeoff") {
     mod=EModifyNeeOff;
     log_info(0,"requested modify nee off");
   } else if (ssub=="_-single") {
     mod=EModifySingleOff;
     log_info(0,"requested modify singlepath off");
   }

   if ((skind=="")||(skind=="-nonordfp-basic")) {
     log_info(0,"Using standard nonordfp.");
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_basic(),mod);
   } else if (skind=="-nonordfp-dense") {
     log_info(0,"requested dense nonordfp.");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_dense(),mod);
   } else if (skind=="-nonordfp-sparse") {
     log_info(0,"requested sparse nonordfp.");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_sparse(),mod);
   } else if (skind=="-nonordfp-nsparse") {
     log_info(0,"requested sparse nonordfp.");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_nsparse(),mod);
   } else if (skind=="-classic-td") {
     log_info(0,"requested classic td.");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_classictd(),mod);
   } else if (skind=="-classic-td-l1") {
     log_info(0,"requested classic td fp-growth with all l1 trees built.");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_classictd_alll1trees(),mod);
   } else if (skind=="-classic-rebuild") {
     log_info(0,"requested classic rebuild fp-growth (note: buildtree is ignored).");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_classicrebuild(),mod);
   } else if (skind=="-classic-rebuild-l1") {
     log_info(0,"requested classic rebuild fp-growth with all l1 trees built (note: buildtree is ignored).");     
     Phase1(&sorted_coder,maxitem,min_supp,o_m,type_nonordfp_classicrebuild_alll1trees(),mod);
   }else {
     log_err(0,"unknown basic type >>%s<< requested.",skind.c_str());
   }


}//fp-growth

#endif
