#ifndef FDRepr_HPP
#define FDRepr_HPP

/**
 *@author Balazs RACZ
 */

#include <stdio.h>
#include <string>
#include <cerrno>
#include <string.h>
#include "io/FileReprBase.hpp"  
#include <sys/types.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include "common/log.h"
#include <iostream>

/**
   The file descriptor based file representation.

   This is the fastest but also the most complicated file representaion.
 */
class FDRepr : public FileReprBase
{
public:
  typedef FileReprBase::params_t params_t;

      FDRepr( const params_t* par ) 
	 throw(std::ios_base::failure) : FileReprBase(par)
      {
	 if(par->mode == WRITE)
	 {
	    fdout=open(par->file_name,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	    if (fdout<0) {
	       std::string s(strerror(errno));
	       log_err(0,"error opening output file %s: %s",par->file_name,s.c_str());
	       throw std::ios_base::failure(s);
	    }
	    bufend=file_buffer+file_buffer_size;
	    buffree=file_buffer;
	 }
	 else
	 {
	    fdout=open(par->file_name,O_RDONLY);
	    if (fdout<0) {
	       std::string s(strerror(errno));
	       log_err(0,"error opening input file %s: %s",par->file_name,s.c_str());
	       throw std::ios_base::failure(s);
	    }
	 }
      }

  ~FDRepr()
  {
    if (mode==WRITE)
      flush();
    close(fdout);
  }
  void rewind()
  {
    lseek(fdout,0,SEEK_SET);
  }
  void flush()
  {
    bufend=buffree;
    buffree=file_buffer;
    while (buffree<bufend) {
      int ret=::write(fdout,buffree,bufend-buffree);
      if (ret<0) {
        perror("write");
      } else
        buffree+=ret;
    }
    bufend=file_buffer+file_buffer_size;
    buffree=file_buffer;
  }

  std::streamsize readFromFile(char* buffer, std::streamsize buffer_size)
  {
    //we make no use of the file buffer in the read case
    ssize_t ret=read(fdout,buffer,buffer_size);
    if (ret==-1) {
      perror("read");
      return 0;
    }
    return ret;
  }

  void writeToFile( const char* buf, std::streamsize len )
  {
    if ((int)len>bufend-buffree) {
      int part=bufend-buffree;
      mempcpy(buffree,buf,part);
      buffree=file_buffer;
      while (buffree<bufend) {
        int ret=::write(fdout,buffree,bufend-buffree);
        if (ret<0) {
          perror("write");
        } else
          buffree+=ret;
      }
      buffree=(char*)mempcpy(file_buffer,buf+part,len-part);
    } else {
      buffree=(char*)mempcpy(buffree,buf,len);
    }
  }
protected :
  /// The file that contain the transactions
  int fdout;
private :
  /// iostream buffer
  char *bufend,*buffree;

};


#endif
