#ifndef FILERepr_HPP
#define FILERepr_HPP

/**
 *@author Ferenc Bodon
 */
#include <stdio.h>
#include <cerrno>
#include "common/log.h"
#include "io/FileReprBase.hpp"  

/**
   The FILE*-based file representation.

   This representation is very simple, readable and fast. Our experiments
   showed, however, that it is still a little bit slower than FDRepr 
   (file descriptor based) representation.
 */
class FILERepr : public FileReprBase
{
   public:
      typedef FileReprBase::params_t params_t;

      FILERepr( const params_t* par ) 
	 throw(std::ios_base::failure) : FileReprBase(par)
      {
	 if(par->mode == READ)
	 {
	    file = fopen(par->file_name,"r");
	    if (!file) 
	    {
	       std::string s(strerror(errno));
	       log_err(0,"error opening input file %s: %s", 
		       par->file_name, s.c_str());
	       throw std::ios_base::failure(strerror(errno));
	    }
	 }
	 else
	 {
	    file = fopen(par->file_name, "wt");
	    if (!file) 
	    {
	       std::string s(strerror(errno));
	       log_err(0,"error opening output file %s: %s", 
		       par->file_name, s.c_str());
	       throw std::ios_base::failure(s);
	    }
	 }
	 setbuffer(file,file_buffer,par->file_buffer_size);
      }

      ~FILERepr()
      {
	 if(file)
	    fclose(file);
      }
      void rewind()
      {
	 fseek(file, SEEK_SET, 0);
      }
      void flush()
      {
	 fflush(file);
      }

      std::streamsize readFromFile(char* buffer, std::streamsize buffer_size)
      {
	 return fread(buffer,1,buffer_size, file);
      }

      void writeToFile( const char* str, std::streamsize n )
      {
	 fwrite( str, n, 1, file );
      }
   protected :
      /// The file that contain the transactions
      FILE* file;

};

#endif
