#ifndef FileReprBase_HPP
#define FileReprBase_HPP

/**
 *@author Ferenc Bodon
 */

#include <fstream>

/** This class represents the base class of a file representation.

There are many ways of representing a file in C++. We created an IO framework
where the file representation is changeable. This way we could test
the efficacy of the different solutions. Our experiment showed that 
file descriptor-based representation is the fastest.
*/
class FileReprBase
{
public:
      enum Mode
      {
	 READ,
	 WRITE,
      };

      /** The parameter class of the IO constructor
       */
      class params_t
      {
	 public:
	    /// the name of the file
	    char* file_name;
	    /// the size of the lowest level buffer
	    std::streamsize file_buffer_size;

	    Mode mode;

	    params_t()
	    {
	       file_name = NULL;
	       file_buffer_size = 128 * 1024;
	       mode = READ;
	    }
      };
      FileReprBase(const params_t* par) : 
	 file_buffer_size(par->file_buffer_size)
      {
	 file_buffer = new char[par->file_buffer_size];
         this->mode=par->mode;
      }

      ~FileReprBase()
      {
	 delete [] file_buffer;
      }
      /** It rewinds the file to the beginning
	  
      Obviously this method makes sense when the file is opened for reading.
       */
      void rewind(){}

      /// It flushes the low-level buffer.
      void flush(){}

      /** This method reads in a block of characters from the file and 
	  places it in a buffer.

	  \param buffer the buffer where the chacters are copied to
	  \param buffer_size the number of characters to read in
	  \return the number of character that has succesfully read in. If the
	  end of the file is reached before the buffer could be filled in, then 
	  this value will be smaller than \p buffer_size.
       */
      std::streamsize readFromFile(char* buffer, std::streamsize buffer_size)
      {
	 return 0;
      }

      /** This method writes out a block of characters to the file.

	  \param buffer the buffer that contains the characters
	  \param n the number of characters to write out
       */
      void writeToFile( const char* buffer, std::streamsize n ){}


   protected:
      /// the low-level buffer
      char* file_buffer;
      /// the size of the lowest level buffer
      std::streamsize file_buffer_size;
      ///operating mode
      Mode mode;

};

#endif
