#ifndef StreamRepr_HPP
#define StreamRepr_HPP

/**
 *@author Ferenc Bodon
 */

#include <fstream>
#include <cerrno>
#include "common/log.h"
#include  "io/FileReprBase.hpp"

/**
   Stream-based file representation.

   This is the simplest, but also the slowest file representation.
 */
class StreamRepr : public FileReprBase
{
   public:
      typedef FileReprBase::params_t params_t;

      StreamRepr( const params_t* par ) 
	 throw(std::ios_base::failure) : FileReprBase(par)
      {
	 if(par->mode == READ)
	 {
	    input_stream.open(par->file_name, std::ios::in);
	    if ( !input_stream.is_open() )
	    {
	       std::string s(strerror(errno));
	       log_err(0,"error opening input file %s: %s", par->file_name, s.c_str());
	       throw std::ios_base::failure(s);
	    }
	    input_stream.rdbuf()->pubsetbuf(file_buffer, par->file_buffer_size);
	 }
	 else
	 {
	    output_stream.open(par->file_name, std::ios::out);
	    if(!output_stream.is_open())
	    {
	       std::string s(strerror(errno));
	       log_err(0,"error opening output file %s: %s", par->file_name, s.c_str());
	       throw std::ios_base::failure(s);
	    }
	    output_stream.rdbuf()->pubsetbuf(file_buffer, par->file_buffer_size);
	 }
      }

      ~StreamRepr()
      {
	 if ( input_stream.is_open() )
	    input_stream.close();
	 if ( output_stream.is_open() )
	    output_stream.close();
      }

      void rewind()
      {
	 input_stream.clear();
	 input_stream.seekg(0, std::ios::beg);
      }
      void flush()
      {
	 output_stream.flush();
      }

      std::streamsize readFromFile(char* buffer, std::streamsize buffer_size)
      {
	 return input_stream.readsome(buffer, buffer_size);
      }
      void writeToFile( const char* str, std::streamsize n )
      {
	 output_stream.write( str, n );
      }
   protected:
      /// stream to read
      std::ifstream input_stream;
      /// steam to write
      std::ofstream output_stream;
};

#endif
