#ifndef StatCoder_HPP
#define StatCoder_HPP

/**
  *@author Ferenc Bodon
  */
  
#include "common.hpp"
#include <iostream>


/** This class assigns 0,1,... new codes 
    to the frequent items. Actually, this class can be regarded 
    as a wrapper around a transaction reader, that changes the items
    of the read in transaction with new codes. Obviuosly, when an
    itemset is written out, the items must be decoded. Writting out a 
    coded itemset, however, is not the duty of the StatCoder class but rather
    the duty of a decoder. After the StatCoder generated the new codes it calls
    the \c setCodeInverse function of the decoder.

    \param T_R the transaction reader
    \param D the decoder
*/
template <class CODER>
class StatCoder : public CODER
{
   public:
      typedef typename CODER::params_t params_t;
      StatCoder(const params_t* par) : CODER(par), first_run(true),
      nr_of_transactions(0), sum_of_tr_sizes(0){}
      template <class BIS> counter_t nextTransactionBIS( BIS& transaction )
      {
	 if(first_run)
	 {
	    counter_t return_value = CODER::nextTransactionBIS(transaction);
	    if(return_value)
	    {
	       ++nr_of_transactions;
	       sum_of_tr_sizes += transaction.size();
	    }
	    else
	    {
	       log_info(0,"Average size of the coded/filtered transactions: %lg", static_cast<double>(sum_of_tr_sizes)/nr_of_transactions);
	       first_run = false;
	    }
	    return return_value;
	 }
	 else 
	    return CODER::nextTransactionBIS(transaction);

      }
      template <class UAC> counter_t nextTransactionUAC( UAC& transaction )
      {
	 if(first_run)
	 {
	    counter_t return_value = CODER::nextTransactionUAC(transaction);
	    if(return_value)
	    {
	       ++nr_of_transactions;
	       sum_of_tr_sizes += transaction.size();
	    }
	    else
	       first_run = false;
	    return return_value;
	 }
	 else 
	    return CODER::nextTransactionUAC(transaction);
      }

   protected:
      bool first_run;
      unsigned int nr_of_transactions;
      unsigned int sum_of_tr_sizes;
};


#endif
