#ifndef CacheDecoderBase_HPP
#define CacheDecoderBase_HPP

/**
  *@author Ferenc Bodon
  */
  
#include "io/codec/decoder/DecoderBase.hpp"
#include "io/output/BufferedOutput.hpp"
#include "util/Formatter.hpp"
#include "io/FDRepr.hpp"  
#include <vector>
#include <iostream>

/** This decoder class stores the string representation 
    of the original codes. With this solution we can spare
    the expensive int to string conversion when writting out
    an itemset.
*/
template <class FR = FDRepr>
class CacheDecoderBase : public DecoderBase< BufferedOutput<FR> >
{
   public:
     
      typedef typename DecoderBase< BufferedOutput<FR> >::params_t params_t;
      CacheDecoderBase(const params_t* par)
	 : DecoderBase< BufferedOutput<FR> >(par){}

      ~CacheDecoderBase()
      {
	 for(std::vector< std::pair<char*, unsigned int> >::iterator it = 
		code_inverse.begin(); it != code_inverse.end(); ++it)
	    delete [] (*it).first;
      }

	 
      void setCodeInverse( const std::vector<item_t>& code_inverse )
      {	 
	 char* temp_string;
	 std::streamsize pos;
	 unsigned int length;
	 std::pair<char*, unsigned int> temp_pair;
	 for( std::vector<item_t>::const_iterator it = code_inverse.begin();
	      it != code_inverse.end(); ++it)
	 {
	    pos = this->buffer_size;
	    Formatter::IntToStringBack(*it, this->buffer, pos);
	    length = this->buffer_size - pos;
	    temp_string = new char[length];
	    memcpy( temp_string, &(this->buffer)[pos], length );
	    temp_pair.first = temp_string;
	    temp_pair.second = length;
	    this->code_inverse.push_back(temp_pair);
	 }
      }	 
      
   protected:
      /** The inverse of code vector.

      \a code_inverse[i] stores the string representation of that item 
      whose new code is \e i.
       */
      std::vector< std::pair<char*, unsigned int> > code_inverse;
};

#endif
