#ifndef SimpleBufferedDFDecoder_HPP
#define SimpleBufferedDFDecoder_HPP

/**
 *@author Ferenc Bodon
 */

#include "io/output/BufferedOutput.hpp"  
#include "io/codec/decoder/SimpleDecoderBase.hpp"  
#include "util/Formatter.hpp"  
#include "io/FDRepr.hpp"  

template <class FR = FDRepr>
class SimpleBufferedDFDecoder : 
   public SimpleDecoderBase< BufferedOutput<FR> >
{
   protected:
      typedef SimpleDecoderBase< BufferedOutput<FR> > PARENT;
   public:
      typedef typename SimpleDecoderBase< BufferedOutput<FR> >::params_t params_t;
      SimpleBufferedDFDecoder(const params_t* par)
	 : PARENT(par)
      {positions.push(0);}

      static bool DINLINE isDFO()
      {
	 return true;
      }  

      void pushItem(const item_t item);

      void write(counter_t support);

      void pushItemWithWrite(item_t item, counter_t support)
      {
	 pushItem(item);
	 write(support);	    
      }
	 
      void pushItemWithPrevSupport(item_t item)
      {
	 pushItemWithWrite(item, last_support);
      }
	 
      void popItem()
      {
	 assert(!positions.empty());
	 positions.pop();
      }
      void popAll()
      {
	 while( !positions.empty() )
	    positions.pop();
	 positions.push(0);
      }

   private:
      std::stack<std::streamsize> positions;
      counter_t last_support;

};
template <class FR> inline void
SimpleBufferedDFDecoder<FR>::pushItem(const item_t item)
{
   register std::streamsize pos = positions.top();
   Formatter::IntToStringFront(PARENT::code_inverse[item], PARENT::buffer, pos);
   PARENT::buffer[pos++] = ' ';
   positions.push(pos);
}
template <class FR> inline void
SimpleBufferedDFDecoder<FR>::write(counter_t support)
{
   last_support = support;
   register std::streamsize pos = positions.top();
   PARENT::buffer[pos++] = '(';
   Formatter::IntToStringFront(support, PARENT::buffer, pos);
   PARENT::buffer[pos++] = ')';
   PARENT::buffer[pos++] = '\n';
   writeToFile( PARENT::buffer, pos );
}
#endif
