/*!
  \file   BinaryNormalDecoder.hpp
  \brief  A normal decoder and output manager 
     that writes the output in binary form.

  \author Ferenc BODON
  \date   2005-05-02
*/
#ifndef BinaryNormalDecoder_HPP
#define BinaryNormalDecoder_HPP


#include <iterator>
#include "io/codec/decoder/SimpleDecoderBase.hpp"  
#include "io/FDRepr.hpp"  

  /**
     FR must be a File representation that 
     implements \c writeToFile function.
  */
  template <class FR = FDRepr>
  class BinaryNormalDecoder : public SimpleDecoderBase<FR>
  {
  public:

    typedef typename SimpleDecoderBase<FR>::params_t params_t;
    BinaryNormalDecoder(const params_t* par)
      : SimpleDecoderBase<FR>(par){}

    static bool DINLINE isDFO()
    {
      return false;
    }  


    template <typename InputIterator> 
    void writeItemsetAndCounter(
       InputIterator first, InputIterator last, 
       const counter_t support )
    {
      static std::vector<item_t> decoded_set;
      decoded_set.resize(2);
      decoded_set[0]=support;
      while(first != last)
        {
          decoded_set.push_back(code_inverse[*first]);
          ++first;
        }
      decoded_set[1]=decoded_set.size()-2;
      writeToFile( (char*)&(decoded_set[0]),
		   decoded_set.size()*sizeof(item_t) );
    };

#endif
