#ifndef DBCache_HPP
#define DBCache_HPP

/**
  *@author Ferenc Bodon
  */
  
#include "../../common.hpp"


/**
   This is the interface class of a database cacher.

   <p>
   A transaction cacher is actually a wrapper around a transaction
   reader (see: class TransactionReader). In general a database cacher
   collects the same transactions together and stores them in an
   efficient way. You can get each transaction one-by-one. For this
   use nextTransactionBIS or nextTransactionUAC methods.
   </p

   <p> There are many transaction cacher. They differ in the data
   structure that stores the transactions. 
   <ul>
   <li>
   OrderedVectorDBCache stores the transactions in a simple ordered vector,
   </li>
   <li>
   OrderedArrayDBCache stores the transactions in a vector of ordered vectors,
   </li>
   <li>
   TrieDBCache stores the transactions in a general trie,
   </li>
   <li>
   RBTreeDBCache is a transaction cacher that store the transactions in a red-black tree,
   </li>
   <li>
   BuildTreeDBCache is a Patricia trie database cacher.
   </li>
   </ul>
   </p>

   
 */
template <class T_R>
class DBCache : public T_R
{
   public:
      typedef typename T_R::params_t params_t;
      DBCache( const params_t* par ):
	 T_R(par){}

      template <class BIS> counter_t nextTransactionBIS( 
	 BIS& transaction ){}

      template <class UAC> bool nextTransactionUAC( 
	 UAC& transaction ){}

      /// see: TransactionReader::rewind
      void rewind(){}   
};

#endif
