#ifndef LBufferedTransactionReader_HPP
#define LBufferedTransactionReader_HPP

/**
  *@author Lars Schmidt-Thieme simplified and embedded by Ferenc Bodon
  */

#include "io/input/transaction_reader/BufferedTransactionReader.hpp"  

// This class represents a buffered TransactionReader.
template <class T_R = TransactionReader<> >
class LBufferedTransactionReader : public BufferedTransactionReader<T_R>
{
   private:
      typedef BufferedTransactionReader<T_R> PARENT;
   public:
      typedef typename BufferedTransactionReader<T_R>::params_t params_t;
      LBufferedTransactionReader( const params_t* par )
	 : BufferedTransactionReader<T_R>( par)
      {
	 bufLength = readFromFile(PARENT::buffer, PARENT::buffer_size);
	 bufPosition = 0;
      }

      /// see: TransactionReader::nextTransactionBIS
      template <class BIS> counter_t nextTransactionBIS( BIS& transaction );

      /// see: TransactionReader::nextTransactionUAC
      template <class UAC> counter_t nextTransactionUAC( UAC& transaction );

      /// see: TransactionReader::rewind
      void rewind()
      {
	 BufferedTransactionReader<T_R>::rewind();
	 bufLength = readFromFile(PARENT::buffer, PARENT::buffer_size);
	 bufPosition = 0;
      }


   private:
      /// The position of the next char to process
      std::streamsize bufPosition;
      /// The number chars that were succesfully read from the file to the buffer
      std::streamsize bufLength;

   private:

      /// see: \ref Bodon::SimpleTransactionReader::getNextItem
      inline bool DINLINE getNextItem(int& item, char& c)
      {
	 item = 0;
	 bool has_next_int = false;
	 c = PARENT::buffer[bufPosition];
	 ++bufPosition;
	 if( bufPosition >= bufLength ) 
	 {
	    bufLength = readFromFile(PARENT::buffer, PARENT::buffer_size);
	    bufPosition = 0;
	 }
	    
	 while( (bufLength != 0) && (c >= '0') && (c <= '9') )
	 {
	    has_next_int = true;
	    item = 10 * item + (c - '0');
	    c = PARENT::buffer[bufPosition];
	    ++bufPosition;
	    if (bufPosition >= bufLength) 
	    {
	       bufLength = readFromFile(PARENT::buffer, PARENT::buffer_size);
	       bufPosition = 0;
	    }
	 }
	 return has_next_int;
      }
};

template <class T_R> template <class BIS> inline counter_t 
LBufferedTransactionReader<T_R>::nextTransactionBIS( BIS& transaction )
{
   transaction.clear();
   if( bufLength == 0 ) return 0;
   char          c;
   int           item;

   do
   {
      if( getNextItem(item, c) )
	 transaction.push_back( (item_t) item );
   }
   while( (bufLength != 0) && (c != '\n') );
   return 1;
}

template <class T_R> template <class UAC> inline counter_t 
LBufferedTransactionReader<T_R>::nextTransactionUAC( UAC& transaction )
{
   transaction.clear();
   if( bufLength == 0 ) return 0;
   int item;
   char c;

   do
   {
      if( getNextItem(item, c) ) 
	 transaction.insert( (item_t) item );
   }
   while( (bufLength != 0) && (c != '\n') );
   return 1;
}


#endif
