#ifndef SimpleTransactionReader_HPP
#define SimpleTransactionReader_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/StreamRepr.hpp"  
#include "io/input/transaction_reader/TransactionReader.hpp"  

/** This is a class that implements 
    StreamTransactionReader::nextTransactionBIS and 
    StreamTransactionReader::nextTransactionBIS procedures. 
    The main purpose of this class is to provide a basic class for 
    the tests. If the size of the input is large, and the performance 
    is important, you should use a BufferedTransactionReader like 
    LBufferedTransactionReader or BufferedTransactionReader
    For more information about the performance of the different input 
    classes test_results.html.
    \see test_results.html 
*/
class SimpleTransactionReader : public TransactionReader<StreamRepr>
{
   public:

      SimpleTransactionReader( const params_t* par ) :
	 TransactionReader<StreamRepr>(par){}

      template <class BIS> counter_t nextTransactionBIS( BIS& transaction );

      template <class UAC> counter_t nextTransactionUAC( UAC& transaction );

   private:
      /** This inline method is responsible to read in one item
       *
       * \param item this will store the item
       * \param c the last character read by the function
       * \return true if an item was succesfully read in
       */
      bool getNextItem(int& item, char& c);
};

template <class BIS> inline counter_t SimpleTransactionReader::
nextTransactionBIS( BIS& transaction )
{
   transaction.clear();
   if( input_stream.eof() ) return 0;
   char          c;
   int           item;
      
   do
   {
      if( getNextItem(item, c) )
	 transaction.push_back( (item_t) item );
   }
   while( !input_stream.eof() && c != '\n' );
//   copy(transaction.begin(),transaction.end(),ostream_iterator<item_t>(cout," "));
//   cout<<"\n";
   return 1;
}

template <class UAC> inline counter_t SimpleTransactionReader::
nextTransactionUAC( UAC& transaction )
{
   transaction.clear();
   if( input_stream.eof() ) return 0;
   int item;
   char c;
   do
   {
      if( getNextItem(item, c) ) 
	 transaction.insert( (item_t) item );
   }
   while( !input_stream.eof() && c != '\n' );
   return 1;
}

inline bool SimpleTransactionReader::getNextItem(int& item, char& c)
{
   item = 0;
   bool has_next_int;
   has_next_int = false;
   input_stream.get(c);
   while(input_stream.good() && (c >= '0') && (c <= '9'))
   {
      has_next_int = true;
      item = 10 * item + int(c) - int('0');
      input_stream.get(c);
   }
   return has_next_int;
}

#endif
