#ifndef SortedTransactionReader_HPP
#define SortedTransactionReader_HPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"
#include "io/input/transaction_reader/TransactionReader.hpp"  
#include <algorithm>

/** Use this class if you want, that a transaction that is read in
is sorted and the duplicates are removed. 

If the UNIQUE template parameter is true than duplicates will be removed, 
otherwise just a sorting is made.
*/
template <class T_R = TransactionReader<>, bool UNIQUE = false, bool ASCENDING=true> 
class SortedTransactionReader : public T_R
{
   public:
      typedef typename T_R::params_t params_t;
      SortedTransactionReader( const params_t* par ) : T_R(par){}

      template <class BIS> counter_t nextTransactionBIS( BIS& transaction )
      {
	 counter_t return_value = T_R::nextTransactionBIS(transaction);
	 if( ASCENDING )
	    std::sort(transaction.begin(), transaction.end());
	 else 
	    std::sort(transaction.begin(), transaction.end(), std::greater<item_t>());
	 if( UNIQUE )
	    transaction.erase(
	       std::unique( transaction.begin(), transaction.end() ),
	       transaction.end() );	 
	 return return_value;
      }
};

#endif
