#ifndef StatInput_HPP
#define StatInput_HPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"  
#include "common/log.h"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"

#include <vector>
/** 
    An input manager class that collects some statistics (number of
    transactions, number of items, average size of the transactions)
    about the transation database.

    This class is a wrapper around a transaction_reader. Collecting
    statistics needs resources thus do not use this class unless you are
    not interesred in performance.
*/
template <class TR = brBufferedTransactionReader< > > 
class StatInput : public TR
{
      unsigned int nr_of_transaction, sum_of_tr_sizes;
      std::vector<bool> items;
      bool first_run;

      counter_t temp_ret_value;
public:
      typedef typename TR::params_t params_t;
      StatInput( const params_t* par ):
	 TR(par), nr_of_transaction(0), 
	 sum_of_tr_sizes(0), first_run(true) { }
      template <class BIS> counter_t nextTransactionBIS( BIS& transaction )
      {

	 if(first_run)
	 {
	    temp_ret_value = TR::nextTransactionBIS( transaction);
	    if(temp_ret_value)
	    {
	       nr_of_transaction += temp_ret_value;
	       sum_of_tr_sizes += transaction.size();
	       for(typename BIS::iterator it = transaction.begin(); 
		  it != transaction.end(); ++it)
	       {
		  if(*it >= items.size())
		     items.resize(*it + 1, false);
		  items[*it] = true;			
	       }
	    }
	    else
	       first_run = false;
	    return temp_ret_value;
	 }
	 else
	    return TR::nextTransactionBIS( transaction);
      }
      ~StatInput()
      {
	 std::cout<<"Number of transactions: "<<nr_of_transaction<<std::endl;
	 std::cout<<"Average size of the transactions: "<<
		  static_cast<double>(sum_of_tr_sizes)/nr_of_transaction<<std::endl;
	 std::cout<<"Number of items: "<< 
	    std::count(items.begin(), items.end(), true)<<std::endl;	       
      }
};

#endif //StatInput_HPP
