#ifndef BinaryDFOutput_HPP
#define BinaryDFOutput_HPP

/*!
  \file   BinaryDFOutput.hpp
  \brief  A DF style output manager that writes the output in binary form.

  \author RACZ Balazs
  \date   2005-04-08
*/
#include <vector>

#include "io/output/WriteOutput.hpp"  

/**
   \brief  A DF style output manager that writes the output in binary form.


   WO must be a descendant of WriteOutput
*/
template <class WO>
class BinaryDFOutput : public WO
{
   public:

      typedef typename WO::params_t params_t;
      BinaryDFOutput(const params_t* par)
	 : WO(par){
	 output_set.push_back(0);//guard for support
	 output_set.push_back(0);//guard for itemset length
      }

      static bool DINLINE isDFO()
      {
	 return true;
      }  

      void pushItem(item_t item)
      {
	 output_set.push_back(item);
      }

      void write(counter_t support)
      {
	 output_set[0]=support;
	 output_set[1]=output_set.size()-2;
	 writeToFile((char*)&(output_set[0]),output_set.size()*sizeof(item_t));
	 last_support = support;
      }

      void pushItemWithWrite(item_t item, counter_t support)
      {
	 pushItem(item);
	 write(support);
      }
	 
      void pushItemWithPrevSupport(item_t item)
      {
	 pushItemWithWrite(item, last_support);
      }
	 
      void popItem()
      {
	 output_set.pop_back();
      }

      void popAll() {
	 output_set.resize(2);
      }

   private:
      std::vector<item_t> output_set; 
      counter_t last_support;

};

#endif
