/*!
  \file   BufferedOutput.hpp
  \brief  An abstract class for output managers that use buffering for the sake of
  efficient outputting the results.

  \author Ferenc BODON
  \date   2005-04-25
*/
#ifndef BufferedOutput_HPP
#define BufferedOutput_HPP


#include "io/output/OutputBase.hpp"
#include "io/FDRepr.hpp"  

template <class FR = FDRepr>
class BufferedOutput : public OutputBase<FR>
{
   public:
      class params_t : public OutputBase<FR>::params_t
      {
	 public:
	    std::streamsize buffer_size;
	    params_t():OutputBase<FR>::params_t()
	    {
	       buffer_size = 4 * 1024;
	    }
      };

      BufferedOutput(const params_t* par):
	 OutputBase<FR>(par),buffer_size(par->buffer_size)
      {
	 buffer =  new char[buffer_size];
      }

      ~BufferedOutput()
      {
	 delete [] buffer;
      }


   protected:
      /// The size of the buffer
      std::streamsize buffer_size;
      /// The buffer
      char* buffer;
};

#endif
