#ifndef NoOutput_HPP
#define NoOutput_HPP

/**
 *@author Ferenc Bodon
 */

#include "io/output/OutputBase.hpp"  
#include "io/FDRepr.hpp"  

/** This class represents a very simple OutputBase.

Actually it does nothing ;) Use this class if no output is needed.
*/
template <bool ISDFO, class FR = FDRepr>
class NoOutput : public OutputBase<FR>
{
   public:
      typedef  typename OutputBase<FR>::params_t params_t;
      NoOutput(const params_t* par) : OutputBase<FR>(par){}

      static bool DINLINE isDFO()
      {
	 return ISDFO;
      }  


      /// see OutputBase::writeItemsetAndCounter()
      template <typename InputIterator> void writeItemsetAndCounter(
 	 InputIterator first, InputIterator last, 
	 counter_t support )
      {
	 if( ISDFO )
	    OutputBase<FR>::writeItemsetAndCounter( first, last, support );
      }

//===== DFO functions
      void pushItem(item_t item)
      {
	 if( !ISDFO )
	    OutputBase<FR>::pushItem(item);
      }

      void write(counter_t support)
      {
	 if( !ISDFO )
	    OutputBase<FR>::write(support);
      }

      void pushItemWithWrite(item_t item, counter_t support)
      {
	 if( !ISDFO )
	    OutputBase<FR>::pushItemWithWrite(item, support);
      }

      void pushItemWithPrevSupport(item_t item)
      {
	 if( !ISDFO )
	    OutputBase<FR>::pushItemWithPrevSupport(item);

      }

      void popItem()
      {
	 if( !ISDFO )
	    OutputBase<FR>::popItem();
      }

      void popAll()
      {
	 if( !ISDFO )
	    OutputBase<FR>::popAll();

      }
//===== End of DFO functions
};

#endif
