#ifndef NonOptDFOutput_HPP
#define NonOptDFOutput_HPP

/*!
  \file   NonOptDFOutput.hpp
  \brief  A DF style output manager that does not write anything, only keeps the compiler from optimizing away output routines.

  \author RACZ Balazs
  \date   2005-04-08
*/
#include <vector>

#include "io/output/OutputBase.hpp"  

//! this variable is to make a reference for dummy data calculated
static volatile int dummy;

namespace bracz
{
  /**
     \brief  A DF style output manager that does not write anything, 
     only keeps the compiler from optimizing away output routines.

  */
   template <class FR>
  class NonOptDFOutput : public OutputBase<FR>
  {
  public:

    typedef  typename OutputBase<FR>::params_t params_t;
    NonOptDFOutput(const params_t* par)
      : OutputBase<FR>(par){
    }

    ~NonOptDFOutput() {
      dummy=dummysum;
    }

    static bool DINLINE isDFO()
    {
      return true;
    }  

    void pushItem(item_t item)
    {
      dummysum+=item;
    }

    void write(counter_t support)
    {
      dummysum+=support;
      last_support=support;
    }

    void pushItemWithWrite(item_t item, counter_t support)
    {
      pushItem(item);
      write(support);
    }
	 
    void pushItemWithPrevSupport(item_t item)
    {
      pushItemWithWrite(item, last_support);
    }
	 
    void popItem()
    {
      dummysum++;
    }

    void popAll() {
      dummysum++;
    }

  private:
    int dummysum;
    counter_t last_support;

  };
}
#endif
