#ifndef OutputBase_HPP
#define OutputBase_HPP

/**
 *@author Ferenc Bodon
 */

#include "common.hpp"  
#include "io/FDRepr.hpp"  
#include <iostream>  

/** This class represents the base class of output manager.

The output manager is responsibe to write out an itemset and its counter.
We distinquish two types of output functions. In the first class, one can 
output an itemset with its counter. The second type of functions belong
to the stack-based (also called by DFO) output solution. Here you can push and pop
an item to or from the stack, and write out the content of the stack together with a
counter.
*/
template <class FR = FDRepr>
class OutputBase : public FR
{
   public:
      typedef typename FR::params_t params_t;

      OutputBase(const params_t* par):FR(par) {}

      static bool DINLINE isDFO()
      {
	 std::cerr<<"Error: using function isDFO"<<std::endl;
	 return false;
      }  


//===== Normal functions 
      
      /** Writes out one itemset and its counter.
       *
       * \param first  An input iterator.
       * \param last  An input iterator.
       * \param support the support of the itemset
       *
       * The elements of the itemset are in the range [\a first, \a last).
       *
       * Here you can see an example of the usage \include example_output.cpp
       */
      template <typename InputIterator> void writeItemsetAndCounter(
 	 InputIterator first, InputIterator last, 
	 counter_t support )
      {
	 std::cerr<<"Error: using Normal function writeItemsetAndCounter"<<std::endl;
      }
//===== End of Normal functions 

//===== DFO functions
      /// Push an item to the top of the stack
      void pushItem(item_t item)
      {
	 std::cerr<<"Error: using DFO function pushItem"<<std::endl;
      }

      /// write the content of the stack together with a counter
      void write(counter_t support)
      {
	 std::cerr<<"Error: using DFO function write"<<std::endl;
      }
      /// Push an item to the top of the stack and 
      /// write the content of it together with a counter
      void pushItemWithWrite(item_t item, counter_t support)
      {
	 std::cerr<<"Error: using DFO function pushItemWithWrite"<<std::endl;
      }

      /// Push an item to the top of the stack and 
      /// write the content of it together with the 
      /// previously used counter
      void pushItemWithPrevSupport(item_t item)
      {
	 std::cerr<<"Error: using DFO function pushItemWithPrevSupport"<<std::endl;
      }

      /// Pop an item from the top of the stack
      void popItem()
      {
	 std::cerr<<"Error: using DFO function popItem"<<std::endl;
      }
      /// remove all items form the stack
      void popAll()
      {
	 std::cerr<<"Error: using DFO function popAll"<<std::endl;
      }

};

#endif
