#ifndef SortOutput_HPP
#define SortOutput_HPP

/**
 *@author Ferenc Bodon
 */

#include <algorithm>
#include "io/output/normal/BufferedNormalOutput.hpp" 

/**
   Use this class if you want the frequent itemset to be ordered when 
   writing them out.

   O must be an output manager that implements \c writeItemsetAndCounter.
 */
template <class O = BufferedNormalOutput<> >
class SortOutput : public O
{
   public:

      typedef typename O::params_t params_t;

      SortOutput(const params_t* par) : 
	 O(par) {}



      template <typename InputIterator> void writeItemsetAndCounter(
	 InputIterator first, InputIterator last, 
	 counter_t support )
      {
	 std::vector<item_t> sorted_itemset(first, last);
	 std::sort(sorted_itemset.begin(), sorted_itemset.end());
	 O::writeItemsetAndCounter(
	    sorted_itemset.begin(), sorted_itemset.end(), support);
      }
};

#endif
