

#include "common.hpp"
#include "common/log.h"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp" 


#include "io/codec/decoder/normal/CacheNormalDecoder.hpp"
//#include "io/output/StatOutput.hpp"
#include "io/output/normal/SortOutput.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"

#include "datastructures/maxvector.hpp"

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/input/transaction_reader/OrderReverser.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"
#include "util/Frequent2FilterOnline.cpp"
#include "apriori/bodon-vector/vector_typedef.hpp"
#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon-vector/SupportCounter.hpp"
#include "apriori/bodon-vector/CandidateGeneratorNoprune.hpp"
#include "apriori/bodon-vector/CandidateGeneratorSimplePrune.hpp"
#include "apriori/bodon-vector/InfrequentRemover.hpp"
#include "apriori/bodon-vector/FrequentItemInserter.hpp"
#include "apriori/bodon-vector/FrequentPairInserterNoprune.hpp"

#include "datastructures/PrefixArrayList.hpp"
#include "apriori/bodon-prefixarray/SupportCounter.hpp"
#include "apriori/bodon-prefixarray/CandidateGeneratorNoprune.hpp"
#include "apriori/bodon-prefixarray/InfrequentRemover.hpp"
#include "apriori/bodon-prefixarray/FrequentPairInserterNoprune.hpp"



#include "apriori/Apriori.hpp"

#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   log_err(0,"Usage: apriori-nontrie transactionfile min_supp outcomefile data_structure [maxsize]");
   log_err(0," transactionfile\t    file, that contains the tranasctions of items");
   log_err(0," outcomefile\t    file to write the outcome");
   log_err(0," min_supp\t    absolute support threshold");
   log_err(0," data_structure\t    data structure to store the candidates: i.e vector, prefixarray");

   std::cerr << file_format;
   log_err(0,"\t\t\tHave a succesful mining ;-)\n\n");
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   <li>4, if proper maxsize cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp, unsigned int& maxsize )
{
   if( argc < 5 )
   {
     log_err(0,"There are 4 mandatory arguments.");
     return 2;
   }
   
   std::string mins=argv[2];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_status(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false;
   int min_supp_i;
   try
   {
      convert(argv[2], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[2]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_status(0,"min_supp is set to %d", min_supp);

   if(argc == 6)
   {
      int maxsize_i;
      try
      {
	 convert(argv[5], maxsize_i);
	 if ( maxsize_i <= 0  )
	 {
	    log_err(0,"%s cannot be converted to a positive integer.",argv[4]);
	    return 4;
	 }
      }
      catch(BadConversion e)
      {
	 log_err(0,"min_supp conversion problem.");
	 return 4;
      }
      maxsize = static_cast<unsigned int>(maxsize_i);
      log_status(0,"maxsize is set to %d", maxsize);
   }
   else
      maxsize = largest_itemsetsize;
   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   unsigned int maxsize;
   bool relative;
   double relminsupp;
      
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, 
			    relative, relminsupp, maxsize );
      if(return_val)
	 return return_val;
   }

   char* input_file = argv[1];
   char* output_file = argv[3];

   try
   {
      // We assume that the transactions does not contain duplicates!!!
      typedef brBufferedTransactionReader< > T_R;
      // Otherwise uncomment this:
      // typedef SortedTransactionReader<brBufferedTransactionReader< >, true> T_R;

      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,  
				   nr_of_transactions, min_supp);

      if(!freq_items_with_counters.empty())
      {
	 log_status(0,"Doing decoder.");
	 typedef CacheNormalDecoder< > D;

	 D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
//	    par_d.numfreq = freq_items_with_counters.size(); // If StatOutput is used!!!
	 D decoder(&par_d);

// ascending order (also change par_c.codemode)
	 typedef SortedTransactionReader< Coder<T_R, D>, false, false > S_C_T_R;
	 typedef OrderReverser< bracz::BuildTreeDBCache< 
	    S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > >S_C;
//   no database caching:
//	 typedef SortedTransactionReader<Coder<T_R, D> >  S_C;
	 typedef Frequent2Filter<S_C> F2F;

	 S_C::params_t par_c;
	 par_c.file_name = input_file;
	 par_c.mode=FileReprBase::READ;
	 par_c.largest_item = tr_reader.getLargestItem();
	 par_c.decoder = &decoder;
	 par_c.freq_items_with_counters = &freq_items_with_counters;
	 par_c.codemode = ASC;
//   par_c.codemode = DESC;
	 log_status(0,"Doing sorted codec.");
	 S_C sorted_coder(&par_c);
	 std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
	    freq_pairs_with_counters;
	 F2F fr_2_filter(&sorted_coder );
	 log_status(0,"Finding frequent pairs.")
	    fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);
	 typedef bool DUMMY;
	 if( strstr(argv[4],"vector") )
	 {
	    log_status(0,"Candidates are stored in a vector");
	    typedef cand_vector_t DATA_STRUCTURE;
	    DATA_STRUCTURE candidates;
	    typedef Bodon::vector_based::FrequentItemInserter<DATA_STRUCTURE, D> FII;
	    FII fii(candidates, decoder);
	    typedef Bodon::vector_based::SupportCounter<DATA_STRUCTURE> SUPP_C_BASE;
	    typedef OneByOneSupportCounter<DATA_STRUCTURE, SUPP_C_BASE, S_C> SUPP_C;

	    typedef Bodon::vector_based::FrequentPairInserterNoprune<D, DUMMY> FPI;
//	    typedef Bodon::vector_based::CandidateGeneratorNoprune<D, DUMMY> CG;
	    typedef Bodon::vector_based::CandidateGeneratorSimplePrune<D, DUMMY> CG;
	    typedef Bodon::vector_based::InfrequentRemover<D, DUMMY> IR;
	    
	    typedef Apriori<S_C, D, DATA_STRUCTURE, DUMMY, FII, 
	       FPI, CG, IR, SUPP_C> A;
	    A apriori(candidates, sorted_coder, decoder, fii);
	    log_status(0,"Finding frequent itemsets.")
	       apriori.findFrequentItemsets( 
		  nr_of_transactions, *par_c.freq_counters,
		  freq_pairs_with_counters, min_supp, maxsize );
	 }
	 else
	 {
	    log_status(0,"Candidates are stored in a prefix-array.");
	    typedef PrefixArrayList<> DATA_STRUCTURE;
	    DATA_STRUCTURE candidates(freq_items_with_counters.size()-1);
	    typedef Bodon::vector_based::FrequentItemInserter<DATA_STRUCTURE, D> FII;
	    FII fii(candidates, decoder);
	    typedef Bodon::prefix_array::SupportCounter<DATA_STRUCTURE> SUPP_C_BASE;
	    typedef OneByOneSupportCounter<DATA_STRUCTURE, SUPP_C_BASE, S_C> SUPP_C;
	    typedef Bodon::prefix_array::FrequentPairInserterNoprune<DATA_STRUCTURE, D, DUMMY> FPI;
	    typedef Bodon::prefix_array::CandidateGeneratorNoprune<DATA_STRUCTURE, D, DUMMY> CG;
	    typedef Bodon::prefix_array::InfrequentRemover<DATA_STRUCTURE, D, DUMMY> IR;
	    typedef Apriori<S_C, D, DATA_STRUCTURE, DUMMY, FII, 
	       FPI, CG, IR, SUPP_C> A;
	    A apriori(candidates, sorted_coder, decoder, fii);
	    log_status(0,"Finding frequent itemsets.")
	       apriori.findFrequentItemsets( 
		  nr_of_transactions, *par_c.freq_counters,
		  freq_pairs_with_counters, min_supp, maxsize );
	 }
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


