#ifndef AprioriSelector_HPP
#define AprioriSelector_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/input/transaction_reader/OrderReverser.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"
//#include "util/Frequent2FilterOnline.cpp"

#include "common/allocators.hpp"


#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/dynamic_trie/trie_manipulators/SupportCounter.hpp"
#include "apriori/OneByOneSupportCounter.hpp"

#include "apriori/bodon/dynamic_trie/trie_manipulators/SimplePruner.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/IntersectProPruner.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/InfreqRemover.hpp"

#include "apriori/Apriori.hpp"

template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class T_R, 
	  class DF_D, NEELevel NEE> 
class AprioriSelector
{
   public:
      AprioriSelector( 
	 counter_t min_supp, char* algorithm, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder, unsigned int maxsize = largest_itemsetsize);
};

template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class T_R, 
	  class DF_D, NEELevel NEE>
AprioriSelector<TRIE_OEL, TRIE_OI, LEAF_WC, T_R, DF_D, NEE>::AprioriSelector( 
   counter_t min_supp, char* algorithm, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder, unsigned int maxsize )
{
   // ascending order (also change par_c.codemode)
   typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
   typedef OrderReverser< typename bracz::BuildTreeDBCache< 
      S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > >S_C;


   //  descending order (also change par_c.codemode):
//   typedef SortedTransactionReader< Coder<T_R, DF_D>, false, true > S_C_T_R; 
//  typedef bracz::BuildTreeDBCache< 
//      S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;

   //no database caching:
//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
//   par_c.codemode = DESC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);

   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
//   Frequent2FilterOnline<S_C> fr_2_filter(
//      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE_OEL main_trie;
   typedef Bodon::dynamic_trie::FrequentItemInserter<DF_D, TRIE_OEL, NEE> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::dynamic_trie::SupportCounter<TRIE_OEL, TRIE_OI> SUPP_C_BASE;
   typedef OneByOneSupportCounter<TRIE_OEL, SUPP_C_BASE, S_C> SUPP_C;
   typedef bracz::singleualloc<LEAF_WC, 64 * 1024> LEAF_ALLOCATOR;
   char *kind=algorithm+7;

   if( strstr(kind,"-noprune") )
   {
      log_info(0,"Complete pruning is disabled.");
      typedef Bodon::dynamic_trie::FrequentPairInserterNoprune<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::dynamic_trie::CandGenInfreqRemoveNopruneMerge<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE > CGIR;

      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, 
	 FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.")
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp, maxsize );
   }
   else if( strstr(kind,"-intersectprune") )
   {
      log_info(0,"Intersection-based pruning is selected.");
      typedef Bodon::dynamic_trie::FrequentPairInserter<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, 
	 LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::dynamic_trie::IntersectProPruner<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::dynamic_trie::CandidateGeneratorPrune<PRUNER, DF_D, 
	 TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::dynamic_trie::InfreqRemover<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp, maxsize );
   }
   else if( strstr(kind,"-simpleprune") )
   {
      log_info(0,"Simple pruning is selected.");
      typedef Bodon::dynamic_trie::FrequentPairInserter<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, 
	 LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::dynamic_trie::SimplePruner<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, 
	 LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::dynamic_trie::CandidateGeneratorPrune<PRUNER, DF_D, 
	 TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::dynamic_trie::InfreqRemover<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, 
	 LEAF_ALLOCATOR, NEE> IR;
      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.")
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp, maxsize );
	    }
   else 
   log_warn(0,"Unknown apriori type requested '%s'.",kind);
}

#endif
