#ifndef LeafRepresentationSelector_HPP
#define LeafRepresentationSelector_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"

#include "common/allocators.hpp"


#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge.hpp"
#include "apriori/OneByOneSupportCounter.hpp"

#include "apriori/bodon/trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/CandidateGeneratorPrune.hpp"

#include "apriori/bodon/trie/trie_manipulators/SimplePruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/SimplePruner.hpp"

#include "apriori/bodon/trie/trie_manipulators/IntersectProPruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/IntersectProPruner.hpp"

#include "apriori/bodon/trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"

#include "apriori/bodon/trie/trie_manipulators/InfreqRemover.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/InfreqRemover.hpp"

#include "apriori/Apriori.hpp"



template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D> 
class LeafRepresentationSelector
{
   public:
      LeafRepresentationSelector( 
	 counter_t min_supp, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);
};
/*template <class TRIE, class LEAF_WC, class T_R, class DF_D> 
class LeafRepresentationSelectorOld
{
   public:
      LeafRepresentationSelectorOld( 
	 counter_t min_supp, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);
};*/

template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D>
LeafRepresentationSelector<TRIE, LEAF_WC, LEAF_ALLOCATOR, T_R, DF_D>::
LeafRepresentationSelector( 
   counter_t min_supp, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef typename bracz::BuildTreeDBCache< SortedTransactionReader<Coder<T_R, DF_D> >,  
      std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;
//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);

   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE main_trie;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, true> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
   typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
//   typedef Bodon::dynamic_trie::SupportCounter<TRIE, INHOMOGENEOUS_TRIE> SC;

   log_info(0,"Complete pruning is disabled.");
   typedef Bodon::FrequentPairInserterNoprune<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, true> FPI;
   typedef Bodon::inhomogeneous_trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, true > CGIR;
   typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
   A apriori(main_trie, sorted_coder, df_decoder, fii);
   log_status(0,"Finding frequent itemsets.")
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
}

/*
template <class TRIE, class LEAF_WC, class T_R, class DF_D>
LeafRepresentationSelectorOld<TRIE, LEAF_WC, T_R, DF_D>::LeafRepresentationSelectorOld( 
   counter_t min_supp, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef typename bracz::BuildTreeDBCache< SortedTransactionReader<Coder<T_R, DF_D> >,  
      std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;
//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);

   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   typedef bracz::NewWrapperAlloc<LEAF_WC> LEAF_ALLOCATOR;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, true> FII;
   typedef Bodon::SupportCounterMerge<TRIE> SC;
   log_info(0,"Complete pruning is disabled.");
   typedef Bodon::FrequentPairInserterNoprune<DF_D, TRIE, TRIE, LEAF_ALLOCATOR, true> FPI;
   typedef Bodon::trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, true> CGIR;
   typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
   A apriori(main_trie, sorted_coder, df_decoder, fii);
   log_status(0,"Finding frequent itemsets.")
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
}
*/
#endif
