#ifndef ThresholdSelector_HPP
#define ThresholdSelector_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"

#include "util/StreamParser.hpp"

#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelistDynLookup.hpp"
#include "apriori/bodon/Trie.hpp"

#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge.hpp"
#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/IntersectProPruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/InfreqRemover.hpp"
#include "apriori/Apriori.hpp"


template <class VECTOR, class T_R, class DF_D> 
class ThresholdSelector
{
   public:
      ThresholdSelector( char* threshold,
	 counter_t min_supp, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& 
	 freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);

   private:
      std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
};

template <class VECTOR, class T_R, class DF_D>
ThresholdSelector<VECTOR, T_R, DF_D>::ThresholdSelector( char* threshold,
   counter_t min_supp, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef typename bracz::BuildTreeDBCache< SortedTransactionReader<Coder<T_R, DF_D>, false>,  
      std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);


   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   typedef Bodon::LeafWithoutConstructor LEAF_WC;	 
   typedef Bodon::Leaf LEAF;	 
   typedef bracz::singleualloc<LEAF_WC, 1024> LEAF_ALLOCATOR;
   const NEELevel NEE = NEE_Off;

   if( strcmp(threshold,"3") == 0 )
   {
      log_info(0,"Threshol is set to 3.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 3> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strcmp(threshold,"6") == 0 )
   {
      log_info(0,"Threshol is set to 6.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 6> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strcmp(threshold,"10") == 0 )
   {
      log_info(0,"Threshol is set to 10.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 10> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strcmp(threshold,"20") == 0 )
   {
      log_info(0,"Threshol is set to 20.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 20> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strcmp(threshold,"40") == 0 )
   {
      log_info(0,"Threshol is set to 40.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 40> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strcmp(threshold,"100") == 0 )
   {
      log_info(0,"Threshol is set to 100.");
      typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<VECTOR, 100> > TRIE;
      TRIE main_trie;
      typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
      FII fii(main_trie, df_decoder);
      typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::inhomogeneous_trie::IntersectProPruner<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::inhomogeneous_trie::InfreqRemover<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else 
      log_warn(0,"Unknown threshold value is requested '%s'.", threshold);
}
#endif
