#include "common.hpp"
#include "common/log.h"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp" 


#include "io/codec/decoder/df/SimpleDFDecoder.hpp"
//#include "io/codec/decoder/df/CacheDFDecoder.hpp"
#include "io/codec/decoder/df/DFDecoderWithEEManagement.hpp"
//#include "io/output/StatOutput.hpp"
#include "io/output/normal/SortOutput.hpp"
//#include "io/codec/decoder/df/DFDecoderWrapper.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"

#include "datastructures/maxvector.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"
//#include "datastructures/trie/edgelist/OrderedEdgelistDynLookup.hpp"
#include "datastructures/trie/edgelist/OffsetIndexVector.hpp"
#include "common/allocators.hpp"

#include "apriori/bodon/Leaf.hpp"
#include "apriori/bodon/Trie.hpp"
#include "apriori/bodon/TrieNEE.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "io/input/transaction_reader/OrderReverser.hpp"
#include "io/codec/coder/Coder.hpp"
#include "util/Frequent2Filter.cpp"
#include "util/Frequent2FilterOnline.cpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/IntersectProPruner.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/InfreqRemoverOrderPresAssumption.hpp"

#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge.hpp"

#include "apriori/Apriori.hpp"

#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   log_err(0,"Usage: orderPresAssumptionRatio transactionfile min_supp outcomefile [options]");
   log_err(0," deadend-option\t    option of dead-end pruning, i.e: on, off");
   log_err(0," transactionfile\t    file, that contains the tranasctions of items");
   log_err(0," outcomefile\t    file to write the outcome");
   log_err(0," min_supp\t    absolute support threshold");

   std::cerr << file_format;
   log_err(0,"\t\t\tHave a succesful mining ;-)\n\n");
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   <li>4, if proper maxsize cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp, unsigned int& maxsize )
{
   if ( argc < 4 )
   {
     log_err(0,"There are 3 mandatory arguments.");
     return 2;
   }
   std::string mins=argv[2];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_info(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false; relminsupp=0;
   int min_supp_i;
   try
   {
      convert(argv[2], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_info(0,"min_supp is set to %d", min_supp);
   if(argc == 5)
   {
      int maxsize_i;
      try
      {
	 convert(argv[4], maxsize_i);
	 if ( maxsize_i <= 0  )
	 {
	    log_err(0,"%s cannot be converted to a positive integer.",argv[4]);
	    return 4;
	 }
      }
      catch(BadConversion e)
      {
	 log_err(0,"max_size conversion problem.");
	 return 4;
      }
      maxsize = static_cast<unsigned int>(maxsize_i);
      log_status(0,"maxsize is set to %d", maxsize);
   }
   else
      maxsize = largest_itemsetsize;
   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   unsigned int maxsize;
   bool relative;
   double relminsupp;
      
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, 
			    relative, relminsupp, maxsize );
      if(return_val)
	 return return_val;
   }

   char* input_file = argv[1];
   char* output_file = argv[3];

   try
   {
      // We assume that the transactions does not contain duplicates!!!
      typedef brBufferedTransactionReader< > T_R;
      // Otherwise uncomment this:
      // typedef SortedTransactionReader<brBufferedTransactionReader< >, true> T_R;

      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,  
				   nr_of_transactions, min_supp);

      if(!freq_items_with_counters.empty())
      {
	 log_status(0,"Doing decoder.");
	 typedef DFDecoderWithEEManagement< > DF_D;
//	    typedef DFDecoderWithEEManagement< SimpleDFDecoder<SortOutput<> > > DF_D;
//	    typedef DFDecoderWithEEManagement< StatOutput<CacheDFDecoder< >, true> > DF_D;

	 DF_D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
//	    par_d.numfreq = freq_items_with_counters.size(); // If StatOutput is used!!!
	 DF_D df_decoder(&par_d);

	 log_status(0,"APRIORI is selected");
	 typedef Bodon::LeafWithoutConstructor LEAF_WC;	 
	 typedef Bodon::Leaf LEAF;	 
	 typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelist<std::vector<Edge> > > TRIE;
//	       typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<std::vector<Edge>, 30 > > TRIE_OEL;

	 	 
	 log_status(0,"NEE_FULL is disabled");
	 log_status(0,"Low memory need option is on.");
	 typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
	 const bool ENDONLY = true;
	 typedef OrderReverser< bracz::BuildTreeDBCache< 
	    S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<ENDONLY>, ENDONLY > >S_C;
	 typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
	 typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
	 typedef Frequent2FilterOnline<S_C> F2F;
	 
	 S_C::params_t par_c;
	 par_c.file_name = input_file;
	 par_c.mode=FileReprBase::READ;
	 par_c.largest_item = tr_reader.getLargestItem();
	 par_c.decoder = &df_decoder;
	 par_c.freq_items_with_counters = &freq_items_with_counters;
	 par_c.codemode = ASC;
//   par_c.codemode = DESC;
	 log_status(0,"Doing sorted codec.");
	 S_C sorted_coder(&par_c);
	 
	 std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
	    freq_pairs_with_counters;
	 F2F fr_2_filter(&sorted_coder );
	 log_status(0,"Finding frequent pairs.")
	    fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

	 TRIE main_trie;
	 typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE_Off> FII;
	 FII fii(main_trie, df_decoder);
	 typedef bracz::singleualloc<LEAF_WC, 64 * 1024> LEAF_ALLOCATOR;
   

	 log_status(0,"Simple pruning is selected.");
	 typedef Bodon::FrequentPairInserter<DF_D, TRIE, LEAF_WC, 
	    LEAF_ALLOCATOR, NEE_Off> FPI;
	 log_status(0,"deadend option is off.");
	 const bool deadend = false;
	 typedef Bodon::inhomogeneous_trie::IntersectProPruner<
	    DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off, deadend> PRUNER;

	 typedef Bodon::inhomogeneous_trie::CandidateGeneratorPrune<PRUNER, DF_D, 
	    TRIE, LEAF_ALLOCATOR, NEE_Off, deadend> CG;
	 typedef Bodon::inhomogeneous_trie::InfreqRemoverOrderPresAssumption<
	    DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR> IR;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder, fii);
	    log_status(0,"Finding frequent itemsets.");
	    apriori.findFrequentItemsets( 
	       nr_of_transactions, *par_c.freq_counters,
	       freq_pairs_with_counters, min_supp, maxsize );
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


