#ifndef RoutingSelector_HPP
#define RoutingSelector_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/input/transaction_reader/OrderReverser.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"

#include "util/StreamParser.hpp"
#include "common/allocators.hpp"

#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/inhomogeneous_trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"

#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge.hpp"
//#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge2.hpp"
//#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge3.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupTrans.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupTransLinBin.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupTransOffsetBitVec.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupTransOffsetIndex.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupEdge.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterLookupEdgePrevMem.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterHybridSimple.hpp"
#include "apriori/Apriori.hpp"

#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentPairInserterNoprune.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/SupportCounter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"

template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D> 
class RoutingSelector
{
   public:
      RoutingSelector( 
	 counter_t min_supp, char* routing, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& 
	 freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);

   private:
      std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
};

template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D> 
class RoutingSelectorOffset
{

   public:
      RoutingSelectorOffset( 
	 counter_t min_supp, char* routing, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& 
	 freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);

   private:
      std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
};
template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D> 
class RoutingSelectorHybrid
{

   public:
      RoutingSelectorHybrid( 
	 counter_t min_supp, char* routing, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& 
	 freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);

   private:
      std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
};
template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D> 
class RoutingSelectorDouble
{


   public:
      RoutingSelectorDouble( 
	 counter_t min_supp, char* routing, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& 
	 freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder);
   private:

      template <unsigned int THRESHOLD, class S_C, 
		class FII, class FPI, class CG, class IR> void 
      selectHybrid(
	 S_C& sorted_coder, DF_D& df_decoder, counter_t nr_of_transactions,
	 std::vector<counter_t>& freq_counters, counter_t min_supp);

   private:
      std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
};
template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D>
RoutingSelector<TRIE, LEAF_WC, LEAF_ALLOCATOR, T_R, DF_D>::RoutingSelector( 
   counter_t min_supp, char* routing, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
   typedef OrderReverser< typename bracz::BuildTreeDBCache< 
      S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > >S_C;
//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);


   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.");
   fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE main_trie;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE_Off> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::FrequentPairInserterNoprune<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off> FPI;

   typedef Bodon::inhomogeneous_trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off > CGIR;
   if( strcmp(routing,"merge") == 0 || strcmp(routing,"default") == 0 )
   {
      log_status(0,"merge is selected for routing.")
      typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }

/*   else if( strcmp(routing,"merge2") == 0 || strcmp(routing,"default") == 0 )
   {
      log_status(0,"merge2 is selected for routing.")
      typedef Bodon::SupportCounterMerge2<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie,sorted_coder, df_decoder);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }

   else if( strcmp(routing,"merge3") == 0 )
   {
      log_status(0,"merge3 is selected for routing.")
      typedef Bodon::SupportCounterMerge3<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie,sorted_coder, df_decoder);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
	 }*/
   else if( strcmp(routing,"lookup_edge") == 0 )
   {
      log_status(0,"Lookup on edges is selected for routing.");
      typedef Bodon::SupportCounterLookupEdge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie,sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }

   else if( strcmp(routing,"lookup_edge_prev_mem") == 0 )
   {
      log_status(0,"Lookup on edges with previously found memorization is selected for routing.");
      typedef Bodon::SupportCounterLookupEdgePrevMem<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie,sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strstr(routing,"bitvector") )
   {
      log_status(0,"Bitvector-based lookup on transaction is selected for routing.");
      typedef Bodon::SupportCounterLookupTransOffsetBitVec<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie,sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else if( strstr(routing,"indexvector") )
   {
      log_status(0,"Indexvector-based lookup on transaction is selected for routing.");
      typedef Bodon::SupportCounterLookupTransOffsetIndex<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
	 }
   else if( !strcmp(routing,"lookup_tr") )
   {
      log_status(0,"Lookup on transaction is selected for routing.");
      typedef Bodon::SupportCounterLookupTrans<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
/*
   else if( strstr(routing,"lookup_tr_linbin") )
   {
      log_status(0,"Speed-up ookup on transaction is selected for routing.");
      char* threshold = routing + strlen("lookup_tr_linbin_");
      if(strcmp(threshold,"3") ==0)
      {
	 log_status(0,"Threshold is set to: 3.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 3> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"6") ==0)
      {
	 log_status(0,"Threshold is set to: 6.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 6> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"9") ==0)
      {
	 log_status(0,"Threshold is set to: 9.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 9> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"12") ==0)
      {
	 log_status(0,"Threshold is set to: 12.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 12> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"16") ==0)
      {
	 log_status(0,"Threshold is set to: 16.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 16> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"22") ==0)
      {
	 log_status(0,"Threshold is set to: 22.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 22> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"32") ==0)
      {
	 log_status(0,"Threshold is set to: 32.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 32> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"45") ==0)
      {
	 log_status(0,"Threshold is set to: 45.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 45> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      else if(strcmp(threshold,"100") ==0)
      {
	 log_status(0,"Threshold is set to: 100.");
	 typedef Bodon::SupportCounterLookupTransLinBin<TRIE, 100> SUPP_C;
	 typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
	 A apriori(main_trie, sorted_coder, df_decoder);
	 log_status(0,"Finding frequent itemsets.");
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp );
      }
      }*/
   else 
      log_warn(0,"Unknown routing type requested '%s'.",routing);
}
template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D>
RoutingSelectorOffset<TRIE, LEAF_WC, LEAF_ALLOCATOR, T_R, DF_D>::RoutingSelectorOffset( 
   counter_t min_supp, char* routing, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
   typedef OrderReverser< typename bracz::BuildTreeDBCache< 
      S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > >S_C;
//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);


   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE main_trie;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE_Off> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::FrequentPairInserterNoprune<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off> FPI;
   typedef Bodon::inhomogeneous_trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off > CGIR;

   if( strcmp(routing,"lookup_edge") == 0 || strcmp(routing,"default") == 0 )
   {
      log_status(0,"Lookup on edges is selected for routing.");
      typedef Bodon::SupportCounterLookupEdge<TRIE> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else 
   {
      log_warn(0,"Only 'lookup_edge' routing strategy");
      log_warn(0,"can be selected for offsetindex edge representation.");
   }
}

template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D>
RoutingSelectorHybrid<TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, T_R, DF_D>::RoutingSelectorHybrid( 
   counter_t min_supp, char* routing, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
   typedef OrderReverser< typename bracz::BuildTreeDBCache< 
      S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > >S_C;

//   typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);


   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE_OEL main_trie;
   typedef Bodon::dynamic_trie::FrequentItemInserter<DF_D, TRIE_OEL, NEE_Off> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::dynamic_trie::FrequentPairInserterNoprune<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE_Off> FPI;

   typedef Bodon::dynamic_trie::CandGenInfreqRemoveNopruneMerge<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE_Off > CGIR;

   if( strcmp(routing,"hybrid") == 0 || strcmp(routing,"default") == 0)
   {
      log_status(0,"Hybrid strategy is selected for routing.");
      typedef Bodon::dynamic_trie::SupportCounter<TRIE_OEL, TRIE_OI> SUPP_C_BASE;
      typedef OneByOneSupportCounter<TRIE_OEL, SUPP_C_BASE, S_C> SUPP_C;
      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
   }
   else 
   {
      log_warn(0,"Only 'hybrid' routing strategy");
      log_warn(0,"can be selected for hybrid edge representation.");
   }
}

template <class TRIE, class LEAF_WC, class LEAF_ALLOCATOR, class T_R, class DF_D>
RoutingSelectorDouble<TRIE, LEAF_WC, LEAF_ALLOCATOR, T_R, DF_D>::RoutingSelectorDouble( 
   counter_t min_supp, char* routing, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder)
{
   typedef typename bracz::BuildTreeDBCache< SortedTransactionReader<Coder<T_R, DF_D>, false>,  
      std::vector<item_t>, bracz::EndPatriciaBuildTree<true> > S_C;

   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);


   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE_Off> FII;
   typedef Bodon::FrequentPairInserterNoprune<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off> FPI;

   typedef Bodon::inhomogeneous_trie::
      CandGenInfreqRemoveNopruneMerge<DF_D, TRIE, LEAF_WC, LEAF_ALLOCATOR, NEE_Off > CGIR;
   if( strstr(routing,"hybrid_simple") )
   {
      log_status(0,"Hybrid solution is selected for routing.");
      char* threshold = routing + strlen("hybrid_simple_");
      if(strcmp(threshold,"6") ==0)
      {
	 log_status(0,"Threshold is set to: 6.");
	 selectHybrid<6, S_C, FII, FPI, CGIR, CGIR>(
	    sorted_coder, df_decoder, nr_of_transactions, 
	    *par_c.freq_counters, min_supp);
      }
      else if(strcmp(threshold,"10") ==0)
      {
	 log_status(0,"Threshold is set to: 10.");
	 selectHybrid<10, S_C, FII, FPI, CGIR, CGIR>(
	    sorted_coder, df_decoder, nr_of_transactions, 
	    *par_c.freq_counters, min_supp);
      }
      else if(strcmp(threshold,"30") ==0)
      {
	 log_status(0,"Threshold is set to: 30.");
	 selectHybrid<30, S_C, FII, FPI, CGIR, CGIR>(
	    sorted_coder, df_decoder, nr_of_transactions, 
	    *par_c.freq_counters, min_supp);
      }
      else if(strcmp(threshold,"100") ==0)
      {
	 log_status(0,"Threshold is set to: 100.");
	 selectHybrid<100, S_C, FII, FPI, CGIR, CGIR>(
	    sorted_coder, df_decoder, nr_of_transactions, 
	    *par_c.freq_counters, min_supp);
      }
      else
      {
	 log_warn(0,"Bad threshold value was selected!");
	 log_warn(0,"Please select 6, 10, 20, 30, 40, 100!");
      }
   }
   else 
   {
      log_warn(0,"For 'double edge representation' only 'hybrid_simple'");
      log_warn(0,"routing strategy can be selected.");
   }
}

#endif
