// LightVector_set_difference_Test.cpp

#include "datastructures/vector/LightVector.hpp"
#include "datastructures/vector/set_difference.hpp"
#include <iostream>
using namespace std;

/**
 * Simple test for set_difference algorithm on a LightVector.
 *
 * @author Lars Schmidt-Thieme
 */
int main(int argc, char** argv) {

  int values[] = { 1, 3, 5, 7, 9, 11, 13 };
  int len = 7;
  int values2[] = { 1, 2, 4, 6, 7, 9 };
  int len2 = 6;

  const LightVector<int> x(values, len), y(values2, len2);
  LightVector<int>* res = LightVector<int>::create(10, 0);

  cout << "x = " << x << endl;
  cout << "y = " << y << endl;
#if DEBUG_LEVEL > 0
  cout << "res.capacity = " << res->capacity() << endl;
#endif

  // 1. ordinary set difference:
  set_difference(x, y, *res);
  cout << "set_difference(x,y) = " << *res << endl;

  // 2. with lower or upper size bounds:
  bool isBelowUpperSizeBound = set_difference_upper_size_bound(x, y, *res, 3);
  cout << "set_difference_upper_size_bound(x,y,3) = " << *res << ", isBelowUpperSizeBound = " << isBelowUpperSizeBound << endl;

  bool isBelowUpperSizeBound2 = set_difference_upper_size_bound(x, y, *res, 4);
  cout << "set_difference_upper_size_bound(x,y,4) = " << *res << ", isBelowUpperSizeBound = " << isBelowUpperSizeBound2 << endl;

  bool isAboveLowerSizeBound = set_difference_lower_size_bound(x, y, *res, 5);
  cout << "set_difference_lower_size_bound(x,y,5) = " << *res << ", isAboveLowerSizeBound = " << isAboveLowerSizeBound << endl;

  bool isAboveLowerSizeBound2 = set_difference_lower_size_bound(x, y, *res, 4);
  cout << "set_difference_lower_size_bound(x,y,4) = " << *res << ", isAboveLowerSizeBound = " << isAboveLowerSizeBound2 << endl;

  // 3. just result sizes:
  cout << "|set_difference(x,y)| = " << size_of_set_difference(x, y) << endl;

  cout << "|set_difference_upper_size_bound(x,y,3)| = " << size_of_set_difference_upper_size_bound(x, y, 3) << endl;
  cout << "|set_difference_upper_size_bound(x,y,4)| = " << size_of_set_difference_upper_size_bound(x, y, 4) << endl;
  cout << "|set_difference_lower_size_bound(x,y,5)| = " << size_of_set_difference_lower_size_bound(x, y, 5) << endl;
  cout << "|set_difference_lower_size_bound(x,y,4)| = " << size_of_set_difference_lower_size_bound(x, y, 4) << endl;


}
