
/**
  *@author Ferenc Bodon
  */

#include "common/log.h"
#include "test/io/codec/decoder/DecoderOutputTester.hpp"

#include "io/output/normal/BufferedNormalOutput.hpp"
#include "io/codec/decoder/normal/SimpleNormalDecoder.hpp"
#include "io/codec/decoder/normal/CacheNormalDecoder.hpp"
#include "io/codec/decoder/df/CacheDFDecoder.hpp"
#include "io/codec/decoder/df/SimpleBufferedDFDecoder.hpp"

#include <ctime>
#include <iostream>


 void DecoderOutputTester::usageDecoderOutput()
{
   std::cerr << "\nUsage: decoder_output_test type writer set_size output_file [random_seed]\n ";
   std::cerr << " type\t {normal, df}"<<std::endl;   
   std::cerr << " If type = normal, then:"<<std::endl;   
   std::cerr << " writer\t {simple, cache}"<<std::endl;  
   std::cerr << " If type = df, then:"<<std::endl;    
   std::cerr << " writer\t {buffered, cache}"<<std::endl;   
}

int DecoderOutputTester::doTest(
   int argc, char *argv[] )
{
   if ( argc < 5 )
   {
      usageDecoderOutput();
      return 3;
   }
   else
   {
      unsigned int set_size = atoi(argv[3]);
      int rand_seed;
      log_info(0,"The size of the itemset is set to %d.", set_size);
      if(argc < 6)
	 rand_seed = static_cast<int>(time(0));
      else
	 rand_seed = atoi(argv[5]);
      log_info(0,"The random seed is set to %d.", rand_seed);
      srand(rand_seed);

      if( strcmp(argv[1],"normal") == 0 )
      {
	 if(strcmp(argv[2],"simple") == 0)
	 {
	    log_info(0,"Simple normal decoder is selected");
	    writeAllCodedSubsetsNormal< SimpleNormalDecoder< > >( 
	       set_size, argv[4] );
	 }
	 else if(strcmp(argv[2],"cache") == 0)
	 {
	    log_info(0,"Cache normal decoder is selected");
	    writeAllCodedSubsetsNormal< CacheNormalDecoder< > >( 
	       set_size, argv[4] );
	 } 
	 else
	 {
	    usageDecoderOutput();
	    log_err(0,"Unknown writer: %s", argv[2]);
	    return 2;
	 }
      }
      else if(strcmp(argv[1],"df") == 0)
      {
	 if(strcmp(argv[2],"buffered") == 0)
	 {
	    log_info(0,"Buffered dfo decoder is selected");
	    writeAllCodedSubsetsDF< SimpleBufferedDFDecoder< > >( 
	       set_size, argv[4] );
	 }
	 else if(strcmp(argv[2],"cache") == 0)
	 {
	    log_info(0,"Cache DF decoder is selected");
	    writeAllCodedSubsetsDF< CacheDFDecoder< > >( 
	       set_size, argv[4] );
	 }
	 else
	 {
	    usageDecoderOutput();
	    log_err(0,"Unknown writer: %s", argv[2]);
	    return 2;
	 }
      }
      else
      {
	 usageDecoderOutput();
	 log_err(0,"Unknown type: %s", argv[1]);
	 return 1;
      }
   }
   return 0;
}

