#ifndef DecoderOutputTester_HPP
#define DecoderOutputTester_HPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"
#include "io/FDRepr.hpp"
#include <vector>

class DecoderOutputTester
{
   public:
      void usageDecoderOutput();
      int doTest( int argc, char *argv[] );

   private:
      template <class D> inline void 
      writeAllCodedSubsetsNormal( const unsigned int set_size, char* file_name );
      template <class DF_D> inline void 
      writeAllCodedSubsetsDF( const unsigned int set_size, char* file_name );
};

template <class D> void DecoderOutputTester::
writeAllCodedSubsetsNormal( const unsigned int set_size, char* file_name )
{
   typename D::params_t par;
   par.file_name = file_name;
   par.mode=FileReprBase::WRITE;
   D decoder(&par);

   std::vector<item_t> code_inverse(set_size);
   for( unsigned int index=0; index < set_size; ++index )
      code_inverse[index] = rand();

   decoder.setCodeInverse(code_inverse);

   std::vector<item_t> transaction;
   std::vector<bool> pattern(set_size, false);
   std::vector<bool>::size_type index;

   counter_t support = 1; //we dont care about the support
   do
   {
      decoder.writeItemsetAndCounter( 
	 transaction.begin(), transaction.end(), support);
      index=0;
      while(index < set_size && pattern[index] == true )
      {
	 pattern[index] = false;
	 transaction.pop_back();
	 index++;
      }
      if( index == set_size )
	 break;
      else
      {
	 pattern[index] = true;
	 transaction.push_back(index);
      }
   }
   while(true);
}
template <class DF_D> void DecoderOutputTester::
writeAllCodedSubsetsDF( const unsigned int set_size, char* file_name )
{
   typename DF_D::params_t par;
   par.file_name = file_name;
   par.mode=FileReprBase::WRITE;
   DF_D df_decoder(&par);

   std::vector<item_t> code_inverse(set_size);
   for( unsigned int index=0; index < set_size; ++index )
      code_inverse[index] = rand();

   df_decoder.setCodeInverse(code_inverse);

   std::vector<bool> pattern(set_size, false);
   std::vector<bool>::size_type index;

   counter_t support = 1; //we dont care about the support
   do
   {
      df_decoder.write(support);
      index=0;
      while(index < set_size && pattern[index] == true )
      {
	 pattern[index] = false;
	 df_decoder.popItem();
	 index++;
      }
      if( index == set_size )
	 break;
      else
      {
	 pattern[index] = true;
	 df_decoder.pushItem(index);
      }
   }
   while(true);
}

#endif
