#ifndef DBCacheTester_HPP
#define DBCacheTester_HPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"
#include "common/log.h"
#include "io/FDRepr.hpp"
#include "io/codec/decoder/normal/CacheNormalDecoder.hpp"
#include "io/codec/coder/Coder.hpp"
#include <vector>




class DBCacheTester
{
   public:
      void usageDBCacheTester();
      void doTest(int argc, char *argv[] );
   private:
      typedef CacheNormalDecoder<> D;

      template <class CACHER> void selectMode(
	 char* codemode, char* file_name, int nr_of_traversals,
	 item_t largest_item, D* decoder,
	 std::vector< std::pair<counter_t, item_t> >* freq_items_with_counters);

      template <class CACHER, CodeMode MODE> void cacheDatabase(
	 char* file_name, int nr_of_traversals, item_t largest_item, D* decoder,
	 std::vector< std::pair<counter_t, item_t> >* freq_items_with_counters);

};


template <class CACHER> void DBCacheTester::selectMode(
   char* codemode, char* file_name, int nr_of_traversals,
   item_t largest_item, D* decoder,
   std::vector< std::pair<counter_t, item_t> >* freq_items_with_counters)
{
   if( strcmp(codemode, "ASC") ==0 )
   {
      log_info(0,"Ascending order is chosen.");
      cacheDatabase<CACHER, ASC>( file_name, nr_of_traversals, 
				  largest_item, decoder,
				  freq_items_with_counters);
   }
   else if( strcmp(codemode, "DESC") ==0 )
   {
      log_info(0,"Descending order is chosen.");
      cacheDatabase<CACHER, DESC>( file_name, nr_of_traversals, 
				   largest_item, decoder,
				   freq_items_with_counters);
   }
   else if( strcmp(codemode, "NOCARE") ==0 )
   {
      log_info(0,"Order is not going to be modified.");
      cacheDatabase<CACHER, NOCARE>( file_name, nr_of_traversals, 
				     largest_item, decoder,
				     freq_items_with_counters);
   }
   else
      usageDBCacheTester();
}


template <class CACHER, CodeMode MODE> void DBCacheTester::cacheDatabase(
   char* file_name, int nr_of_traversals, 
   item_t largest_item, D* decoder,
   std::vector< std::pair<counter_t, item_t> >* freq_items_with_counters)
{
   typename CACHER::params_t par;
   par.file_name = file_name;
   par.mode=FileReprBase::READ;
   par.largest_item = largest_item;
   par.decoder = decoder;
   par.freq_items_with_counters = freq_items_with_counters;
   par.codemode = MODE;

   log_info(0,"Doing database cacher.");
   CACHER t_r(&par);
   std::cout<<"\t\t\t\t\tDone!"<<std::endl;

   std::cout<<"Traversing through the cached transactions "
	    << nr_of_traversals<<" times."<<std::endl;
   std::vector<item_t> transaction;
   counter_t multiplicity;
   for(int trav_index = 1; trav_index <= nr_of_traversals; ++trav_index)
   {
      std::cout<<trav_index<<". travel"<<std::flush;
      t_r.rewind();
      while( (multiplicity = t_r.nextTransactionBIS(
		 transaction)) !=0 )
      {
	 // uncomment the following lines if you want to see the output
	 decoder->writeItemsetAndCounter( 
	    transaction.begin(), transaction.end(), multiplicity );
      }
      std::cout<<"\t\t\t\t\t\tDone!"<<std::endl;
   }
}

#endif
