
/**
  *@author Ferenc Bodon
  */

#include "test/io/input/InputreaderTester.hpp"
#include "io/input/transaction_reader/SimpleTransactionReader.hpp"
#include "io/StreamRepr.hpp"
#include "io/FILERepr.hpp"
#include "io/FDRepr.hpp"

#include <iostream>

void InputreaderTester::usageSimpleInput()
{
   std::cerr << "\nUsage: input_test file_repr reader input_file_name lowlevel_buffer_size [nr_of_scan]\n ";
   std::cerr << " The possible file representations are:"<<std::endl;
   std::cerr << " file_repr\t {filedes, FILE, stream}"<<std::endl;
   std::cerr << " reader\t {simple, lars, bracz}"<<std::endl;
   std::cerr << " The size of the low level buffer will be lowlevel_buffer_size."<<std::endl;
}

void InputreaderTester::doTest( int argc, char *argv[] )
{
   if( argc < 5 )
      usageSimpleInput();
   else
   {
      std::streamsize lowlevel_buffer_size = atoi(argv[4]);
      log_info(0,"Low level buffer size is set to %d.", lowlevel_buffer_size);
      unsigned int nr_of_scan;
      if(argc == 6)
	 nr_of_scan = atoi(argv[5]);
      else 
	 // Most FIM algorithms read the original transaction database two times.
	 nr_of_scan = 2;
      log_info(0,"The number of scan is set to %d.", nr_of_scan);
      if( strcmp(argv[2], "simple") == 0)
      {
	 log_info(0,"Simple transaction reader is selected for test.");
	 if(strcmp(argv[1],"stream") == 0)
	    readDatabase<SimpleTransactionReader>(argv[3], nr_of_scan, lowlevel_buffer_size);
	 else
	    std::cerr<<"\nOnly stream file representation is implemented for simple reader ;("<<std::endl;
      }
      else if( strcmp(argv[1],"filedes") == 0) 
      {
	 log_info(0,"filedes is selected as a file representation.");
	 selectReader<FDRepr>(argv[2], argv[3], nr_of_scan, lowlevel_buffer_size);
      }
      else if( strcmp(argv[1],"FILE") == 0) 
      {
	 log_info(0,"FILE* is selected as a file representation.");
	 selectReader<FILERepr>(argv[2], argv[3], nr_of_scan, lowlevel_buffer_size);
      }
      else if( strcmp(argv[1],"stream") == 0)
      {
	 log_info(0,"std::ifstream is selected as a file representation.");
	 selectReader<StreamRepr>(argv[2], argv[3], nr_of_scan, lowlevel_buffer_size);
      }
      else
	 usageSimpleInput();
   }
}
