/**
  *@author Ferenc Bodon
  */

#include "common/log.h"
#include "test/io/output/SimpleOutputTester.hpp"
#include "io/StreamRepr.hpp"
#include "io/FILERepr.hpp"
#include "io/FDRepr.hpp"
#include "io/output/normal/SimpleOutput.hpp"
#include "io/output/normal/BufferedNormalOutput.hpp"

#include <iostream>

void SimpleOutputTester::usageSimpleOutput()
{
   std::cerr << "\nUsage: simple_output_test file_repr writer ";
   std::cerr << "ouput_file set_size lowlevel_buffer_size highlevel_buffer_size [random_seed]"<<std::endl;
   std::cerr << " file_repr\t {filedes, FILE, stream}"<<std::endl;
   std::cerr << " writer\t {simple, buffered}";
   std::cerr << " The size of the low level buffer will be lowlevel_buffer_size."<<std::endl;
   std::cerr << " The size of the high level buffer (in case of buffered writer) ";
   std::cerr << " will be highlevel_buffer_size."<<std::endl;
}

void SimpleOutputTester::doTest(
   int argc, char *argv[]) throw(std::ios_base::failure)
{
   if ( argc < 7 )
      usageSimpleOutput();
   else
   {
      unsigned int set_size = atoi(argv[4]);
      log_info(0,"The size of the itemset is set to %d.", set_size);
      std::streamsize lowlevel_buffer_size = atoi(argv[5]);
      log_info(0,"Low level buffer size is set to %d.", lowlevel_buffer_size);
      std::streamsize highlevel_buffer_size = atoi(argv[6]);
      log_info(0,"High level buffer size is set to %d.", highlevel_buffer_size);
      int rand_seed;
      if(argc < 8)
	 rand_seed = static_cast<int>(time(0));
      else
	 rand_seed = atoi(argv[7]);
      srand(rand_seed);
      log_info(0,"The random seed is set to %d.", rand_seed);

      if( strcmp(argv[2],"simple") == 0 )
      {
	 log_info(0,"Simple writer is selected for test.");
	 log_info(0,"High level buffer size parameter is ignored.");
	 if(strcmp(argv[1],"stream") == 0)
	 {
	    SimpleOutput::params_t par;
	    par.file_name = argv[3];
	    par.mode = FileReprBase::WRITE;
	    par.file_buffer_size = lowlevel_buffer_size;
	    SimpleOutput o_manager(&par);
	    writeAllSubsets( 
	       o_manager, set_size, rand_seed );
	 }
	 else
	    std::cerr<<"\nOnly stream file representation is implemented for simple writer ;("<<std::endl;
      }
      else if( strcmp(argv[2],"buffered") == 0 )
      {
	 log_info(0,"Buffered writer is selected for test.");
	 if( strcmp(argv[1],"filedes") == 0 )
	 {
	    log_info(0,"filedes is selected as a file representation.");
	    makeBufferedWriter< BufferedNormalOutput< OutputBase<FDRepr> > >( 
	       argv[3], lowlevel_buffer_size, highlevel_buffer_size, set_size, rand_seed );
	 }
	 else if( strcmp(argv[1],"FILE") == 0 )
	 {
	    log_info(0,"FILE* is selected as a file representation.");
	    makeBufferedWriter< BufferedNormalOutput< OutputBase<FILERepr> > >( 
	       argv[3], lowlevel_buffer_size, highlevel_buffer_size, set_size, rand_seed );
	 }
	 else if( strcmp(argv[1],"stream") == 0 )
	 {
	    log_info(0,"std::ofstream is selected as a file representation.");
	    makeBufferedWriter< BufferedNormalOutput< OutputBase<StreamRepr> > >( 
	       argv[3], lowlevel_buffer_size, highlevel_buffer_size, set_size,rand_seed );
	 }
	 else
	    usageSimpleOutput();
      }
      else
	 usageSimpleOutput();
   }
}
