#ifndef SimpleOutputTester_HPP
#define SimpleOutputTester_HPP

/**
  *@author Ferenc Bodon
  */

#include "io/FileReprBase.hpp"
#include "common.hpp"
#include "common/log.h"
#include <vector>

/**
   The purpose of this class is to measure the effect of buffering
   in outputting the results, and the effect of the differrent file 
   representation.

   In the test we write out all subset of a set, whose size is given by
   a parameter. The element of the itemset will be random numbers.
   The random seed can be given by a parameter.
 */
class SimpleOutputTester
{
   public:
      void usageSimpleOutput();
      void doTest( int argc, char *argv[] )
	 throw(std::ios_base::failure);
   private:
      template <class O_M> void makeBufferedWriter( 
	 char* file_name, const std::streamsize lowlevel_buffer_size,
	 const std::streamsize highlevel_buffer_size, const unsigned int set_size,
	 const int rand_seed) throw(std::ios_base::failure);

      /// This function writes out all subset of an itemset.
      template <class O_M> void writeAllSubsets( 
	 O_M& o_manager, const unsigned int set_size,
	 const int rand_seed );
	 
};

template <class O_M> void SimpleOutputTester::
makeBufferedWriter( char* file_name, const std::streamsize lowlevel_buffer_size,
		    const std::streamsize highlevel_buffer_size, const unsigned int set_size,
		    const int rand_seed) throw(std::ios_base::failure)
{
   typename O_M::params_t par;
   par.file_name = file_name;
   par.mode = FileReprBase::WRITE;
   par.file_buffer_size = lowlevel_buffer_size;
   par.buffer_size = highlevel_buffer_size;
   O_M o_manager(&par);
   writeAllSubsets( o_manager, set_size,
		       rand_seed );
}
/**
   \param file_name the name of the file output is written to.
   \param set_size the size of the set whose subsets have to be written out.
   \param lowlevel_buffer_size the low level buffer size
   \param rand_seed the seed for random number generation
 */
template <class O_M> void SimpleOutputTester::
writeAllSubsets( O_M& o_manager, const unsigned int set_size,
		 const int rand_seed )
{
   std::vector<item_t> itemset(set_size);
   std::vector<item_t> subset;
   std::vector<bool> pattern(set_size, false);
   std::vector<bool>::size_type index;

   log_info(0,"Generating random items...");
   for( unsigned int index=0; index < set_size; ++index )
      itemset[index] = rand();

   counter_t support = 1; //we dont care about the support
   log_info(0,"Generating and writting out the subsets...");
   do
   {
      o_manager.writeItemsetAndCounter( 
	 subset.begin(), subset.end(), support);
      index=0;
      while(index < set_size && pattern[index] == true )
      {
	 pattern[index] = false;
	 subset.pop_back();
	 index++;
      }
      if( index == set_size )
	 break;
      else
      {
	 pattern[index] = true;
	 subset.push_back(itemset[index]);
      }
   }
   while(true);
}

#endif
