#ifndef Formatter_CPP
#define Formatter_CPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"
#include <fstream>

/// This class provides some basic int to string conversion methods
class Formatter
{
   public:
 
      /// Convers an integer to string; the characters are
      /// placed in a backward manner.
      template <class T> static void IntToStringBack( 
	 T n, char* buffer, std::streamsize& pos );

      /// Convers an integer to string; the characters are
      /// placed in a forward manner.
      template <class T> static void IntToStringFront( 
	 T n, char* buffer, std::streamsize& pos );

      /// Generates the string representation of a support, i.e
      /// it converts the int to string and puts parenthesis around it
      static void SupportToStringBack( 
	 counter_t support, char* buffer, std::streamsize& pos );

      /// Convers a sequence of integers to string; the characters are
      /// placed in a backward manner.
      template <typename InputIterator> static void ItemsetToStringBack( 
	 InputIterator first, InputIterator last, char* buffer, 
	 std::streamsize& pos );


   private: 
};
/**
   \param n the integer
   \param buffer the character buffer that will store the string
   representation of the integer
   \param pos this parameter gives the position of the buffer where the last
   character of the representation is put. When the function 
   exits, \c pos will store the position before the first character
   of the representation.
 */
template <class T> inline void Formatter::IntToStringBack( 
   T n, char* buffer, std::streamsize& pos )
{
   if( n == 0 ) 
      buffer[--pos] = '0';
   else 
   {
      while( n > 0 ) 
      {
	 buffer[--pos] = n % 10 + '0';
	 n /= 10;
      }
   }
}

/**
   \param n the integer
   \param buffer the character buffer that will store the string
   representation of the integer
   \param pos this parameter gives the position of the buffer where the first
   character of the representation is put. When the function 
   exits, \c pos will store the position after the last character
   of the representation.
 */
template <class T> inline void Formatter::IntToStringFront( 
   T n, char* buffer, std::streamsize& pos )
{
   if( n == 0 ) 
      buffer[pos++] = '0';
   else
   {
      size_t length=0;
      {
	 register T n2(n);
	 while(n2)
	 {
	    length++;
	    n2 /= 10;
	 }
      }
      char* buffer2 = buffer + pos + length -1;
      while(n) 
      {
	 *buffer2-- = '0' + n%10;
	 n /= 10;
      }
      pos += length;
   }	 
}
/**
   \param support the support value
   \param buffer the character buffer that will store the string
   representation
   \param pos this parameter gives the position of the buffer where 
   the closing parenthesis is put. When the function 
   exits, \c pos will store the position before opening parenthesis.
 */
inline void Formatter::SupportToStringBack( 
   counter_t support, char* buffer, std::streamsize& pos )
{
   buffer[--pos] = '\n';
   buffer[--pos] = ')';

   while (support > 0)
   {
      buffer[--pos] = support % 10 + '0';
      support /= 10;
   }
   buffer[--pos] = '(';
}

template <typename InputIterator> inline void Formatter::ItemsetToStringBack( 
   InputIterator first, InputIterator last, char* buffer, 
   std::streamsize& pos )
{
   while(last != first)
   {
      buffer[--pos] = ' ';
      --last;
      IntToStringBack(*last, buffer, pos);
   }
}



#endif
