#ifndef Frequent2FilterOnline_CPP
#define Frequent2FilterOnline_CPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"  
#include <vector>
//#include <iostream> //for tests

/** This class finds the frequent pairs with the array technique.

   \param IT_R the type of the transaction reader.
 */
template <class IT_R>
class Frequent2FilterOnline
{
   public:
      Frequent2FilterOnline<IT_R>(IT_R* it_r):
	 it_r(it_r), largest_item(it_r->getLargestItem()) { }

      /// This function finds the frequent pairs
      void findFrequentPairs(
	 std::vector< std::pair< counter_t, std::pair<item_t, item_t> > >& 
	 freq_pairs_with_counters,
	 counter_t min_supp);


   private:
      IT_R* it_r;
      const item_t largest_item;
};

/**
   \param freq_pairs_with_counters the vector that stores the frequent 
   pairs. Each element of a vector is a pair, where the first element
   is the counter, and the second element stores the elements of the pair 
   in an ascending order.
   \param min_supp the support threshold
 */
template <class IT_R> inline void Frequent2FilterOnline<IT_R>::
findFrequentPairs(
   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >& 
   freq_pairs_with_counters, counter_t min_supp)
{
   std::vector< std::vector< std::pair<item_t, counter_t> > > temp_counter_array;
   temp_counter_array.reserve(largest_item);
   temp_counter_array.resize(largest_item);

   freq_pairs_with_counters.clear();
   it_r->rewind();

   std::vector<item_t> transaction;
   std::vector<item_t>::iterator it1, it2;
   std::vector< std::pair<item_t, counter_t> >::iterator it_row;
   counter_t counter;
   /// Determining the support of the ipairs
   while( (counter = it_r->nextTransactionBIS( transaction )) != 0 )
   {      
      if( transaction.size() > 1)
      {
	 for( it1 = transaction.begin(); it1 != transaction.end() - 1; ++it1)
	 {
	    it_row = temp_counter_array[*it1].begin();
	    for( it2 = it1+1; it2 != transaction.end(); ++it2)
	    {
	       while( it_row != temp_counter_array[*it1].end() 
		      && (*it_row).first < *it2 )
		  ++it_row;
	       if( it_row != temp_counter_array[*it1].end() &&
		  (*it_row).first == *it2)
	       {
		  (*it_row).second += counter;
		  ++it_row;
	       }
	       else
	       {
		  it_row = temp_counter_array[*it1].insert(
		     it_row, std::pair<item_t, counter_t>(*it2, counter) );
		  ++it_row;
	       }
	    }
	 }
      }
   }
   /// Finding the frequent pairs
   item_t index1;
   std::pair<item_t, item_t> temp_pair;
   std::pair< counter_t, std::pair<item_t, item_t> > temp_pair2;
   for( index1 = 0; index1 < largest_item ; ++index1 )
   {
      for( it_row = temp_counter_array[index1].begin(); 
	   it_row != temp_counter_array[index1].end(); ++it_row )
      {
	 if( (*it_row).second >= min_supp )
	 {
	    temp_pair.first = index1;
	    temp_pair.second = (*it_row).first;
	    temp_pair2.first = (*it_row).second;
	    temp_pair2.second = temp_pair;
	    freq_pairs_with_counters.push_back(temp_pair2);
	 }
      }
      temp_counter_array[index1].clear();
      std::vector< std::pair<item_t, counter_t> >().swap(temp_counter_array[index1]);   
      // temp_counter_array[index1] will never be used again!
   }
}

#endif
